import logging
import socket
import threading
from threading import Timer

import spread_core.mqtt as mqtt
import spread_core.tools.settings as settings
from spread_core.tools.service_launcher import Launcher


class TcpSocket:
    HOST = None
    PORT = None
    TIMEOUT = None
    KILL_TIMEOUT = None

    def __init__(self):
        self._socket = None
        self._killer = None
        self._creator = None

    def update_config(self):
        self.HOST = settings.config[settings.BUS][settings.HOST]
        self.PORT = settings.config[settings.BUS][settings.PORT]
        self.TIMEOUT = settings.config[settings.BUS][settings.TIMEOUT]
        self.KILL_TIMEOUT = settings.config[settings.BUS][settings.KILL_TIMEOUT]

    def create(self):
        self.stop_creator()
        self._creator = threading.Timer(self.TIMEOUT, self._create)
        self._creator.start()

    def stop_creator(self):
        if isinstance(self._creator, threading.Timer):
            self._creator.cancel()
            self._creator = None

    def _create(self):
        if self._socket is None:
            logging.debug('CONNECT: try')
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.settimeout(self.TIMEOUT)
            self._socket.connect((self.HOST, self.PORT))
        except BaseException as ex:
            if self._socket:
                logging.info('CONNECT: ' + str(ex))
            self.create()
        else:
            logging.debug('CONNECT: SUCCESS')
            self.stop_creator()

    def start_timer(self):
        if self.KILL_TIMEOUT > 0:
            self._killer = Timer(self.KILL_TIMEOUT, self.kill)
            self._killer.start()

    def stop_timer(self):
        if self._killer:
            self._killer.cancel()
            self._killer = None

    def kill(self):
        if isinstance(self._socket, socket.socket):
            logging.debug('Kill socket')
            self._socket.close()
            self._socket = None

    def send_message(self, data, r_size):
        self.stop_timer()
        out = b''
        if self._socket is None:
            self._create()
        if self._socket:
            try:
                self._socket.send(data)
            except BaseException as ex:
                logging.error('Send error: ' + str(ex))
                if self._creator is None:
                    self.create()
            try:
                while len(out) < r_size:
                    out += self._socket.recv(r_size - len(out))
            except socket.timeout:
                pass
            except BaseException as ex:
                logging.error('Recv error: ' + ' '.join([hex(b)[2:].rjust(2, '0') for b in out]) + ' ' + str(ex))
                if self._creator is None:
                    self.create()
        self.start_timer()
        return out


class TCPAdapterLauncher(Launcher):
    _dumped = False
    _def_start_delay = 0

    def __init__(self):
        self._manager = self
        self.sock = TcpSocket()
        self._dump_topic = None
        self._send_topic = None

        super(TCPAdapterLauncher, self).__init__()

    def start(self):
        self.mqttc.subscribe(self._send_topic)
        logging.debug('Subscribed to {}'.format(self._send_topic))

    def on_message(self, mosq, obj, msg):
        data, flags = msg.payload.decode().split('#')
        data = bytes.fromhex(data)
        data += crc16(data)
        flags = flags.split(':')
        size = 0
        for flag in flags:
            if 'RS' in flag:
                size = int(flag[2:])

        logging.debug('[->  ]: {}'.format(' '.join(hex(b)[2:].rjust(2, '0') for b in data)).upper())
        out = self.sock.send_message(data, size)
        topic = self._dump_topic
        out_str = ' '.join(hex(b)[2:].rjust(2, '0') for b in out).upper()
        if len(out) == 0:
            payload = '#W'
        else:

            if out[-2:] != crc16(out[:-2]):
                logging.warning('Неверная контрольная сумма ответа!')
                payload = '#E'
            else:
                # без crc16!!
                payload = out_str[:-6]

        logging.debug('[  <-]: {}'.format(payload))
        self.mqttc.publish(topic=topic, payload=payload)

    def on_exit(self, sig, frame):
        self.sock.stop_timer()
        self.sock.stop_creator()
        super(TCPAdapterLauncher, self).on_exit(sig, frame)

    def update_config(self, sig=None, frame=None):
        super(TCPAdapterLauncher, self).update_config(sig, frame)
        self._send_topic = str(mqtt.bus.TopicBus.of_send(settings.config[settings.BUS][settings.SUBTOPIC]))
        self._dump_topic = str(mqtt.bus.TopicBus.of_dump(settings.config[settings.BUS][settings.SUBTOPIC]))
        self.sock.update_config()


def crc16(data):
    crc = 0xFFFF
    l = len(data)
    i = 0
    while i < l:
        j = 0
        crc = crc ^ data[i]
        while j < 8:
            if crc & 0x1:
                mask = 0xA001
            else:
                mask = 0x00
            crc = ((crc >> 1) & 0x7FFF) ^ mask
            j += 1
        i += 1
    if crc < 0:
        crc -= 256
    result = crc.to_bytes(2, 'little')
    return result


def run():
    TCPAdapterLauncher().connect()


if __name__ == '__main__':
    run()
