import logging
from datetime import datetime

from .providers import providers, lights, sensors
from ..managers import Manager
from ...protocols import dali
from ...protocols.dali.device import events
from ...protocols.dali.driver.dali_interface import DaliCanAdapter, BrokerAdapter
from ...tools.settings import config, LOCAL_ADAPTER, LEVEL_SYNC_DELAY

DALI1 = 1
DALI2 = 2


class Dali(Manager):
    DTR0 = 0
    DTR1 = 1
    DTR2 = 2

    level_sync_delay = 1

    def __init__(self, p_id, m_id):
        super().__init__(p_id, m_id)
        self._groups = (None, [[] for i in range(0, 16)], [[] for i in range(0, 32)])
        self._scan_time = None
        self._broadcast = [None, None, None]
        self.group_providers = (None, [None for i in range(0, 16)], [None for i in range(0, 32)])
        self.dtr_values = (None, [None, None, None], [None, None, None])

    def _sorted_info(self):
        _all = self._providers

        def s(_id):
            if _all[_id].dali_type == DALI2:
                return _all[_id].dali_type, 10-_all[_id].instance._value, _id
            else:
                return _all[_id].dali_type, _id

        return sorted(self._providers, key=s)

    def set_interface(self):
        self.send_interface = DaliCanAdapter() if config.get(LOCAL_ADAPTER, False) else BrokerAdapter(self.mqttc)
        logging.info('Bus interface is "%s"' % self._send_interface.__class__.__name__)

    def update_config(self):
        has_local_adapter = config.get(LOCAL_ADAPTER, False)
        super(Dali, self).update_config()
        if has_local_adapter != config.get(LOCAL_ADAPTER, False):
            self.set_interface()

        self.level_sync_delay = config.get(LEVEL_SYNC_DELAY, self.level_sync_delay)
        from spread_core.bam.dali.tuner import TunerLimited
        TunerLimited.update_config()

    @property
    def scan_time(self):
        if self._scan_time is None:
            self._scan_time = datetime.now()
        return self._scan_time

    def add_provider(self, data):
        super(Dali, self).add_provider(data)
        provider = self._providers[data['id']]
        if provider.is_device:
            addr = provider.binding.addr.address
            self._addresses.setdefault(addr, [])
            self._addresses[provider.binding.addr.address].append(provider.id)
        elif provider.is_group:
            self.group_providers[provider.dali_type][provider.binding.addr.group] = provider
        elif provider.is_broadcast:
            self._broadcast[provider.dali_type] = provider

    def broadcast(self, dali_type):
        return self._broadcast[dali_type]

    def add_to_group(self, dali_type, group_id, provider_id):
        if provider_id not in self._groups[dali_type][group_id]:
            self._groups[dali_type][group_id].append(provider_id)

    def remove_from_group(self, dali_type, group_id, provider_id):
        if provider_id in self._groups[dali_type][group_id]:
            self._groups[dali_type][group_id].remove(provider_id)

    def get_group_size(self, dali_type, group_id):
        return len(self._groups[dali_type][group_id])

    def get_group_members(self, dali_type, group_id):
        for p_id in self._groups[dali_type][group_id]:
            _provider = self.get_provider(p_id)
            if _provider:
                yield _provider

    def get_all_devices(self, dali_type):
        for _provider in self._providers.values():
            if _provider.is_device and _provider.dali_type == dali_type:
                yield _provider

    def on_external_command(self, command, response, *args):
        if isinstance(command, dali.gear.general.DTR0):
            self.dtr_values[DALI1][0] = command.param
        elif isinstance(command, dali.gear.general.DTR1):
            self.dtr_values[DALI1][1] = command.param
        elif isinstance(command, dali.gear.general.DTR2):
            self.dtr_values[DALI1][2] = command.param
        elif isinstance(command, dali.device.general.DTR0):
            self.dtr_values[DALI2][0] = command._opcode
        elif isinstance(command, dali.device.general.DTR1):
            self.dtr_values[DALI2][1] = command._opcode
        elif isinstance(command, dali.device.general.DTR2):
            self.dtr_values[DALI2][2] = command._opcode
        elif isinstance(command, dali.device.general.DTR1DTR0):
            self.dtr_values[DALI2][0] = command._opcode
            self.dtr_values[DALI2][1] = command._instance
        elif isinstance(command, dali.device.general.DTR2DTR1):
            self.dtr_values[DALI2][1] = command._opcode
            self.dtr_values[DALI2][2] = command._instance
        elif isinstance(command, (dali.gear.general._StandardCommand, dali.gear.general.DAPC)):
            if isinstance(command.destination, dali.address.Short):
                devices = self.providers_by_address(DALI1, command.destination.address)
            elif isinstance(command.destination, dali.address.Group):
                group = self.group_providers[DALI1][command.destination.group]
                if group:
                    devices = (group,)
                else:
                    devices = self.get_group_members(DALI1, command.destination.group)
            elif isinstance(command.destination, dali.address.Broadcast):
                if self._broadcast[DALI1]:
                    devices = (self._broadcast[DALI1],)
                else:
                    devices = self.get_all_devices(DALI1)
            else:
                devices = ()

            for _provider in devices:
                if _provider is None: continue
                if not isinstance(_provider, lights.DaliLight): continue

                if command.is_query:
                    [_provider.on_request_answer(funit_type, response, None, *args) for funit_type in command.funit_type]
                else:
                    if _provider.is_valid:
                        _provider.on_command_sended(command)

        elif isinstance(command, dali.device.general._DeviceCommand):
            # DALI_2
            if isinstance(command, dali.device.general._SpecialDeviceCommand):
                pass
            else:
                if isinstance(command.destination, dali.address.Short):
                    devices = self.providers_by_address(DALI2, command.destination.address)
                elif isinstance(command.destination, dali.address.Broadcast):
                    devices = self.get_all_devices(DALI2)
                elif isinstance(command.destination, dali.address.Group):
                    group = self.group_providers[DALI2][command.destination.group]
                    if group:
                        devices = (group,)
                    else:
                        devices = self.get_group_members(DALI2, command.destination.group)
                else:
                    devices = ()

                for _provider in devices:
                    if isinstance(command, dali.device.general._StandardInstanceCommand):
                        if not isinstance(command.instance, (dali.address.InstanceGroup, dali.address.InstanceBroadcast)):
                            if not _provider.compare_instance(command.instance):
                                continue

                    if command.is_query:
                        [_provider.on_request_answer(funit_type, response, None, *args) for funit_type in command.funit_type]
                    elif command.is_config:
                        if _provider.is_valid:
                            _provider.on_command_sended(command)

    def on_event(self, event_data):
        event = events.Event.of(event_data)
        providers = []

        if event.scheme_id == dali.device.events.SCHEME_ADDR_INST_TYPE:
            if event.short_addr.address in self._addresses:
                for p_id in self._addresses[event.short_addr.address]:
                    _provider = self._providers[p_id]
                    if isinstance(_provider, sensors.DaliSensor) and _provider.compare_instance(event.instance_type):
                        providers.append(_provider)
                        event.data = _provider.type_id
        elif event.scheme_id == dali.device.events.SCHEME_ADDR_INST_NUMBER:
            if event.short_addr.address in self._addresses:
                for p_id in self._addresses[event.short_addr.address]:
                    _provider = self._providers[p_id]
                    if isinstance(_provider, sensors.DaliSensor) and _provider.compare_instance(event.instance_number):
                        providers.append(_provider)
                        event.data = _provider.type_id

        logging.debug(f'<- {event}')

        for _p in providers:
            _p.on_event(event)

    def on_bus_status(self, ready):
        for dali_type in (DALI1, DALI2):
            for p in self.get_all_devices(dali_type):
                if isinstance(p, providers.DaliProvider):
                    p.set_valid(p.VALID if ready else p.INVALID)
                    if ready:
                        p.commit_states(ready)

    # def GetScan(self, sig):
    #     value = {"managerID": self.id,
    #              "managerType": self.__class__.__name__,
    #              "scanData": dict(devices=[], devices2=[]),
    #              "scanIndex": 0,
    #              "scanName": "",
    #              "scanTime": str(self.scan_time).replace(' ', 'T')}
    #
    #     for addr, ids in self._addresses.items():
    #         device2 = None
    #         for p_id in ids:
    #             if p_id in self._providers:
    #                 provider = self._providers[p_id]
    #                 d = provider.serialize()
    #                 if d:
    #                     if isinstance(provider, lights.DaliLight):
    #                         value['scanData']['devices'].append(d)
    #                     elif isinstance(provider, sensors.DaliSensor):
    #                         if device2 is None:
    #                             device2 = d
    #                             device2['instances'] = []
    #                         inst = provider.serialize_instance()
    #                         if inst:
    #                             device2['instances'].append(inst)
    #
    #         if device2:
    #             value['scanData']['devices2'].append(device2)
    #
    #     jocket = mqtt.spread.variable.Variable(value, key=sig.key)
    #     topic = mqtt.spread.topic.Reply(self.project_id, sig.session_id)
    #     self.publish(topic, jocket, retain=False)
