import logging

import math

from spread_core.mqtt.spread.variable import Variable
from spread_core.tools.timer import Timer
import spread_core.mqtt as mqtt
import spread_core.protocols.dali as dali
from spread_core.bam.dali import bindings
from spread_core.bam.dali.providers.providers import DaliProvider, DaliFunit, Groups, Validity, Discovery
from spread_core.bam.dali.tuner import TunerLimited
from spread_core.bam.multi_command import MultiCommand

""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                                  DALI 1                                """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

DISCOVERY_DELAY = 0.5


class DaliLight(DaliProvider):
    _dali_type = 1

    def __init__(self, data):
        super(DaliLight, self).__init__(data)
        self._fadeTime = 2
        self._fadeRate = 0
        self._last_level = 0

        self._level_timer = None
        self.is_unaddress_tunned = False
        self._tuner = None
        self._level_timer = None

        self.__level_before_discovery = None

    @property
    def fadeTime(self): return self._fadeTime

    @property
    def fadeRate(self): return self._fadeRate

    @fadeTime.setter
    def fadeTime(self, value):
        self._fadeTime = value

        if self.is_device:
            for _group in self.groups():
                _group.fadeTime = max((provider.fadeTime for provider in _group.get_members()))

            broadcast = self.manager.broadcast(self._dali_type)
            if broadcast and isinstance(broadcast, DaliLight) and broadcast.is_broadcast:
                broadcast.fadeTime = max((provider.fadeTime for provider in broadcast.get_members()))

    @fadeRate.setter
    def fadeRate(self, value):
        self._fadeRate = value

        if self.is_device:
            for _group in self.groups():
                _group.fadeRate = max((provider.fadeRate for provider in _group.get_members()))

            broadcast = self.manager.broadcast(self._dali_type)
            if broadcast and isinstance(broadcast, DaliLight) and broadcast.is_broadcast:
                broadcast.fadeRate = max((provider.fadeRate for provider in broadcast.get_members()))

    def repair_on_fail(self) -> tuple:
        return (*super(DaliLight, self).repair_on_fail(), CurrentLevelRaw, BrightnessLevel, Tuning,)

    @property
    def tuner(self) -> TunerLimited:
        return self._tuner

    def on_survey(self):
        yield from super(DaliLight, self).on_survey()
        yield dali.gear.general.QueryStatus(self.short_addr.address)

        if self.is_valid_strong: yield dali.gear.general.QueryActualLevel(self.short_addr.address)

        if self.is_valid_strong: yield dali.gear.general.QueryFadeTimeFadeRate(self.short_addr.address)
        if self.is_valid_strong: yield dali.gear.general.QueryMinLevel(self.short_addr.address)
        if self.is_valid_strong: yield dali.gear.general.QueryMaxLevel(self.short_addr.address)
        if self.is_valid_strong: yield dali.gear.led.QueryDimmingCurve(self.short_addr.address)
        if self.is_valid_strong: yield dali.gear.general.QueryPowerOnLevel(self.short_addr.address)
        if self.is_valid_strong: yield dali.gear.general.QuerySystemFailureLevel(self.short_addr.address)
        if self.is_valid_strong: yield dali.gear.general.QueryPhysicalMinimum(self.short_addr.address)
        if self.is_valid_strong: yield dali.gear.general.QueryGroupsZeroToSeven(self.short_addr)
        if self.is_valid_strong: yield dali.gear.general.QueryGroupsEightToFifteen(self.short_addr)

        for i in range(0, 16):
            if self.is_valid_strong: yield dali.gear.general.QuerySceneLevel(self.short_addr, i)

    def request(self, cmd, *args, **kwargs):
        if isinstance(cmd, (TunerLimited.UP, TunerLimited.DOWN)):
            if self.tuner and not self.tuner.is_active:
                logging.debug(f'skip {cmd}')
                return

        super(DaliLight, self).request(cmd, *args, **kwargs)

        if cmd.is_send_notify:
            self.on_command_sended(cmd)

    def on_request_answer(self, funit_type, response, sig=None, *args):
        if isinstance(response, dali.gear.general.QueryFadeTimeAndRateResponse) and response:
            ft = None if response.fade_time is None else 'ft%s' % int(FadeTime.values[response.fade_time]*10)
            self.on_update(FadeTime, ft, sig=sig)

            fr = None if response.fade_rate is None else 'fr%s' % int(FadeRate.values[response.fade_rate]*10)
            self.on_update(FadeRate, fr, sig=sig)
        elif funit_type == CurrentLevelRaw and response is None:
            return
        elif funit_type == CurrentLevelRaw and response.value.as_integer == 255:
            self.on_update(funit_type, None, sig=sig)
        elif funit_type == Types.part_funit and response:
            self.on_update(*Types.update_part(self, response.value.as_integer), sig=sig)
        elif isinstance(response, dali.gear.general.QueryStatusResponse):
            self.on_update(Validity, not response.error, sig=sig)
        elif funit_type == SceneLevelsRaw:
            level = response.value.as_integer if response else 255
            self.on_update(SceneLevelsRaw, self._parse_scenes(level, args[0]), sig=sig)
        else:
            super(DaliLight, self).on_request_answer(funit_type, response, sig, *args)

    def on_exit(self):
        if self._tuner:
            self._tuner.stop()
        self.stop_level_query()
        super(DaliLight, self).on_exit()

    def check_valid(self):
        super(DaliLight, self).check_valid()
        if self.is_device:
            self.request(dali.gear.general.QueryStatus(self.short_addr.address))

    def get_info(self):
        super(DaliLight, self).get_info()

        if self.is_device:
            if FadeTime not in self.values or FadeRate not in self.values:
                self.request(dali.gear.general.QueryFadeTimeFadeRate(self.short_addr.address))
            if MinLevelRaw not in self.values:
                self.request(dali.gear.general.QueryMinLevel(self.short_addr.address))
            if MaxLevelRaw not in self.values:
                self.request(dali.gear.general.QueryMaxLevel(self.short_addr.address))
            if DimmingCurve not in self.values:
                self.request(dali.gear.led.QueryDimmingCurve(self.short_addr.address))
            if SceneLevelsRaw not in self.values:
                self.GetSceneLevelsRaw()

            self.request(dali.gear.general.QueryActualLevel(self.short_addr.address))

            if Types not in self.values:
                self.GetTypes()
            if PowerOnLevelRaw not in self.values:
                self.request(dali.gear.general.QueryPowerOnLevel(self.short_addr.address))
            if SystemFailureLevelRaw not in self.values:
                self.request(dali.gear.general.QuerySystemFailureLevel(self.short_addr.address))
            if PhysicalMinLevelRaw not in self.values:
                self.request(dali.gear.general.QueryPhysicalMinimum(self.short_addr.address))
        else:
            # fadeTime меняется при изменении fadeTime e объектов
            self.on_update(FadeTime, self.get_value(FadeTime, FadeTime.index_of[1]))
            self.on_update(FadeRate, self.get_value(FadeRate, FadeRate.index_of[1]))
            self.on_update(MinLevelRaw, self.get_value(MinLevelRaw, 1))
            self.on_update(MaxLevelRaw, self.get_value(MaxLevelRaw, 254))
            self.on_update(DimmingCurve, self.get_value(DimmingCurve, DimmingCurve.index_of[0]))
            self.on_update(SceneLevelsRaw, self.get_value(SceneLevelsRaw, []))
            self.on_update(CurrentLevelRaw, self.get_value(CurrentLevelRaw, 0))
            self.on_update(PowerOnLevelRaw, self.get_value(PowerOnLevelRaw, 0))
            self.on_update(SystemFailureLevelRaw, self.get_value(SystemFailureLevelRaw, 80))
            self.on_update(PhysicalMinLevelRaw, self.get_value(PhysicalMinLevelRaw, 1))

    def query_level_async(self, sig=None):
        if self.get_value(Discovery):
            return

        for _provider in self.get_members():
            _provider.stop_level_query()
            if self._manager.level_sync_delay:
                _provider._level_timer = Timer.set(max(_provider.fadeTime, self._manager.level_sync_delay), _provider.query_level, sig)

    def stop_level_query(self):
        if self._level_timer:
            try:
                self._level_timer.close()
            except ValueError:
                pass
            finally:
                self._level_timer = None

    def query_level(self, sig=None):
        self._level_timer = None
        if self.is_device:
            cmd = dali.gear.general.QueryActualLevel(self.short_addr)
            cmd.sig = sig
            self.manager.add_command(self, cmd)
        else:
            for _provider in self.get_members():
                _provider.query_level()

    def on_update(self, funit_type, value, sig=None, **kwargs):
        super(DaliLight, self).on_update(funit_type, value, sig, **kwargs)

        if funit_type in [CurrentLevelRaw, DimmingCurve]:
            self.GetBrightnessLevel(sig, kwargs.get(Variable.PRELIM, False))
        elif funit_type == FadeTime:
            try:
                self.fadeTime = FadeTime.values[FadeTime.index_of.index(value)]
            except:
                pass
        elif funit_type == FadeRate:
            try:
                self.fadeRate = FadeRate.values[FadeRate.index_of.index(value)]
            except:
                pass

    def on_command_sended(self, cmd):
        if cmd.is_levels_dependent:
            for _provider in self.get_members(True):
                value = None
                cl = _provider.get_value(CurrentLevelRaw, 0)

                if isinstance(cmd, dali.gear.general.DAPC):
                    value = cmd.power
                elif isinstance(cmd, dali.gear.general.Off):
                    value = 0
                elif isinstance(cmd, dali.gear.general.RecallMinLevel):
                    value = 1
                elif isinstance(cmd, dali.gear.general.RecallMaxLevel):
                    value = 254
                elif isinstance(cmd, dali.gear.general.GoToLastActiveLevel):
                    value = _provider._last_level
                elif isinstance(cmd, dali.gear.general.Up):
                    if cl > 0:
                        value = _provider.get_value(CurrentLevelRaw, 0) + max(1, round(self.fadeRate/5))
                elif isinstance(cmd, dali.gear.general.Down):
                    if cl > 0:
                        value = max(1, _provider.get_value(CurrentLevelRaw, 0) - max(1, round(self.fadeRate/5)))
                elif isinstance(cmd, dali.gear.general.StepUp):
                    if cl > 0:
                        value = _provider.get_value(CurrentLevelRaw, 0) + 1
                elif isinstance(cmd, dali.gear.general.OnAndStepUp):
                    value = _provider.get_value(CurrentLevelRaw, 0) + 1
                elif isinstance(cmd, dali.gear.general.StepDown):
                    if cl > 0:
                        value = _provider.get_value(CurrentLevelRaw, 0) - 1
                elif isinstance(cmd, dali.gear.general.StepDownAndOff):
                    value = _provider.get_value(CurrentLevelRaw, 0) - 1
                elif isinstance(cmd, dali.gear.general.GoToScene):
                    scenes = _provider.get_value(SceneLevelsRaw, None)
                    scene_id = cmd.frame.pack[1] - 0x10
                    if scenes is not None and scene_id in scenes and scenes[scene_id] in range(0, 254):
                        value = scenes[scene_id]

                if value is not None:
                    if _provider.is_valid:
                        if value > 0:
                            value = max(_provider.get_value(MinLevelRaw, 0), value)
                            value = min(_provider.get_value(MaxLevelRaw, 254), value)
                            _provider._last_level = value
                        _provider.on_update(CurrentLevelRaw, value, **{Variable.PRELIM: True})

            if self._tuner:
                self._tuner.check_state(cmd)

            if isinstance(cmd, (Discovery.DOWN_CMD, Discovery.UP_CMD)) and self.get_value(Discovery):
                Timer.set(DISCOVERY_DELAY, self.switch_level_by_discovery, has_cmd=cmd.__class__)

            self.query_level_async(cmd.sig)

        elif isinstance(cmd, dali.gear.general.SetFadeTime):
            if self.dtr(0) is not None:
                ft = 'ft%s' % int(FadeTime.values[self.dtr(0)] * 10)
                self.update_recursive(FadeTime, ft, **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.gear.general.SetFadeRate):
            if self.dtr(0) is not None:
                fr = 'fr%s' % int(FadeRate.values[self.dtr(0)] * 10)
                self.update_recursive(FadeRate, fr, **{Variable.PRELIM: True})
        elif isinstance(cmd, (dali.gear.general.AddToGroup, dali.gear.general.RemoveFromGroup)):
            included = isinstance(cmd, dali.gear.general.AddToGroup)
            for part in self.get_members(return_self=True):
                if isinstance(part, DaliLight):
                    value = Groups.def_value(part._dali_type)
                    value = [{**i} for i in part.get_value(Groups, value)]
                    cpl = False
                    for i in value:
                        if i['index'] == cmd.param:
                            i['value'] = included
                            cpl = True
                            break
                    if cpl is False:
                        value.append({'index': cmd.param, 'value': included})

                    part.on_update(Groups, value, **{Variable.PRELIM: True})
        elif isinstance(cmd, (dali.gear.general.SetScene, dali.gear.general.RemoveFromScene)):
            level = self.dtr(0) if isinstance(cmd, dali.gear.general.SetScene) else 255
            for part in self.get_members(return_self=True):
                if isinstance(part, DaliLight):
                    value = SceneLevelsRaw.def_value()
                    value = [{**i} for i in part.get_value(SceneLevelsRaw, value)]

                    cpl = False

                    for i in value:
                        if i['index'] == cmd.param:
                            i['value'] = level
                            cpl = True
                            break

                    if cpl is False:
                        value.append({'index': cmd.param, 'value': level})

                    part.on_update(SceneLevelsRaw, value, **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.gear.led.SelectDimmingCurve):
            if self.dtr(0) is not None:
                self.update_recursive(DimmingCurve, DimmingCurve.of_raw(self.dtr(0)), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.gear.general.SetSystemFailureLevel):
            if self.dtr(0) is not None:
                self.update_recursive(SystemFailureLevelRaw, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.gear.general.SetPowerOnLevel):
            if self.dtr(0) is not None:
                self.update_recursive(PowerOnLevelRaw, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.gear.general.SetMaxLevel):
            if self.dtr(0) is not None:
                self.update_recursive(MaxLevelRaw, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.gear.general.SetMinLevel):
            if self.dtr(0) is not None:
                self.update_recursive(MinLevelRaw, self.dtr(0), **{Variable.PRELIM: True})
        else:
            super(DaliLight, self).on_command_sended(cmd)

    def prs2raw(self, p):
        if p == 0:
            return p

        p = abs(p)
        _min = self.get_value(MinLevelRaw, 1)
        _max = self.get_value(MaxLevelRaw, 254)

        if self.get_value(DimmingCurve, None) == DimmingCurve.LINEAR:
            return min(max(_min, round(254 * p / 100)), _max)
        else:
            return min(max(_min, round(((math.log10(p) + 1) * 253 / 3 + 1))), _max)

    def raw2prs(self, r):
        if self.get_value(DimmingCurve, None) == DimmingCurve.LINEAR:
            return min(100, round(100 * r / 254))
        else:
            if r == 0:
                return 0
            elif r < 1:
                return 1
            else:
                p = 10 ** ((r - 1) * 3 / 253 - 1)
                if 0 < p < 1:
                    p = 1
                return round(p)

    def read_memory_location(self, bank_id, start_address, end_address) -> bytes:
        self.check_command_for_device()
        i = self.manager.send_interface

        i.send(dali.gear.general.DTR1(bank_id))

        for _dtr0 in range(start_address, end_address):
            i.send(dali.gear.general.DTR0(_dtr0))
            res = i.send(dali.gear.general.ReadMemoryLocation(self.short_addr.address))
            if isinstance(res, dali.command.Response):
                yield res.value.as_integer
            else:
                break

    def SetShortAddress(self, new_addr, sig=None):
        if self.is_device:
            if new_addr != self.short_addr.address:
                values = None
                old_short = self.short_addr
                i = self.manager.send_interface
                if new_addr is None:
                    i.send(dali.gear.general.DTR0(0xff))
                    i.send(dali.gear.general.SetShortAddress(old_short))
                    self.reset()
                    return
                else:
                    new_short = dali.address.Short(new_addr)
                    res = i.send(dali.gear.general.QueryDeviceType(new_short))
                    if res:
                        i.send(dali.gear.general.DTR0((new_short.address << 1) | 1))
                        i.send(dali.gear.general.SetShortAddress(old_short))
                    else:
                        for _provider in self.manager.providers_by_address(self._dali_type, new_addr):
                            if isinstance(_provider, self.__class__):
                                old_long = []
                                new_long = []

                                i.send(dali.gear.general.Terminate())

                                i.send(dali.gear.general.Initialise(broadcast=False, address=old_short.address))
                                i.send(dali.gear.general.Initialise(broadcast=False, address=new_short.address))

                                old_long.append(i.send(dali.gear.general.QueryRandomAddressH(old_short)))
                                old_long.append(i.send(dali.gear.general.QueryRandomAddressM(old_short)))
                                old_long.append(i.send(dali.gear.general.QueryRandomAddressL(old_short)))

                                new_long.append(i.send(dali.gear.general.QueryRandomAddressH(new_short)))
                                new_long.append(i.send(dali.gear.general.QueryRandomAddressM(new_short)))
                                new_long.append(i.send(dali.gear.general.QueryRandomAddressL(new_short)))

                                i.send(dali.gear.general.SetSearchAddrH(old_long[0].value.as_integer))
                                i.send(dali.gear.general.SetSearchAddrM(old_long[1].value.as_integer))
                                i.send(dali.gear.general.SetSearchAddrL(old_long[2].value.as_integer))
                                i.send(dali.gear.general.ProgramShortAddress(new_short.address))

                                i.send(dali.gear.general.SetSearchAddrH(new_long[0].value.as_integer))
                                i.send(dali.gear.general.SetSearchAddrM(new_long[1].value.as_integer))
                                i.send(dali.gear.general.SetSearchAddrL(new_long[2].value.as_integer))
                                i.send(dali.gear.general.ProgramShortAddress(old_short.address))

                                i.send(dali.gear.general.Terminate())

                                values = _provider.values.copy()

                                for funit_type, value in self.values.items():
                                    _provider.on_update(funit_type, value)

                                break

                self.SetBindingDevice(new_addr, sig=sig, values=values)
            else:
                raise dali.exceptions.CommandGenerateError('You can not give an address to {}'.format(self.binding.b_type))

    def SetDiscovery(self, value, sig=None):
        _discovery = self.get_value(Discovery, False)
        if _discovery != value:
            self.on_update(Discovery, value, sig=sig)
            if value:
                self.__level_before_discovery = self.get_value(BrightnessLevel, 0)
                self.switch_level_by_discovery()
            else:
                has_level = self.__level_before_discovery
                if has_level is not None:
                    self.SetBrightnessLevel(has_level)

    def switch_level_by_discovery(self, has_cmd=None):
        if not self.get_value(Discovery):
            return

        if has_cmd == Discovery.UP_CMD:
            cmd = Discovery.DOWN_CMD
        else:
            cmd = Discovery.UP_CMD

        self.on_command_prepared(cmd(self.short_addr))

    def _parse_scenes(self, value, index) -> list:
        scenes = [
            {key: value for key, value in item.items()}
            for item in self.get_value(SceneLevelsRaw, SceneLevelsRaw.def_value())
        ]
        if len(scenes) != SceneLevelsRaw.length:
            scenes = SceneLevelsRaw.def_value()

        scenes[index]['value'] = value

        return scenes

    def GetSceneLevelsRaw(self, sig=None):
        if self.is_device:
            for i in range(0, 16):
                self.request(dali.gear.general.QuerySceneLevel(self.short_addr, i))

        self.on_update(SceneLevelsRaw, self.get_value(SceneLevelsRaw, SceneLevelsRaw.def_value()), sig=sig)

    def SetSceneLevelsRaw(self, val, sig=None):
        for item in val:
            self.request(dali.gear.general.DTR0(max(0, min(255, item['value']))))
            self.request(dali.gear.general.SetScene(self.short_addr, item['index']))

        if self.is_device:
            self.GetSceneLevelsRaw(sig=sig)
        else:
            self.update_recursive(SceneLevelsRaw, val, sig=sig)

    def GetGroups(self, sig=None):
        if self.is_device:
            self.request(dali.gear.general.QueryGroupsZeroToSeven(self.short_addr))
            self.request(dali.gear.general.QueryGroupsEightToFifteen(self.short_addr))

        _groups = [{'index': self.binding.addr.group, 'value': True}] if self.is_group else []
        _groups = self.get_value(Groups, _groups)

        self.on_update(Groups, _groups, sig=sig)

    def SetGroups(self, val, sig=None):
        for item in val:
            if item['value']:
                self.request(dali.gear.general.AddToGroup(self.short_addr, item['index']))
            else:
                self.request(dali.gear.general.RemoveFromGroup(self.short_addr, item['index']))

        if self.is_device:
            self.GetGroups(sig)
        else:
            self.update_recursive(Groups, val, sig=sig)

    def GetTypes(self, sig=None):
        _types = None

        if self.is_device:
            i = self._manager.send_interface

            _type = i.send(dali.gear.general.QueryDeviceType(self.short_addr))
            if _type:
                if _type.value.as_integer == 0xff:
                    _types = []
                    while _type is not None:
                        _type = i.send(dali.gear.general.QueryNextDeviceType(self.short_addr))
                        if _type:
                            if _type.value.as_integer == 0xfe:
                                break
                            else:
                                _types.append(_type.value.as_integer)

                    if len(_types) == 0:
                        _types.append(0xff)
                else:
                    _types = [_type.value.as_integer]

        self.on_update(Types, _types, sig=sig)

    def SetGroupLevel(self, value, sig=None):
        try:
            value = int(value)
        except:
            raise dali.exceptions.CommandGenerateError('Value({}) must be integer'.format(value))
        else:
            group_ids = self.get_value(Groups, None)
            group_id = None
            group_size = 999999999
            if group_ids is None:
                self.GetGroups()
                group_ids = self.get_value(Groups, [])

            for item in group_ids:
                if item['value'] is True and 'index' in item:
                    _group_id = item['index']
                    _group_size = self.manager.get_group_size(self._dali_type, _group_id)
                    if _group_size < group_size:
                        group_size = _group_size
                        group_id = _group_id

            if group_id is None:
                self.SetBrightnessLevel(value, sig=sig)
            else:
                group = self.manager.group_providers[self._dali_type][group_id]
                if group:
                    group.SetBrightnessLevel(value, sig=sig)
                else:
                    self.set_pause_for_group(group_id)
                    cmd = dali.gear.general.DAPC(bindings.Group(group_id), self.prs2raw(value))
                    self.request(cmd)
                    for _participant in self.manager.get_group_members(self.dali_type, group_id):
                        _participant.on_command_sended(cmd)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        if self.is_valid is False:
            raise dali.exceptions.CommandGenerateError('Device is invalid!')

        method = kwargs.pop(mqtt.const.METHOD, None)
        if method == dali.gear.general.RecallMinLevel.__name__:
            cmd = dali.gear.general.RecallMinLevel(self.short_addr)
        elif method == dali.gear.general.RecallMaxLevel.__name__:
            cmd = dali.gear.general.RecallMaxLevel(self.short_addr)
        elif method == 'Scene':
            cmd = dali.gear.general.GoToScene(self.short_addr, value)
        elif method == dali.gear.general.Off.__name__:
            cmd = dali.gear.general.Off(self.short_addr)
        elif method == dali.gear.general.GoToLastActiveLevel.__name__:
            cmd = dali.gear.general.GoToLastActiveLevel(self.short_addr)
        elif method == dali.gear.general.StepUp.__name__:
            cmd = dali.gear.general.StepUp(self.short_addr)
        elif method == dali.gear.general.StepDown.__name__:
            cmd = dali.gear.general.StepDown(self.short_addr)
        elif method == dali.gear.general.OnAndStepUp.__name__:
            cmd = dali.gear.general.OnAndStepUp(self.short_addr)
        elif method == dali.gear.general.StepDownAndOff.__name__:
            cmd = dali.gear.general.StepDownAndOff(self.short_addr)
        elif method == dali.gear.general.Up.__name__:
            cmd = dali.gear.general.Up(self.short_addr)
        elif method == dali.gear.general.Down.__name__:
            cmd = dali.gear.general.Down(self.short_addr)
        else:
            cmd = dali.gear.general.DAPC(self.short_addr, self.prs2raw(value))
        cmd.sig = sig
        if self.is_device:
            if self.is_unaddress_tunned:
                groups = self.get_value(Groups, [])
                for item in groups:
                    if item['value'] is True:
                        group_id = item['index']
                        self.set_pause_for_group(group_id)
                self.set_pause_for_group(-1)
            elif self.tuner:
                self.tuner.set_pause(2+self.fadeTime)
        else:
            if self.tuner:
                self.tuner.set_pause(2+self.fadeTime)

        self.manager.add_command(self, cmd, **kwargs)

    def GetBrightnessLevel(self, sig=None, preliminary=False):
        _level = self.get_value(CurrentLevelRaw, None)
        if _level is not None:
            _level = self.raw2prs(_level)
        self.on_update(BrightnessLevel, _level, sig=sig, preliminary=preliminary)

    def SetTuning(self, value, speed=None, sig=None):
        _c_val = self.get_value(Tuning, 0)
        speed = 0.1 if speed is None else speed

        if _c_val * value <= 0:
            self.on_update(Tuning, value, sig=sig)
            if value == 0:
                if self._tuner:
                    self._tuner.stop()
                    self._tuner = None
            else:
                if self._tuner is None:
                    self._tuner = TunerLimited(self)

                self._tuner.set_direction(TunerLimited.UP if value > 0 else TunerLimited.DOWN, speed)
        elif self._tuner:
            self._tuner.speed = speed

    def set_pause_for_group(self, group_id):
        for p_id, _provider in self.manager.providers.items():
            if isinstance(_provider, DaliDimmer) and _provider.tuner:
                if ((_provider.is_group and _provider.short_addr.group == group_id)
                        or (_provider.is_broadcast and group_id == -1)):
                    _provider.tuner.set_pause(2+self.fadeTime)

    def serialize(self) -> dict:
        d = super(DaliLight, self).serialize()
        if d:
            for funit_type in [Types, PhysicalMinLevelRaw, MinLevelRaw, MaxLevelRaw, PowerOnLevelRaw, Groups,
                               SystemFailureLevelRaw, FadeTime, FadeRate, DimmingCurve, SceneLevelsRaw]:
                value = self.get_value(funit_type)
                if value is not None:
                    key = str(funit_type)[0].lower() + str(funit_type)[1:]
                    d[key] = value
        return d


""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                                 DIMMER                                 """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


class DaliDimmer(DaliLight):
    def __init__(self, data):
        super(DaliDimmer, self).__init__(data)
        self.step_cmd = None


""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                              COLOR CONTROL                             """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


class DaliTunableWhite(DaliDimmer):
    def __init__(self, *args, **kwargs):
        super(DaliTunableWhite, self).__init__(*args, **kwargs)
        self.temperature_timer = None

    def repair_on_fail(self) -> tuple:
        return (*super(DaliLight, self).repair_on_fail(), Temperature,)

    def get_info(self):
        super(DaliTunableWhite, self).get_info()
        if self.is_device:
            self.GetTemperature()
            if Coolest not in self.values:
                self.GetCoolest()
            if Warmest not in self.values:
                self.GetWarmest()
        else:
            self.on_update(Temperature, self.get_value(Temperature, 5500))
            self.on_update(Coolest, self.get_value(Coolest, 2000))
            self.on_update(Warmest, self.get_value(Warmest, 9000))

    def on_survey(self):
        yield from super(DaliTunableWhite, self).on_survey()

        if self.is_valid_strong: yield MultiCommand(self.GetTemperature)
        if self.is_valid_strong: yield MultiCommand(self.GetCoolest)
        if self.is_valid_strong: yield MultiCommand(self.GetWarmest)

    def on_command_sended(self, cmd):
        if isinstance(cmd, dali.gear.colour_control.SetTemperature):
            if self.dtr(0) is not None and self.dtr(1) is not None:
                mirek = int.from_bytes(bytes([self.dtr(0), self.dtr(1)]), 'little')
                temperature = int(1000000 / mirek)
                self.update_recursive(Temperature, temperature, **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.gear.colour_control.StoreColourTemperatureLimit):
            mirek = int.from_bytes(bytes([self.dtr(0), self.dtr(1)]), 'little')
            value_type = self.dtr(2)
            value = int(1000000 / mirek)
            funit = Coolest if value_type == dali.gear.colour_control.StoreColourTemperatureLimit.COOLEST else Warmest
            self.update_recursive(funit, value, **{Variable.PRELIM: True})
        else:
            super(DaliTunableWhite, self).on_command_sended(cmd)

    def _get_color_value(self, funit_type, value_type: int, sig=None):
        res = None
        try:
            self.request(dali.gear.general.DTR0(value_type))
            res = self.manager.send_interface.send(dali.gear.colour_control.QueryColorValue(self.short_addr)).value.as_integer
            if res != 0xff:
                b0 = self.manager.send_interface.send(dali.gear.general.QueryContentDTR0(self.short_addr)).value.as_integer
                b1 = self.manager.send_interface.send(dali.gear.general.QueryContentDTR1(self.short_addr)).value.as_integer
                res = int(1000000 / int.from_bytes([b0, b1], 'little'))
        except:
            pass
        finally:
            self.on_update(funit_type, res, sig=sig)

    def _set_color_limit(self, value_type: int, value: int):
        if not 1000 <= value <= 20000:
            raise ValueError("temperature must be in the range 1000..20000")
        mirek = int(1000000 / value)
        self._params = [int(b) for b in mirek.to_bytes(2, 'little')] + [value_type]
        self.request(dali.gear.colour_control.StoreColourTemperatureLimit(self.short_addr))

    def SetTemperature(self, temperature=1, sig=None):
        if not 1000 <= temperature <= 20000:
            raise ValueError("temperature must be in the range 1000..20000")
        try:
            mirek = int(1000000 / temperature)
            bts = [int(b) for b in mirek.to_bytes(2, 'little')]
            self.request(dali.gear.general.DTR0(bts[0]))
            self.request(dali.gear.general.DTR1(bts[1]))
            self.request(dali.gear.colour_control.SetTemperature(self.short_addr))
            self.request(dali.gear.colour_control.Activate(self.short_addr))
        except BaseException as ex:
            logging.error(ex)
        else:
            self.query_temperature()

    def query_temperature(self):
        self.stop_temperature_timer()

        mc = MultiCommand(self.GetTemperature)
        self.temperature_timer = Timer.set(self.fadeTime, self.manager.add_command, self, mc)

    def stop_temperature_timer(self):
        if self.temperature_timer:
            try:
                self.temperature_timer.close()
            except ValueError:
                pass
            finally:
                self.temperature_timer = None

    def GetTemperature(self, sig=None):
        self._get_color_value(Temperature, dali.gear.colour_control.QueryColorValue.TEMPERATURE, sig=sig)

    def SetCoolest(self, sig=None, value=0):
        try:
            self._set_color_limit(dali.gear.colour_control.StoreColourTemperatureLimit.COOLEST, value)
        except BaseException as ex:
            logging.error(ex)
        else:
            self.GetCoolest(sig)

    def GetCoolest(self, sig=None):
        self._get_color_value(Coolest, dali.gear.colour_control.QueryColorValue.COOLEST, sig=sig)

    def SetWarmest(self, sig=None, value=0):
        try:
            self._set_color_limit(dali.gear.colour_control.StoreColourTemperatureLimit.WARMEST, value)
        except BaseException as ex:
            logging.error(ex)
        else:
            self.GetWarmest(sig)

    def GetWarmest(self, sig=None):
        self._get_color_value(Warmest, dali.gear.colour_control.QueryColorValue.WARMEST, sig=sig)

    def on_exit(self):
        self.stop_temperature_timer()
        super(DaliTunableWhite, self).on_exit()


""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                                 SWITCH                                 """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


class DaliRelay(DaliLight):
    pass


class DaliLightFunit(DaliFunit):
    entity_type = DaliLight
    _dtr0 = dali.gear.general.DTR0
    _dtr1 = dali.gear.general.DTR1
    _dtr2 = dali.gear.general.DTR2

    @property
    def entity(self) -> DaliLight: return super(DaliLightFunit, self).entity

    def _prepare_cmd(self, provider: DaliLight, cmd, *args):
        if cmd._hasparam:
            return cmd(provider.short_addr, *args)
        else:
            return cmd(provider.short_addr)


class BrightnessLevel(DaliLightFunit):
    set_func = DaliLight.SetBrightnessLevel
    get_func = DaliLight.GetBrightnessLevel


class Types(DaliLightFunit):
    part_funit = 'NextType'
    get_func = DaliLight.GetTypes

    @classmethod
    def update_part(cls, entity: DaliLight, value: int) -> (DaliFunit, str):
        v = entity.get_value(cls, [])
        if value != 254 and value not in v:
            v.append(value)

        if value == 254 and 255 in v and len(v) > 1:
            v.remove(255)

        return cls, sorted(v)


class CurrentLevelRaw(DaliLightFunit):
    get_after_set = False
    set_cmd = dali.gear.general.DAPC
    get_cmd = dali.gear.general.QueryActualLevel


class PhysicalMinLevelRaw(DaliLightFunit):
    get_cmd = dali.gear.general.QueryPhysicalMinimum


class MinLevelRaw(DaliLightFunit):
    set_cmd = dali.gear.general.SetMinLevel
    get_cmd = dali.gear.general.QueryMinLevel

    def set(self):
        super(MinLevelRaw, self).set()
        self.entity.on_command_prepared(
            dali.gear.general.QueryActualLevel(
                self.entity.short_addr
            )
        )


class MaxLevelRaw(DaliLightFunit):
    set_cmd = dali.gear.general.SetMaxLevel
    get_cmd = dali.gear.general.QueryMaxLevel

    def set(self):
        super(MaxLevelRaw, self).set()
        self.entity.on_command_prepared(
            dali.gear.general.QueryActualLevel(
                self.entity.short_addr
            )
        )


class PowerOnLevelRaw(DaliLightFunit):
    set_cmd = dali.gear.general.SetPowerOnLevel
    get_cmd = dali.gear.general.QueryPowerOnLevel


class SystemFailureLevelRaw(DaliLightFunit):
    set_cmd = dali.gear.general.SetSystemFailureLevel
    get_cmd = dali.gear.general.QuerySystemFailureLevel


class SceneLevelsRaw(DaliLightFunit):
    length = 16
    set_func = DaliLight.SetSceneLevelsRaw
    get_func = DaliLight.GetSceneLevelsRaw

    @staticmethod
    def def_value() -> list:
        return [{'index': i, 'value': 255} for i in range(0, 16)]


class FadeTime(DaliLightFunit):
    values = [0, 0.7, 1, 1.4, 2, 2.8, 4, 5.7, 8, 11.3, 16, 22.6, 32, 45.3, 64, 90.5]
    index_of = ["ft0", "ft7", "ft10", "ft14", "ft20", "ft28", "ft40", "ft57", "ft80", "ft113", "ft160", "ft226",
                "ft320", "ft453", "ft640", "ft905"]
    set_cmd = dali.gear.general.SetFadeTime
    get_cmd = dali.gear.general.QueryFadeTimeFadeRate


class FadeRate(DaliLightFunit):
    values = [0, 358, 253, 179, 127, 89.4, 63.3, 44.7, 31.6, 22.4, 15.8, 11.2, 7.9, 5.6, 4, 2.8]
    index_of = ["", "fr3580", "fr2530", "fr1790", "fr1270", "fr894", "fr633", "fr447", "fr316", "fr224", "fr158",
                "fr112", "fr79", "fr56", "fr40", "fr28"]
    set_cmd = dali.gear.general.SetFadeRate
    get_cmd = dali.gear.general.QueryFadeTimeFadeRate


class DimmingCurve(DaliLightFunit):
    LOGARITHMIC = 'Logarithmic'
    LINEAR = 'Linear'
    index_of = [LOGARITHMIC, LINEAR]
    set_cmd = dali.gear.led.SelectDimmingCurve
    get_cmd = dali.gear.led.QueryDimmingCurve


class GroupLevel(DaliLightFunit):
    set_func = DaliLight.SetGroupLevel


class Tuning(DaliLightFunit):
    entity_type = DaliLight
    set_func = entity_type.SetTuning


class Temperature(DaliFunit):
    entity_type = DaliTunableWhite
    set_func = entity_type.SetTemperature
    get_func = entity_type.GetTemperature


class Warmest(DaliFunit):
    entity_type = DaliTunableWhite
    set_func = entity_type.SetWarmest
    get_func = entity_type.GetWarmest


class Coolest(DaliFunit):
    entity_type = DaliTunableWhite
    set_func = entity_type.SetCoolest
    get_func = entity_type.GetCoolest
