import logging

import spread_core.protocols.dali as dali
from spread_core.bam.dali import bindings
from spread_core.bam.entities import Funit
from spread_core.bam.multi_command import MultiCommand
from spread_core.bam.providers import Provider
from spread_core.errors.project_errors import JocketError
from spread_core.tools import settings


class DaliProvider(Provider):
    _dali_type = 0
    dali_type = property(fget=lambda self: self._dali_type)

    def __init__(self, data):
        super(DaliProvider, self).__init__(data)

        f_unit = bindings.BINDING
        _exist, _ower = settings.dump.get_dump(data['type'], data['id'], f_unit)
        data['attributes']['def_' + f_unit.lower()] = data['attributes'][f_unit.lower()]
        if _exist:
            data['attributes'][f_unit] = _ower

        f_unit = bindings.BINDING + bindings.DEVICE
        data['attributes']['def_' + bindings.DEVICE.lower()] = data['attributes'][bindings.DEVICE.lower()]
        _exist, _ower = settings.dump.get_dump(data['type'], data['id'], f_unit)
        if _exist:
            data['attributes'][bindings.DEVICE.lower()] = _ower

        f_unit = bindings.BINDING + bindings.GROUP
        data['attributes']['def_' + bindings.GROUP.lower()] = data['attributes'][bindings.GROUP.lower()]
        _exist, _ower = settings.dump.get_dump(data['type'], data['id'], f_unit)
        if _exist:
            data['attributes'][bindings.GROUP.lower()] = _ower

        self._binding = bindings.of(data['attributes'])

    def __str__(self):
        return super(DaliProvider, self).__str__() + f' {self.binding}'

    def set_value(self, key, value):
        super(DaliProvider, self).set_value(key, value)
        if key in (BindingDevice, BindingGroup, Binding):
            if (key == BindingDevice and value == self.binding.def_device) or \
                    (key == BindingGroup and value == self.binding.def_group) or \
                    (key == Binding and value == self.binding.def_binding):
                settings.dump.set_dump(self, key, settings.KILL_ITEM)
            else:
                settings.dump.set_dump(self, key, value)

    def groups(self):
        """
        :return: Список групповых провайдеров, в группы которых включён provider
        """
        for g_id in self.get_value(Groups, []):
            _group = self.manager.group_providers[self.dali_type][g_id]
            if _group:
                yield _group

    def not_repaired(self) -> tuple:
        return (*super(DaliProvider, self).not_repaired(), Binding, BindingDevice, BindingGroup,)

    def repair_on_fail(self) -> tuple:
        return (*super(DaliProvider, self).repair_on_fail(), Discovery)

    @property
    def surveyable(self):
        if self.is_device:
            return super(DaliProvider, self).surveyable

        return False

    @property
    def binding(self):
        return self._binding

    @property
    def short_addr(self):
        return self.binding.addr

    @property
    def is_device(self):
        return isinstance(self.binding, bindings.BindingDevice)

    @property
    def is_group(self):
        return isinstance(self.binding, bindings.BindingGroup)

    @property
    def is_broadcast(self):
        return isinstance(self.binding, bindings.BindingBroadcast)

    @property
    def is_unaddressed(self):
        return isinstance(self.binding, bindings.BroadcastUnaddressed)

    def get_members(self, return_self=False):
        if self.is_group:
            yield from self.manager.get_group_members(self.dali_type, self.short_addr.group)
            if return_self:
                yield self
        elif self.is_broadcast:
            yield from self.manager.get_all_devices(self.dali_type)
            if return_self:
                yield self
        else:
            yield self

    def dtr(self, index):
        return self._manager.dtr_values[self._dali_type][index]

    def request(self, cmd: dali.command.Command, *args, **kwargs):
        try:
            sig = kwargs.pop('sig', cmd.sig)
            response = self._manager.send_interface.send(cmd, *args, **kwargs)
            if cmd.is_query:
                if isinstance(cmd, dali.gear.general._GearCommand) and cmd._hasparam:
                    cmd_args = (cmd.param, )
                else:
                    cmd_args = ()
                [self.on_request_answer(funit_type, response, sig, *cmd_args) for funit_type in cmd.funit_type]
        except BaseException as ex:
            logging.exception(ex)

    def on_request_answer(self, funit_type, response, sig=None, *args):
        if isinstance(response, dali.command.YesNoResponse):
            self.on_update(funit_type, response.value, sig=sig)
        elif funit_type == 'MemoryPart' and response:
            self.on_update(*_MemoryValue.update_part(self, response.value.as_integer, *args), sig=sig)
        elif isinstance(response, dali.command.Response):
            funit = DaliFunit.of(funit_type)
            if funit and issubclass(funit, DaliFunit):
                if funit.index_of:
                    value = funit.index_of[response.value.as_integer]
                elif funit.mask:
                    value = [funit.mask[i] for i in range(0, 8) if response.value.as_integer & 2**i != 0]
                else:
                    value = response.value.as_integer

                if funit.value_type:
                    if value.__class__ != funit.value_type:
                        if funit.value_type == list:
                            value = [value]
                        else:
                            value = funit.value_type(value)
            else:
                value = response.value.as_integer

            self.on_update(funit_type, value, sig=sig)
        else:
            Provider.on_update(self, funit_type, None, sig=sig)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        if funit_type == Validity:
            if self.is_valid and not value:
                self.set_valid(self.SUSPICIOUS)
            elif self._status != self.VALID and value:
                self.set_valid(self.VALID)
        elif funit_type == '%s0-7' % Groups:
            self._parse_groups(0, value, sig=sig)
        elif funit_type == '%s8-15' % Groups:
            self._parse_groups(8, value, sig=sig)
        elif funit_type == '%s16-23' % Groups:
            self._parse_groups(16, value, sig=sig)
        elif funit_type == '%s24-31' % Groups:
            self._parse_groups(24, value, sig=sig)
        else:
            if funit_type == Groups and value is not None and self.is_device:
                for item in value:
                    if item['value']:
                        self.manager.add_to_group(self._dali_type, item['index'], self.id)
                    else:
                        self.manager.remove_from_group(self._dali_type, item['index'], self.id)

            super().on_update(funit_type, value, sig, **kwargs)

    def get_bindings(self):
        self.GetBinding()
        self.GetBindingDevice()
        self.GetBindingGroup()

    def get_info(self):
        if Discovery not in self.values:
            self.on_update(Discovery, False)
        if Groups not in self.values:
            self.GetGroups()

    def on_survey(self):
        yield from ()

    def check_valid(self):
        super(DaliProvider, self).check_valid()
        self._status = self.SUSPICIOUS

    def update_recursive(self, funit_type, value, sig=None, **kwargs):
        self.on_update(funit_type, value, sig, **kwargs)

        if not self.is_device:
            for _provider in self.get_members():
                _provider.on_update(funit_type, value, sig=None, **kwargs)

    def set_valid(self, validity):
        if validity == self._status == self.SUSPICIOUS:
            validity = self.INVALID

        super(DaliProvider, self).set_valid(validity)

        if validity != self.SUSPICIOUS:
            for funit_type in self.values.copy():
                if funit_type not in (Binding, BindingDevice, BindingGroup):
                    self.on_update(funit_type, None if validity == self.INVALID else self.get_value(funit_type))

    def reset(self):
        self.publish_all(invalid=True)
        self.values = {}
        self.check_valid()
        self.get_bindings()
        if self.is_valid:
            self.get_info()

    def check_command_for_device(self):
        if not self.is_device:
            raise JocketError('The GET command cannot be applied to an object that is not an DEVICE!')

    def read_memory_location(self, bank_id, start_address, end_address) -> bytes:
        raise NotImplementedError()

    def SetShortAddress(self, new_addr, sig=None):
        raise NotImplementedError()

    def SetDiscovery(self, value, sig=None):
        raise NotImplementedError()

    def GetBinding(self, sig=None):
        self.on_update(Binding, self.binding.b_type, sig=sig)

    def SetBinding(self, value, sig=None):
        if value not in bindings.ALL:
            raise dali.exceptions.CommandGenerateError('Binding({}) is not present in available values({})'.format(value, bindings.ALL))

        _c_binding = self.get_value(bindings.BINDING, bindings.DEVICE)
        if _c_binding != value:
            self._binding = bindings.of({
                bindings.BINDING.lower(): value,
                value.lower(): self.get_value((str(Binding) + value), -1),
                'def_' + bindings.BINDING.lower(): self.binding.def_binding
            })

            self.on_update(Binding, value, sig=sig)

    def GetBindingDevice(self, sig=None):
        if self.get_value(bindings.BINDING) == bindings.DEVICE:
            self.on_update(BindingDevice, self.binding.addr.address, sig=sig)
        else:
            self.on_update(BindingDevice, None, sig=sig)

    def SetBindingDevice(self, address, sig=None, values=None):
        if address is None:
            address = self.binding.def_device

        try:
            address = int(address)
        except:
            raise dali.exceptions.CommandGenerateError('Value({}) must be integer!'.format(address))
        else:

            if self.get_value(BindingDevice) != address:
                if self.get_value(bindings.BINDING) == bindings.DEVICE:
                    if values is None:
                        for _provider in self.manager.providers_by_address(self._dali_type, address):
                            if isinstance(_provider, self.__class__):
                                values = _provider.values.copy()
                                break

                    try:
                        self.manager._addresses[self.short_addr.address].remove(self.id)
                        if len(self.manager._addresses[self.short_addr.address]) == 0:
                            self.manager._addresses.pop(self.short_addr.address)
                    finally:
                        self.binding.set_addr(address)
                        self.manager._addresses.setdefault(address, [])
                        self.manager._addresses[address].append(self.id)

                    if values is None:
                        self.reset()
                    else:
                        for funit_type, value in values.items():
                            self.on_update(funit_type, value)

            self.on_update(BindingDevice, address, sig=sig)

    def GetBindingGroup(self, sig=None):
        if self.get_value(bindings.BINDING) == bindings.GROUP:
            self.on_update(BindingGroup, self.binding.addr.group, sig=sig)
        else:
            self.on_update(BindingGroup, None, sig=sig)

    def SetBindingGroup(self, value, sig=None):
        try:
            value = int(value)
        except:
            raise dali.exceptions.CommandGenerateError('Value({}) must be integer!'.format(value))

        if self.get_value(BindingGroup) != value:
            if self.get_value(bindings.BINDING) == bindings.GROUP:
                self.binding.set_addr(value)
            self.on_update(BindingGroup, value, sig=sig)

    def _parse_groups(self, first_index: int, value, sig=None):
        _groups = [{key: value for key, value in item.items()} for item in self.get_value(Groups, [])]
        if len(_groups) != Groups.length[self._dali_type]:
            _groups = Groups.def_value(self._dali_type)

        for _i in range(0, 8):
            _groups[first_index + _i]['value'] = bool(value & (1 << _i))

        self.update_recursive(Groups, _groups, sig=sig)

    def GetGroups(self, sig=None):
        raise NotImplementedError()

    def SetGroups(self, val, sig=None):
        raise NotImplementedError()

    def serialize(self) -> dict:
        d = {}
        if self.is_device:
            d['address'] = self.binding.addr.address
            for funit_type in [Gtin, Serial, GtinOem, SerialOem, FirmwareVersion, HardwareVersion]:
                value = self.get_value(funit_type)
                key = str(funit_type)[0].lower() + str(funit_type)[1:]
                if value is not None:
                    d[key] = value
        return d


class DaliFunit(Funit):
    get_for_not_devices = True
    set_cmd = None
    get_cmd = None
    set_func = None
    get_func = None
    easy_get = False
    get_after_set = True
    _dtr0 = None
    _dtr1 = None
    _dtr2 = None

    def _prepare_cmd(self, provider, cmd, *args):
        raise NotImplemented()

    def _apply_cmd(self, provider: DaliProvider, cmd, *args):
        cmds = []

        if cmd._uses_dtr0:
            cmds.append(self._dtr0(args[0]))
            args = args[1:]

        if cmd._uses_dtr1:
            cmds.append(self._dtr1(0))
            logging.warning('NEED DTR1 for %s of %s' % (cmd.__name__, provider))

        if cmd._uses_dtr2:
            cmds.append(self._dtr2(0))
            logging.warning('NEED DTR2 for %s of %s' % (cmd.__name__, provider))

        cmd_instance = self._prepare_cmd(provider, cmd, *args)

        cmd_instance.sig = self.signature
        cmds.append(cmd_instance)

        return cmds

    def set(self):
        if self.set_cmd:
            cmds = self._apply_cmd(self.entity, self.set_cmd, self.set_value)
            self.entity.on_command_prepared(*cmds)
            if not self.entity.is_device:
                self.entity.on_update(self.funit_type, self.set_value)

        elif self.set_func:
            self.entity.manager.add_command(self.entity, self._apply_func(self.entity, self.set_func.__name__))

        if self.get_cmd and self.get_after_set:
            self.get()

    def get(self):
        if self.easy_get:
            self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
        else:
            if not self.entity.is_device:
                # return
                if self.get_for_not_devices:
                    self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
                    self._variable.clear_signature()
                else:
                    return

            for _provider in self.entity.get_members():
                if isinstance(_provider, self.entity_type):
                    if self.get_cmd:
                        _provider.on_command_prepared(*self._apply_cmd(_provider, self.get_cmd))
                    elif self.get_func:
                        _provider.manager.add_command(_provider, self._apply_func(_provider, self.get_func.__name__))


DaliProvider.funit_class = DaliFunit


class Address(DaliFunit):
    get_for_not_devices = False
    set_func = DaliProvider.SetShortAddress


class Discovery(DaliFunit):
    UP_CMD = dali.gear.general.RecallMaxLevel
    DOWN_CMD = dali.gear.general.Off
    get_for_not_devices = False
    easy_get = True
    set_func = DaliProvider.SetDiscovery


class Groups(DaliFunit):
    length = {1: 16, 2: 32}
    @staticmethod
    def def_value(dali_type) -> list:
        return {
            1: [{'index': i, 'value': False} for i in range(0, Groups.length[1])],
            2: [{'index': i, 'value': False} for i in range(0, Groups.length[2])]
        }.get(dali_type)

    set_func = DaliProvider.SetGroups
    get_func = DaliProvider.GetGroups


class _MemoryValue(DaliFunit):
    get_for_not_devices = False
    bank_id = 0
    start = None
    end = None

    def _get(self, sig=None):
        self.entity.on_update(
            self.funit_type,
            self.to_string(self.entity.read_memory_location(self.bank_id, self.start, self.end)),
            sig=sig
        )

    def get(self):
        if self.entity.is_device:
            cmd = MultiCommand(self._get)
            cmd.set_signature(self.funit_type, session_id=self._variable.session_id, key=self._variable.key)
            self.entity.manager.add_command(self.entity, cmd)

    @classmethod
    def as_hexed(cls, _bytes): return ''.join([hex(b)[2:].rjust(2, '0').upper() for b in _bytes])

    @classmethod
    def as_int(cls, _bytes): return '.'.join([str(b) for b in _bytes])

    to_string = as_hexed

    @classmethod
    def update_part(cls, entity: DaliProvider, value: int, location_id: int, bank_id: int) -> (DaliFunit, str):
        for _f in (Gtin, Serial, GtinOem, SerialOem, FirmwareVersion, HardwareVersion):
            if bank_id == _f.bank_id and _f.start <= location_id < _f.end:
                index = location_id-_f.start
                if _f.to_string == _f.as_hexed:
                    v = entity.get_value(_f, '..'*(_f.end - _f.start))
                    v = v[:index*2] + hex(value)[2:].rjust(2, '0').upper() + v[(index+1)*2:(_f.end - _f.start)*2]
                    return _f, v
                else:
                    v = entity.get_value(_f, '.').split('.')
                    v[index] = str(value)
                    return _f, '.'.join(v)

        raise BaseException('Unknown memory part of [%s, %s]' % (bank_id, location_id))


class Gtin(_MemoryValue):
    start = 0x3
    end = 0x9


class Serial(_MemoryValue):
    start = 0xb
    end = 0x13


class GtinOem(_MemoryValue):
    bank_id = 1
    start = 0x3
    end = 0x9


class SerialOem(_MemoryValue):
    bank_id = 1
    start = 0x9
    end = 0x11


class FirmwareVersion(_MemoryValue):
    start = 0x9
    end = 0xb
    to_string = _MemoryValue.as_int


class HardwareVersion(_MemoryValue):
    start = 0x13
    end = 0x15
    to_string = _MemoryValue.as_int


class Binding(DaliFunit):
    get_for_not_devices = False
    set_func = DaliProvider.SetBinding
    get_func = DaliProvider.GetBinding


class BindingDevice(DaliFunit):
    get_for_not_devices = False
    set_func = DaliProvider.SetBindingDevice
    get_func = DaliProvider.GetBindingDevice


class BindingGroup(DaliFunit):
    get_for_not_devices = False
    set_func = DaliProvider.SetBindingGroup
    get_func = DaliProvider.GetBindingGroup


class Validity(DaliFunit):
    easy_get = True
