import struct

from spread_core.bam import MultiCommand
from spread_core.mqtt.spread.variable import Variable
from spread_core.tools.timer import Timer
import spread_core.protocols.dali as dali
from spread_core.bam.dali.providers.providers import DaliProvider, DaliFunit, Groups, GtinOem, Gtin, Serial, SerialOem, \
    HardwareVersion, FirmwareVersion, Validity, Discovery

""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                                  DALI 2                                """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


class DaliSensor(DaliProvider):
    _type_id = None
    _dali_type = 2

    def __init__(self, data):
        super(DaliSensor, self).__init__(data)

        if 'instanceIndex' in data['attributes']:
            self._instance = dali.address.InstanceNumber(data['attributes']['instanceIndex'])
        else:
            self._instance = dali.address.InstanceType(self.type_id)

    @property
    def instance(self):
        return self._instance

    def compare_instance(self, instance: dali.address.Instance):
        if isinstance(instance, dali.address.InstanceType):
            return self.type_id == instance.value
        elif isinstance(instance, dali.address.InstanceNumber):
            return self.get_value(InstanceIndex) == instance.value

        return False

    @property
    def type_id(self):
        return self._type_id

    def repair_on_fail(self) -> tuple:
        return (*super(DaliSensor, self).repair_on_fail(), On,)

    def check_valid(self):
        super(DaliSensor, self).check_valid()
        if self.is_device:
            self.request(dali.device.general.QueryInstanceStatus(self.short_addr.address, self.instance))

    def on_survey(self):
        yield from super(DaliSensor, self).on_survey()
        yield dali.device.general.QueryInstanceStatus(self.short_addr.address, self.instance)
        
        if self.is_valid_strong: yield dali.device.general.QueryDeviceGroupsZeroToSeven(self.short_addr)
        if self.is_valid_strong: yield dali.device.general.QueryDeviceGroupsEightToFifteen(self.short_addr)
        if self.is_valid_strong: yield dali.device.general.QueryDeviceGroupsSixteenToTwentyThree(self.short_addr)
        if self.is_valid_strong: yield dali.device.general.QueryDeviceGroupsTwentyFourToThirtyOne(self.short_addr)
        if self.is_valid_strong: yield dali.device.general.QueryInstanceEnabled(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.general.QueryNumberOfInstances(self.short_addr)
        if self.is_valid_strong: yield dali.device.general.QueryEventScheme(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.general.QueryResolution(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.general.QueryEventPriority(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.general.QueryEventFilterL(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.general.QueryOperatingMode(self.short_addr)
        if self.is_valid_strong: yield dali.device.general.QueryPrimaryInstanceGroup(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.general.QueryInstanceGroup1(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.general.QueryInstanceGroup2(self.short_addr, self.instance)
        if self.is_valid_strong: yield MultiCommand(self.get_event_value)

    def get_info(self):
        super(DaliSensor, self).get_info()
        self.on_update(InstanceType, self.type_id)

        if self.is_device:
            self.request(dali.device.general.QueryInstanceEnabled(self.short_addr, self.instance))

            if InstancesNumber not in self.values:
                self.request(dali.device.general.QueryNumberOfInstances(self.short_addr))
            if InstanceIndex not in self.values:
                self.GetInstanceIndex()
            if EventScheme not in self.values:
                self.request(dali.device.general.QueryEventScheme(self.short_addr, self.instance))
            if self.get_value(Resolution, 0) == 0:
                self.request(dali.device.general.QueryResolution(self.short_addr, self.instance))
            if self.get_value(On, False):
                self.get_event_value()
            if EventPriority not in self.values:
                self.request(dali.device.general.QueryEventPriority(self.short_addr, self.instance))
            if EventFilter not in self.values:
                self.request(dali.device.general.QueryEventFilterL(self.short_addr, self.instance))
            if OperationMode not in self.values:
                self.request(dali.device.general.QueryOperatingMode(self.short_addr))
            if Group0 not in self.values:
                self.request(dali.device.general.QueryPrimaryInstanceGroup(self.short_addr, self.instance))
            if Group1 not in self.values:
                self.request(dali.device.general.QueryInstanceGroup1(self.short_addr, self.instance))
            if Group2 not in self.values:
                self.request(dali.device.general.QueryInstanceGroup2(self.short_addr, self.instance))
            if FeatureTypes not in self.values:
                self.fill_feature_types()
        else:
            self.on_update(On, False)

            for funit_type in (InstancesNumber, InstanceIndex, EventScheme, Resolution, OperationMode,
                               Group0, Group1, Group2, EventPriority):
                self.on_update(funit_type, 0)

            for funit_type in (EventScheme, FeatureTypes):
                self.on_update(funit_type, [])

    def get_event_value(self, sig=None):
        resolution = self.get_value(Resolution, 0)
        if resolution > 0:
            r = self.manager.send_interface.send(dali.device.general.QueryInputValue(self.short_addr, self.instance))
            if r:
                _val = bin(r.value.as_integer)[2:].rjust(8, '0')
                while r:
                    r = self.manager.send_interface.send(dali.device.general.QueryInputValueLatch(self.short_addr, self.instance))
                    if r:
                        _val += bin(r.value.as_integer)[2:].rjust(8, '0')

                _val = _val.ljust(resolution, '0')[:resolution]
                _val = int(_val[:10], 2)
                return _val

    def fill_feature_types(self, sig=None):
        self.check_command_for_device()
        resp = self.manager.send_interface.send(dali.device.general.QueryFeatureType(self.short_addr, self.instance))
        _arr = []
        if resp:
            if resp.value.as_integer == 0xff:
                _r = None
                while _r is None or _r.value.as_integer == 0xfe or _r.value.as_integer == 0xff:
                    _r = self.request(dali.device.general.QueryNextFeatureType(self.short_addr, self.instance))
                    if _r is None or _r.value.as_integer == 0xfe:
                        break
                    _arr.append(_r.value.as_integer)
            elif resp.value.as_integer != 0xfe:
                _arr = [resp.value.as_integer]

        self.on_update(FeatureTypes, _arr, sig=sig)

    def on_command_sended(self, cmd):
        if isinstance(cmd, (dali.device.general.AddToDeviceGroupsZeroToFifteen,
                            dali.device.general.AddToDeviceGroupsSixteenToThirtyOne,
                            dali.device.general.RemoveFromDeviceGroupsZeroToFifteen,
                            dali.device.general.RemoveFromDeviceGroupsSixteenToThirtyOne)):
            start_index = 0 if isinstance(cmd, (dali.device.general.AddToDeviceGroupsZeroToFifteen,
                                                dali.device.general.RemoveFromDeviceGroupsZeroToFifteen)) else 16

            adding = isinstance(cmd, (dali.device.general.AddToDeviceGroupsZeroToFifteen,
                                      dali.device.general.AddToDeviceGroupsSixteenToThirtyOne))
            raw_value = self.dtr(1) + (self.dtr(2) << 8)

            value = Groups.def_value(self._dali_type)
            value = [{**i} for i in self.get_value(Groups, value)]
            for i in range(0, 16):
                if 1 << i & raw_value:
                    value[start_index+i]['value'] = adding
            self.update_recursive(Groups, value, **{Variable.PRELIM: True})
        elif isinstance(cmd, (dali.device.general.EnableInstance, dali.device.general.DisableInstance)):
            self.update_recursive(On, isinstance(cmd, dali.device.general.EnableInstance), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.general.SetEventScheme):
            self.update_recursive(EventScheme, EventScheme.of_raw(self.dtr(0)), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.general.SetEventPriority):
            self.update_recursive(EventPriority, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.general.SetEventFilter):
            value = 0
            for i in range(0, 3):
                value += (self.dtr(i) >> i)
            mask = EventFilter.FILTERS.get(self.type_id, [])
            self.update_recursive(EventFilter, [mask[i] for i in range(0, 8) if value & 2 ** i != 0], **{Variable.PRELIM: True})
        else:
            super(DaliSensor, self).on_command_sended(cmd)

    def on_request_answer(self, funit_type, response, sig=None, *args):
        if isinstance(response, dali.device.general.QueryInstanceStatusResponse):
            self.update_recursive(Validity, not response.error, sig=sig)
        elif funit_type == EventFilter and response:
            mask = EventFilter.FILTERS.get(self.type_id, [])
            self.update_recursive(funit_type, [mask[i] for i in range(0, 8) if response.value.as_integer & 2 ** i != 0], sig=sig)
        elif funit_type == HoldTime:
            self.update_recursive(funit_type, response.value.as_integer if response else None, sig=sig)
            self.update_recursive(EngineryHoldTime, (response.value.as_integer * 10) if response else None)
        else:
            super(DaliSensor, self).on_request_answer(funit_type, response, sig, *args)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        super(DaliSensor, self).on_update(funit_type, value, sig, **kwargs)

        if funit_type in [InstancesNumber, Groups, OperationMode, Gtin, GtinOem, Serial, SerialOem, FirmwareVersion, HardwareVersion]:
            for _provider in self.manager.providers_by_address(self._dali_type, self.short_addr.address):
                if isinstance(_provider, DaliSensor):
                    if self.id != _provider.id:
                        if _provider.get_value(funit_type) != value:
                            _provider.on_update(funit_type, value, **kwargs)

    def on_event(self, event):
        raise NotImplementedError()

    def read_memory_location(self, bank_id, start_address, end_address) -> bytes:
        self.check_command_for_device()
        i = self.manager.send_interface

        i.send(dali.device.general.DTR1(bank_id))

        for _dtr0 in range(start_address, end_address):
            i.send(dali.device.general.DTR0(_dtr0))
            res = i.send(dali.device.general.ReadMemoryLocation(self.short_addr.address))
            if isinstance(res, dali.command.Response):
                yield res.value.as_integer
            else:
                break

    def SetShortAddress(self, new_addr, sig=None):
        if self.is_device:
            if new_addr != self.short_addr.address:
                values = None
                old_short = self.short_addr
                i = self.manager.send_interface
                if new_addr is None:
                    i(dali.device.general.DTR0(0xff))
                    i(dali.device.general.SetShortAddress(old_short))
                    for _old in self.manager.providers_by_address(self._dali_type, old_short.address):
                        if isinstance(_old, DaliSensor):
                            _old.reset()
                    return
                else:
                    new_short = dali.address.Short(new_addr)
                    res = i(dali.device.general.QueryNumberOfInstances(new_short))
                    if res is None:
                        i(dali.device.general.DTR0(new_addr))
                        i(dali.device.general.SetShortAddress(old_short))

                        for _old in self.manager.providers_by_address(self._dali_type, old_short.address):
                            if isinstance(_old, DaliSensor):
                                _old.SetBindingDevice(new_addr)
                    else:
                        for _provider in self.manager.providers_by_address(self._dali_type, new_addr):
                            if isinstance(_provider, self.__class__) and _provider.instance == self.instance:
                                old_long = []
                                new_long = []

                                i(dali.device.general.Terminate())

                                i(dali.device.general.Initialise(old_short.address))
                                i(dali.device.general.Initialise(new_short.address))

                                old_long.append(i(dali.device.general.QueryRandomAddressH(old_short)))
                                old_long.append(i(dali.device.general.QueryRandomAddressM(old_short)))
                                old_long.append(i(dali.device.general.QueryRandomAddressL(old_short)))

                                new_long.append(i(dali.device.general.QueryRandomAddressH(new_short)))
                                new_long.append(i(dali.device.general.QueryRandomAddressM(new_short)))
                                new_long.append(i(dali.device.general.QueryRandomAddressL(new_short)))

                                i(dali.device.general.SearchAddrH(old_long[0].value.as_integer))
                                i(dali.device.general.SearchAddrM(old_long[1].value.as_integer))
                                i(dali.device.general.SearchAddrL(old_long[2].value.as_integer))
                                i(dali.device.general.ProgramShortAddress(new_short.address))

                                i(dali.device.general.SearchAddrH(new_long[0].value.as_integer))
                                i(dali.device.general.SearchAddrM(new_long[1].value.as_integer))
                                i(dali.device.general.SearchAddrL(new_long[2].value.as_integer))
                                i(dali.device.general.ProgramShortAddress(old_short.address))

                                i(dali.device.general.Terminate())

                                values = _provider.values.copy()

                                for funit_type, value in self.values.items():
                                    _provider.on_update(funit_type, value)

                                for _old in self.manager.providers_by_address(self._dali_type, old_short.address):
                                    if _old != _provider and isinstance(_old, DaliSensor):
                                        for _new in self.manager.providers_by_address(self._dali_type, new_addr):
                                            if _new != _provider and isinstance(_new, _old.__class__):
                                                _values = _new.values.copy()

                                                for funit_type, value in _old.values.items():
                                                    _new.on_update(funit_type, value)

                                                for funit_type, value in _values.items():
                                                    _old.on_update(funit_type, value)

                                                break

                                break

                self.SetBindingDevice(new_addr, sig=sig, values=values)
            else:
                raise dali.exceptions.CommandGenerateError('You can not give an address to {}'.format(self.binding.b_type))

    def SetEnable(self, value, sig=None):
        if self.get_value('On') != value:
            if value is True:
                self.request(dali.device.general.EnableInstance(self.short_addr, self.instance))
            else:
                self.request(dali.device.general.DisableInstance(self.short_addr, self.instance))

            for _provider in self.get_members():
                cmd = dali.device.general.QueryInstanceEnabled(_provider.short_addr, _provider.instance)
                cmd.sig = sig
                self.request(cmd)

    def SetDiscovery(self, value, sig=None):
        _discovery = self.get_value(Discovery, False)
        if _discovery != value:
            self.on_update(Discovery, value, sig=sig)

            if _discovery is False and value is True:
                self.check_discovery()
            else:
                self.request(dali.device.general.Terminate())

    def check_discovery(self):
        if self.get_value(Discovery) is True:
            self.request(dali.device.general.IdentifyDevice(self.short_addr))
            Timer.set(7, self.check_discovery)

    def GetGroups(self, sig=None):
        if self.is_device:
            self.request(dali.device.general.QueryDeviceGroupsZeroToSeven(self.short_addr), sig=sig)
            self.request(dali.device.general.QueryDeviceGroupsEightToFifteen(self.short_addr), sig=sig)
            self.request(dali.device.general.QueryDeviceGroupsSixteenToTwentyThree(self.short_addr), sig=sig)
            self.request(dali.device.general.QueryDeviceGroupsTwentyFourToThirtyOne(self.short_addr), sig=sig)
        else:
            _groups = [{'index': self.binding.addr.group, 'value': True}] if self.is_group else []
            _groups = self.get_value(Groups, _groups)
            self.on_update(Groups, _groups, sig=sig)

    def SetGroups(self, val, sig=None):
        dtrs = [0, 0, 0, 0]
        for item in val:
            if item['value']:
                dtrs[int(item['index'] / 16)] += 1 << (item['index'] % 16)
            else:
                dtrs[2+int(item['index'] / 16)] += 1 << (item['index'] % 16)

        for i in range(0, 4):
            if dtrs[i]:
                _dtr1, _dtr2 = struct.pack('<H', dtrs[i])
                self.request(dali.device.general.DTR2DTR1(_dtr2, _dtr1))
                self.request(
                    (
                        dali.device.general.AddToDeviceGroupsZeroToFifteen,
                        dali.device.general.AddToDeviceGroupsSixteenToThirtyOne,
                        dali.device.general.RemoveFromDeviceGroupsZeroToFifteen,
                        dali.device.general.RemoveFromDeviceGroupsSixteenToThirtyOne
                    )[i](self.short_addr)
                )

        if self.is_device:
            self.GetGroups(sig)
        else:
            self.update_recursive(Groups, val)

    def GetInstanceIndex(self, sig=None):
        self.check_command_for_device()
        val = None
        if isinstance(self.instance, dali.address.InstanceType):
            if InstancesNumber in self.values:
                for instance_num in range(0, self.values[InstancesNumber]):
                    tp = self.manager.send_interface.send(
                        dali.device.general.QueryInstanceType(self.short_addr, dali.address.InstanceNumber(instance_num)))
                    if tp:
                        if tp.value.as_integer == self.instance._value:
                            val = instance_num
                            break
        elif isinstance(self.instance, dali.address.InstanceNumber):
            val = self.instance._value

        self.on_update(InstanceIndex, self.get_value(InstanceIndex, val), sig=sig)

    def serialize(self) -> dict:
        d = super(DaliSensor, self).serialize()
        if d:
            for funit_type in [OperationMode, Groups, InstancesNumber]:
                value = self.get_value(funit_type)
                if value is not None:
                    key = str(funit_type)[0].lower() + str(funit_type)[1:]
                    d[key] = value
        return d

    def serialize_instance(self) -> dict:
        i = {}
        value = self.get_value(InstanceIndex)
        if value is None: return {}
        else: i['index'] = value

        value = self.get_value(InstanceIndex)
        if value: i['type'] = value

        value = self.get_value(On)
        if value is not None: i['enabled'] = value

        for funit_type in [Group0, Group1, Group2, Resolution, EventScheme, EventPriority, FeatureTypes,
                           EventFilter, DeadTime, HoldTime, ReportTime]:
            value = self.get_value(funit_type)
            if value is not None:
                key = str(funit_type)[0].lower() + str(funit_type)[1:]
                i[key] = value
        return i


""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                               Light Sensor                             """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


class DaliLightSensor(DaliSensor):
    _type_id = 4

    def repair_on_fail(self) -> tuple:
        return (*super(DaliSensor, self).repair_on_fail(), CurrentLuminosity,)

    def on_event(self, event):
        self.on_update(CurrentLuminosity, event.data.value)

    def get_info(self):
        super(DaliLightSensor, self).get_info()
        if self.is_device:
            if DeadTime not in self.values:
                self.request(dali.device.light_sensor_ext.QueryDeadTime(self.short_addr, self.instance))
            if ReportTime not in self.values:
                self.request(dali.device.light_sensor_ext.QueryReportTime(self.short_addr, self.instance))
            if Hysteresis not in self.values:
                self.request(dali.device.light_sensor_ext.QueryHysteresis(self.short_addr, self.instance))
            if HysteresisMin not in self.values:
                self.request(dali.device.light_sensor_ext.QueryHysteresisMin(self.short_addr, self.instance))
        else:
            for funit_type in (DeadTime, ReportTime, Hysteresis, HysteresisMin):
                self.on_update(funit_type, 0)

    def on_command_sended(self, cmd):
        if isinstance(cmd, dali.device.light_sensor_ext.SetHysteresis):
            self.update_recursive(Hysteresis, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.light_sensor_ext.SetHysteresisMin):
            self.update_recursive(HysteresisMin, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.light_sensor_ext.SetDeadTime):
            self.update_recursive(DeadTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.light_sensor_ext.SetReportTime):
            self.update_recursive(ReportTime, self.dtr(0), **{Variable.PRELIM: True})
        else:
            super(DaliLightSensor, self).on_command_sended(cmd)
                
    def on_survey(self):
        yield from super(DaliLightSensor, self).on_survey()
        
        if self.is_valid_strong: yield dali.device.light_sensor_ext.QueryDeadTime(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.light_sensor_ext.QueryReportTime(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.light_sensor_ext.QueryHysteresis(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.light_sensor_ext.QueryHysteresisMin(self.short_addr, self.instance)

    def serialize_instance(self) -> dict:
        d = super(DaliLightSensor, self).serialize_instance()
        for funit_type in [Hysteresis, HysteresisMin]:
            value = self.get_value(funit_type)
            if value is not None:
                key = str(funit_type)[0].lower() + str(funit_type)[1:]
                d[key] = value
        return d

    def get_event_value(self, sig=None):
        self.on_update(CurrentLuminosity, super().get_event_value(sig))


""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                             Presence Sensor                            """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""


class DaliPresenceSensor(DaliSensor):
    _type_id = 3

    def repair_on_fail(self) -> tuple:
        return (*super(DaliSensor, self).repair_on_fail(), CurrentPresence,)

    def on_event(self, event):
        self.on_update(CurrentPresence, event.data.value)

    def get_info(self):
        super(DaliPresenceSensor, self).get_info()
        if self.is_device:
            if DeadTime not in self.values:
                self.request(dali.device.presence_sensor_ext.QueryDeadTime(self.short_addr, self.instance))
            if ReportTime not in self.values:
                self.request(dali.device.presence_sensor_ext.QueryReportTime(self.short_addr, self.instance))
            if HoldTime not in self.values:
                self.request(dali.device.presence_sensor_ext.QueryHoldTime(self.short_addr, self.instance))
        else:
            for funit_type in (DeadTime, ReportTime, HoldTime):
                self.on_update(funit_type, 0)

    def on_survey(self):
        yield from super(DaliPresenceSensor, self).on_survey()

        if self.is_valid_strong: yield dali.device.presence_sensor_ext.QueryDeadTime(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.presence_sensor_ext.QueryReportTime(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.presence_sensor_ext.QueryHoldTime(self.short_addr, self.instance)

    def get_event_value(self, sig=None):
        value = super().get_event_value(sig)
        if value is None:
            result = None
        else:
            result = dali.device.events.PresenceEvent(value=value).value

        self.on_update(CurrentPresence, result)

    def on_command_sended(self, cmd):
        if isinstance(cmd, dali.device.presence_sensor_ext.SetHoldTime):
            self.update_recursive(HoldTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.presence_sensor_ext.SetDeadTime):
            self.update_recursive(DeadTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.presence_sensor_ext.SetReportTime):
            self.update_recursive(ReportTime, self.dtr(0), **{Variable.PRELIM: True})
        else:
            super(DaliPresenceSensor, self).on_command_sended(cmd)


class DaliPushButton(DaliSensor):
    _type_id = 1

    def repair_on_fail(self) -> tuple:
        return (*super(DaliSensor, self).repair_on_fail(), LastEvent,)

    def on_event(self, event):
        self.values.pop(LastEvent, None)
        self.on_update(LastEvent, event.data.value)

    def get_info(self):
        super(DaliPushButton, self).get_info()
        if self.is_device:
            if ShortTimeMin not in self.values:
                self.request(dali.device.push_button_ext.QueryShortTimeMin(self.short_addr, self.instance))
            if ShortTime not in self.values:
                self.request(dali.device.push_button_ext.QueryShortTime(self.short_addr, self.instance))
            if DoubleTime not in self.values:
                self.request(dali.device.push_button_ext.QueryDoubleTime(self.short_addr, self.instance))
            if DoubleTimeMin not in self.values:
                self.request(dali.device.push_button_ext.QueryDoubleTimeMin(self.short_addr, self.instance))
            if RepeatTime not in self.values:
                self.request(dali.device.push_button_ext.QueryRepeatTime(self.short_addr, self.instance))
            if StuckTime not in self.values:
                self.request(dali.device.push_button_ext.QueryStuckTime(self.short_addr, self.instance))
        else:
            for funit_type in (ShortTimeMin, ShortTime, DoubleTime, DoubleTimeMin, RepeatTime, StuckTime):
                self.on_update(funit_type, 0)
    
    def on_survey(self):
        yield from super(DaliPushButton, self).on_survey()
        
        if self.is_valid_strong: yield dali.device.push_button_ext.QueryShortTimeMin(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.push_button_ext.QueryShortTime(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.push_button_ext.QueryDoubleTime(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.push_button_ext.QueryDoubleTimeMin(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.push_button_ext.QueryRepeatTime(self.short_addr, self.instance)
        if self.is_valid_strong: yield dali.device.push_button_ext.QueryStuckTime(self.short_addr, self.instance)

    def get_event_value(self, sig=None):
        event = dali.device.events.ButtonEvent(super().get_event_value(sig))
        self.on_update(LastEvent, event.value, sig=sig)

    def on_command_sended(self, cmd):
        if isinstance(cmd, dali.device.push_button_ext.SetShortTime):
            self.update_recursive(ShortTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.push_button_ext.SetDoubleTime):
            self.update_recursive(DoubleTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.push_button_ext.SetDoubleTime):
            self.update_recursive(DoubleTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.push_button_ext.SetRepeatTime):
            self.update_recursive(RepeatTime, self.dtr(0), **{Variable.PRELIM: True})
        elif isinstance(cmd, dali.device.push_button_ext.SetStuckTime):
            self.update_recursive(StuckTime, self.dtr(0), **{Variable.PRELIM: True})
        else:
            super(DaliPushButton, self).on_command_sended(cmd)


class DaliSensorFunit(DaliFunit):
    entity_type = DaliSensor
    _dtr0 = dali.device.general.DTR0
    _dtr1 = dali.device.general.DTR1
    _dtr2 = dali.device.general.DTR2

    @property
    def entity(self) -> DaliSensor:
        return super(DaliSensorFunit, self).entity

    def _prepare_cmd(self, provider: DaliSensor, cmd, *args):
        if issubclass(cmd, dali.device.general._StandardInstanceCommand):
            return cmd(provider.short_addr, provider.instance, *args)
        elif issubclass(cmd, dali.device.general._StandardDeviceCommand):
            return cmd(provider.short_addr, *args)
        else:
            return cmd(*args)


class On(DaliSensorFunit):
    set_func = DaliSensor.SetEnable
    get_cmd = dali.device.general.QueryInstanceEnabled


class OperationMode(DaliSensorFunit):
    set_cmd = dali.device.general.SetOperatingMode
    get_cmd = dali.device.general.QueryOperatingMode


class Group0(DaliSensorFunit):
    set_cmd = dali.device.general.SetPrimaryInstanceGroup
    get_cmd = dali.device.general.QueryPrimaryInstanceGroup


class Group1(DaliSensorFunit):
    set_cmd = dali.device.general.SetInstanceGroup1
    get_cmd = dali.device.general.QueryInstanceGroup1


class Group2(DaliSensorFunit):
    set_cmd = dali.device.general.SetInstanceGroup2
    get_cmd = dali.device.general.QueryInstanceGroup2


class EventScheme(DaliSensorFunit):
    set_cmd = dali.device.general.SetEventScheme
    get_cmd = dali.device.general.QueryEventScheme
    index_of = ["esIndexAndType", "esAddressAndType", "esAddressAndIndex", "esLowestGroupAndType", "esPrimaryGroupAndType"]


class EventPriority(DaliSensorFunit):
    set_cmd = dali.device.general.SetEventPriority
    get_cmd = dali.device.general.QueryEventPriority


class EventFilter(DaliSensorFunit):
    FILTERS = {
        1: ["buttonReleasedEvents", "buttonPressedEvents", "shortPressEvents", "doublePressEvents",
            "longPressStartEvents", "longPressRepeatEvents", "longPressStopEvents", "buttonStuckFreeEvents"],
        3: ["occupiedEvents", "vacantEvents", "repeatEvents", "movementEvents", "noMovementEvents"],
        4: ["illuminanceLevelEvents"],
    }
    set_cmd = dali.device.general.SetEventFilter
    get_cmd = dali.device.general.QueryEventFilterL
    @property
    def mask(self): return self.FILTERS.get(self.entity.type_id, [])


class InstancesNumber(DaliSensorFunit):
    get_cmd = dali.device.general.QueryNumberOfInstances


class InstanceIndex(DaliSensorFunit):
    get_func = DaliSensor.GetInstanceIndex


class InstanceType(DaliSensorFunit):
    pass


class FeatureTypes(DaliSensorFunit):
    get_func = DaliSensor.fill_feature_types


class Resolution(DaliSensorFunit):
    get_cmd = dali.device.general.QueryResolution


class CurrentLuminosity(DaliSensorFunit):
    entity_type = DaliLightSensor
    easy_get = True


class CurrentPresence(DaliSensorFunit):
    entity_type = DaliPresenceSensor
    easy_get = True


class Hysteresis(DaliSensorFunit):
    entity_type = DaliLightSensor
    set_cmd = dali.device.light_sensor_ext.SetHysteresis
    get_cmd = dali.device.light_sensor_ext.QueryHysteresis


class HysteresisMin(DaliSensorFunit):
    entity_type = DaliLightSensor
    set_cmd = dali.device.light_sensor_ext.SetHysteresisMin
    get_cmd = dali.device.light_sensor_ext.QueryHysteresisMin


class HoldTime(DaliSensorFunit):
    entity_type = DaliPresenceSensor
    set_cmd = dali.device.presence_sensor_ext.SetHoldTime
    get_cmd = dali.device.presence_sensor_ext.QueryHoldTime


class EngineryHoldTime(DaliSensorFunit):
    entity_type = DaliPresenceSensor
    set_cmd = dali.device.presence_sensor_ext.SetHoldTime
    get_cmd = dali.device.presence_sensor_ext.QueryHoldTime
    value_range = range(0, 254*10+1)

    @property
    def set_value(self): return int(super(EngineryHoldTime, self).set_value / 10)


class DeadTime(DaliSensorFunit):
    @property
    def set_cmd(self):
        if isinstance(self.entity, DaliPresenceSensor):
            return dali.device.presence_sensor_ext.SetDeadTime
        elif isinstance(self.entity, DaliLightSensor):
            return dali.device.light_sensor_ext.SetDeadTime

    @property
    def get_cmd(self):
        if isinstance(self.entity, DaliPresenceSensor):
            return dali.device.presence_sensor_ext.QueryDeadTime
        elif isinstance(self.entity, DaliLightSensor):
            return dali.device.light_sensor_ext.QueryDeadTime


class ReportTime(DaliSensorFunit):
    @property
    def set_cmd(self):
        if isinstance(self.entity, DaliPresenceSensor):
            return dali.device.presence_sensor_ext.SetReportTime
        elif isinstance(self.entity, DaliLightSensor):
            return dali.device.light_sensor_ext.SetReportTime

    @property
    def get_cmd(self):
        if isinstance(self.entity, DaliPresenceSensor):
            return dali.device.presence_sensor_ext.QueryReportTime
        elif isinstance(self.entity, DaliLightSensor):
            return dali.device.light_sensor_ext.QueryReportTime


class _DaliPushButtonFunit(DaliSensorFunit):
    entity_type = DaliPushButton


class LastEvent(_DaliPushButtonFunit):
    retained = False
    get_func = DaliPushButton.get_event_value


class ShortTime(_DaliPushButtonFunit):
    value_range = range(10, 256)
    get_cmd = dali.device.push_button_ext.QueryShortTime
    set_cmd = dali.device.push_button_ext.SetShortTime


class ShortTimeMin(_DaliPushButtonFunit):
    get_cmd = dali.device.push_button_ext.QueryShortTimeMin


class DoubleTime(_DaliPushButtonFunit):
    value_range = range(10, 101)
    get_cmd = dali.device.push_button_ext.QueryDoubleTime
    set_cmd = dali.device.push_button_ext.SetDoubleTime


class DoubleTimeMin(_DaliPushButtonFunit):
    get_cmd = dali.device.push_button_ext.QueryDoubleTimeMin


class RepeatTime(_DaliPushButtonFunit):
    value_range = range(5, 101)
    get_cmd = dali.device.push_button_ext.QueryRepeatTime
    set_cmd = dali.device.push_button_ext.SetRepeatTime


class StuckTime(_DaliPushButtonFunit):
    value_range = range(5, 256)
    get_cmd = dali.device.push_button_ext.QueryStuckTime
    set_cmd = dali.device.push_button_ext.SetStuckTime
