import datetime
import json
import logging
import os
import threading
from threading import Timer

from spread_core import mqtt
from spread_core.bam import Funit, subgineries
from spread_core.errors.project_errors import ClassifierError, CommandError
from spread_core.tools.settings import config, FADE_TIME
from .items import incubator as i
from .handler import Handler
from spread_core.mqtt.spread.topic import State, Set, Renew
from spread_core.mqtt.spread.variable import Variable

CONF_TIMER = 'conf_timer'
WORK_TIMER = 'timer'
LOCAL_READER_TIMER = 'local_read_timer'
STATES_READER_TIMER = 'states_read_timer'

STATE_STARTED = 0
STATE_INIT = 1
STATE_WORK = 2
STATE_EXITED = 3


class Cycler(Handler):
    LOCAL_PATH = '/var/project'

    def __init__(self, project_id, _mqttc):
        super(Cycler, self).__init__(project_id, _mqttc)
        self.cycles = {}
        self.incubators = {}
        self.locations = None
        self.triggers = None
        self.timer = None
        self.conf_timer = None
        self.state = STATE_STARTED
        self.local_read_timer = None
        self.states_read_timer = None

    def update_config(self):
        super(Cycler, self).update_config()
        i.PeriodicalTrigger.FADE_TIME = config[FADE_TIME]

    def update(self, incubator, funit_type, value, sig=None):
        if self.state != STATE_WORK: return
        address = mqtt.spread.address.SubgineryAddress(self.project_id,
                                                       incubator.id,
                                                       subgineries.Incubator.__name__,
                                                       funit_type)
        if value is None:
            variable = value
            if funit_type == Cycle:
                self.incubators.pop(incubator.id, None)
                self.update(incubator, Active, None)
                self.update(incubator, Recursive, None)
                self.update(incubator, StartTime, None)
        else:
            variable = Variable(value, key=sig.key if sig else None)
            if sig and sig.session_id:
                self.publish(mqtt.spread.topic.Reply(self.project_id, sig.session_id), variable, retain=False)

        self.publish(State(address), variable, retain=True)

        self.dump(os.path.join(self.LOCAL_PATH, str(address)), value)

    def read_states(self):
        logging.warning('Reading states data...')
        dest = os.path.join(self.LOCAL_PATH, str(self.project_id), 'Location')
        for dirpath, dirnames, filenames in os.walk(dest):
            if i.Cycle.__name__ in filenames:
                for file in (i.Cycle.__name__, *(fn for fn in filenames if fn != i.Cycle.__name__)):
                    with open(os.path.join(dirpath, file), 'r') as f:
                        data = json.load(f)
                    topic = '/'.join((dirpath.replace(self.LOCAL_PATH, '/'.join((State.protocol, State.action))), file))
                    # print(topic, data)
                    self.publish(topic, Variable(data), retain=True)

        self.restart_timer(1, CONF_TIMER, self.on_ready)

    def read_local(self):
        logging.warning('Reading local data...')
        dest = os.path.join(self.LOCAL_PATH, str(self.project_id), mqtt.local.Local.protocol, subgineries.Incubator.__name__)
        for dirpath, dirnames, filenames in os.walk(dest):
            for file in filenames:
                with open(os.path.join(dirpath, file), 'r') as f:
                    data = json.load(f)
                topic = '/'.join((dirpath.replace(self.LOCAL_PATH, mqtt.local.Local.p_protocol), file))
                # print(topic, data)
                triggers = data.pop('triggers', {})
                self.publish(topic, data, retain=True)
                for t in triggers:
                    self.publish('/'.join((topic, mqtt.local.CycleTrigger.type, t)), triggers[t], retain=True)

    def restart_timer(self, time, timer_name, handler):
        timer = self.__getattribute__(timer_name)
        if isinstance(timer, threading.Timer):
            timer.cancel()

        if handler:
            timer = threading.Timer(time, handler)
            timer.start()
        else:
            timer = None

        self.__setattr__(timer_name, timer)

    def start(self):
        self.state = STATE_INIT
        self.locations = {}
        self.triggers = []
        self.subscribe(mqtt.project.TopicProject.locations(self.project_id))

    def subscribe_states(self):
        if self.state != STATE_WORK:
            entity = mqtt.spread.address.SubgineryAddress(self.project_id, '+', i.Incubator.__name__, '+')
            self.restart_timer(10, STATES_READER_TIMER, self.read_states)

            self.unsubscribe(Set(entity))
            self.unsubscribe(Renew(entity))
            self.subscribe(State(entity))

    def on_spread(self, topic: mqtt.spread.topic.SpreadTopic, data: bytes):
        self.restart_timer(0, STATES_READER_TIMER, None)

        variable = Variable.of(data)
        command = _IncFunit.of(topic.entity_address.funit_type)
        self.incubators.setdefault(
            topic.entity_address.id,
            i.Incubator(topic.entity_address.id, **self.locations.get(topic.entity_address.id, {}))
        )

        if command and issubclass(command, Funit):
            try:
                if isinstance(topic, (Set, State)):
                    command(self, self.incubators[topic.entity_address.id], variable).set()
                elif isinstance(topic, Renew):
                    command(self, self.incubators[topic.entity_address.id], variable).get()
                else:
                    raise Funit.NotImplemented()
            except Funit.NotImplemented:
                raise ClassifierError.of_action(topic.__class__.__name__, self, topic.entity_address.funit_type)
            except BaseException as ex:
                raise CommandError.of_exec(str(ex))

            if isinstance(topic, State):
                self.restart_timer(1, CONF_TIMER, self.on_ready)

        else:
            raise ClassifierError.of_funit(topic.entity_address.funit_type, self)

    def on_project(self, topic: mqtt.project.TopicProject, data: bytes):
        if topic.file_name == topic.LOCATIONS + '.json':
            self.locations = {
                loc['id']: {'name': loc['name']}
                for loc in json.loads(data.decode())[topic.LOCATIONS]
            }
            self.subscribe(mqtt.project.TopicProject.subgineries(self.project_id))

        elif topic.file_name == topic.SUBGINERIES + '.json':
            for sb in json.loads(data.decode())[topic.SUBGINERIES]:
                try:
                    if sb['type'] == subgineries.Lighting.__name__ and sb['locationID'] in self.locations:
                        self.locations[sb['locationID']]['subginery_id'] = sb['id']
                except KeyError:
                    continue

            self.unsubscribe(mqtt.project.TopicProject.subgineries(self.project_id))
            self.restart_timer(10, LOCAL_READER_TIMER, self.read_local)
            self.subscribe(mqtt.local.Cycles.broadcast(self.project_id))

    def on_cycle(self, topic: mqtt.local.Cycles, data: bytes):
        self.restart_timer(0, LOCAL_READER_TIMER, None)
        to_dump = None
        if data:
            self.cycles.setdefault(topic.entity_id, i.Cycle(topic.entity_id))
            self.cycles[topic.entity_id].update(json.loads(data.decode()))
            to_dump = self.cycles[topic.entity_id].as_dict
        else:
            self.cycles.pop(topic.entity_id, None)

        self.dump(str(topic).replace(topic.p_protocol, self.LOCAL_PATH), to_dump)

        if self.state == STATE_WORK:
            for inc in tuple(self.incubators.values()):
                if inc.cycle and inc.cycle.id == topic.entity_id:
                    if inc.id in self.incubators:
                        Cycle(self, self.incubators[inc.id], Variable(topic.entity_id if data else None)).set()
        else:
            self.restart_timer(1, CONF_TIMER, self.subscribe_states)

    def on_trigger(self, topic: mqtt.local.CycleTrigger, data: bytes):
        self.restart_timer(0, LOCAL_READER_TIMER, None)
        data = None if data == b'' else json.loads(data.decode())
        if topic.cycle_id in self.cycles:
            if data:
                self.cycles[topic.cycle_id].triggers[topic.entity_id] = data
            else:
                self.cycles[topic.cycle_id].triggers.pop(topic.entity_id, None)

            self.dump(
                str(mqtt.local.Cycles(self.project_id, topic.cycle_id)).replace(topic.p_protocol, self.LOCAL_PATH),
                self.cycles[topic.cycle_id].as_dict
            )
        else:
            logging.warning('Cycle(%s) not exist' % topic.cycle_id)

        if self.state == STATE_WORK:
            for inc in tuple(self.incubators.values()):
                if inc.cycle and inc.cycle.id == topic.cycle_id:
                    if inc.id in self.incubators:
                        Cycle(self, inc, Variable(topic.cycle_id)).set()

        if self.state == STATE_INIT:
            self.restart_timer(1, CONF_TIMER, self.subscribe_states)

    def on_ready(self):
        self.state = STATE_WORK
        entity = mqtt.spread.address.SubgineryAddress(self.project_id, '+', 'Incubator', '+')

        self.unsubscribe(State(entity))
        self.subscribe(Set(entity))
        self.subscribe(Renew(entity))

    def on_timer(self):
        try:
            while len(self.triggers) > 0:
                self.triggers.sort()
                trigger = self.triggers[0]
                interval = trigger.time_left + 1
                if interval < 1:
                    self.increment_trigger(trigger)
                else:
                    logging.info(f'SET TIMER to {datetime.timedelta(seconds=interval)} sec FOR {trigger}')
                    if self.timer:
                        self.timer.cancel()
                    self.timer = Timer(interval=interval, function=self.exec_task, args=[trigger])
                    self.timer.start()
                    break
            if len(self.triggers) == 0:
                logging.info('NO MOVE')
        except BaseException as ex:
            logging.exception(ex)

    def exec_task(self, _: i.PeriodicalTrigger):
        for _trigger in self.triggers.copy():
            if _trigger.time_left > 0: break
            logging.info(f'EXECUTE {_trigger}:')
            self.publish(
                _trigger.topic(self.project_id),
                Variable(_trigger.current_value, session=self.sessionID),
                log_level=logging.INFO
            )
            self.increment_trigger(_trigger)

        self.on_timer()

    def increment_trigger(self, trigger: i.PeriodicalTrigger):
        trigger.increment_datetime()
        if not trigger.has_next:
            try:
                self.triggers.remove(trigger)
            except:
                pass

    def on_exit(self):
        super(Cycler, self).on_exit()
        self.restart_timer(0, CONF_TIMER, None)
        self.restart_timer(0, WORK_TIMER, None)
        self.restart_timer(0, LOCAL_READER_TIMER, None)
        self.restart_timer(0, STATES_READER_TIMER, None)

    @staticmethod
    def dump(file_path: str, data):
        try:
            if data is not None:
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
                with open(file_path, "w") as f:
                    json.dump(data, f)
            elif os.path.exists(file_path):
                os.remove(file_path)

        except BaseException as ex:
            logging.error(ex)


class _IncFunit(Funit):
    entity_type = i.Incubator
    attname = None
    _entity = None

    def __init__(self, cycler: Cycler, entity: i.Incubator, variable: Variable):
        self.cycler = cycler
        super(_IncFunit, self).__init__(entity, variable)

    @property
    def entity(self) -> i.Incubator:
        return super(_IncFunit, self).entity

    def set(self):
        if self.entity.calculated:
            for trigger in self.entity.triggers:
                try:
                    if trigger in self.cycler.triggers:
                        self.cycler.triggers.remove(trigger)
                except:
                    pass

        setattr(self.entity, self.attname, self.set_value)

        if self.entity.active and self.entity.triggers:
            for trigger in self.entity.triggers:
                if trigger.has_next:
                    self.cycler.triggers.append(trigger)

        self.cycler.restart_timer(1, WORK_TIMER, self.cycler.on_timer)
        self.cycler.update(self.entity, self.funit_type, self._variable.value, sig=self.signature)


class Cycle(_IncFunit):
    attname = 'cycle'

    @property
    def set_value(self):
        return self.cycler.cycles.get(super(Cycle, self).set_value, None)


class Recursive(_IncFunit):
    attname = 'recursive'


class Active(_IncFunit):
    attname = 'active'


class StartTime(_IncFunit):
    attname = 'startTime'
