import datetime

from spread_core import mqtt
from spread_core.bam import subgineries
from .triggers import DateTrigger


class Cycle:
    __slots__ = ['id', 'name', 'period', 'duration', 'startLevel', 'triggers']

    def __init__(self, _id: int):
        self.id = _id
        self.triggers = {}

    def update(self, data: dict):
        for attname, value in data.items():
            if attname in self.__slots__:
                setattr(self, attname, value)

    def __str__(self):
        return '%s<%s>' % (self.name, self.id)

    def __repr__(self): return self.__str__()

    @property
    def as_dict(self):
        return dict(
            name=self.name,
            period=self.period,
            duration=self.duration,
            startLevel=self.startLevel,
            triggers=self.triggers
        )


class Incubator:
    __slots__ = ["id", "name", "subginery_id", "_startTime", "_cycle", "_active", "_recursive", "_triggers"]

    def __init__(self, _id, name, subginery_id):
        self.id = _id
        self.name = name
        self.subginery_id = subginery_id
        self._cycle = None
        self._recursive = False
        self._active = False
        self._triggers = None
        self._startTime = None

    def __str__(self):
        return '%s of %s' % (self.subginery_id, self._cycle)

    def __repr__(self): return self.__str__()

    def set_trigger(self, trigger):
        self._triggers.append(trigger)

    def sort(self):
        self._triggers.sort(key=lambda tr: tr.datetime)

    @property
    def ready(self):
        return self._cycle is not None and self._startTime is not None

    @property
    def calculated(self): return self._triggers is not None

    @property
    def triggers(self):
        if self._triggers is None and self.ready:
            self._triggers = []

            for t_id, trigger_data in self.cycle.triggers.items():
                trigger = PeriodicalTrigger.of({'id': t_id, **trigger_data}, self)
                if trigger.day < self._cycle.period:
                    self._triggers.append(trigger)

            self.sort()
            last_trigger = None
            for trigger in self._triggers:
                if isinstance(trigger, PeriodicalTrigger):
                    trigger.increment_datetime()
                    if last_trigger:
                        trigger.last_trigger = last_trigger
                        last_trigger.next_trigger = trigger
                    else:
                        trigger.last_trigger = self.cycle.startLevel

                    last_trigger = trigger

        return self._triggers

    @property
    def recursive(self): return self._recursive

    @recursive.setter
    def recursive(self, value):
        self._triggers = None
        self._recursive = value

    @property
    def active(self): return self._active

    @active.setter
    def active(self, value):
        self._active = value

    @property
    def startTime(self): return self._startTime

    @startTime.setter
    def startTime(self, value):
        self._triggers = None
        self._startTime = value

    @property
    def cycle(self) -> Cycle: return self._cycle

    @cycle.setter
    def cycle(self, cycle: Cycle):
        self._triggers = None
        self._cycle = cycle


class PeriodicalTrigger(DateTrigger):
    FADE_TIME = 2

    def __init__(self, data):
        super(PeriodicalTrigger, self).__init__(**data)
        self.day = None
        self.time = None
        self._incubator = None
        self._datetime = None
        self._step_datetime = None
        self.actions = []
        self.level = None
        self.last_trigger = None
        self.next_trigger = None

    def __str__(self): return '<[%s]%s>(%sй день %s)%s..%s' % (self.incubator.id, self.incubator.cycle.name, self.day, self.datetime.time(), self.level_from, self.level)

    def topic(self, project_id):
        return str(mqtt.spread.topic.Set(
            mqtt.spread.address.SubgineryAddress(
                project_id,
                self.incubator.subginery_id,
                subgineries.Lighting.__name__,
                subgineries.BrightnessLevel
            )
        ))

    @property
    def level_from(self) -> int:
        if isinstance(self.last_trigger, PeriodicalTrigger):
            return self.last_trigger.level
        elif isinstance(self.last_trigger, int):
            return self.last_trigger
        else:
            return 0

    @property
    def current_value(self):
        return int(self.level_from + (self.level - self.level_from)*self.current_progress)

    @property
    def current_progress(self):
        duration = self.incubator.cycle.duration
        if duration == 0:
            duration = 1/60
        progress = (datetime.datetime.now() - self._datetime).total_seconds() / duration / 60
        return min(1, max(0, progress))

    @property
    def has_next(self):
        if self.incubator.recursive:
            return True
        if self.has_next_step:
            return True
        if self.datetime > datetime.datetime.now():
            return True

        return False

    @property
    def has_next_step(self):
        return self._step_datetime is not None

    @property
    def time_left(self):
        now = datetime.datetime.now()
        if self._step_datetime:
            return int((self._step_datetime - now).total_seconds())

        return int((self.datetime - now).total_seconds())

    def check_progress(self):
        next_action = datetime.datetime.now() + datetime.timedelta(seconds=self.FADE_TIME)
        if 0 < self.current_progress < 1 and (self.next_trigger is None or next_action < self.next_trigger.datetime):
            self._step_datetime = next_action
        else:
            self._step_datetime = None

    @property
    def incubator(self) -> Incubator: return self._incubator

    @incubator.setter
    def incubator(self, incubator: Incubator): self._incubator = incubator

    @property
    def datetime(self):
        if self._datetime is None:
            self._datetime = datetime.datetime.combine(
                datetime.datetime.fromisoformat(self.incubator.startTime),
                datetime.time.fromisoformat(self.time)
            )
            self._datetime += datetime.timedelta(days=self.day)

        return self._datetime

    def increment_datetime(self):
        now = datetime.datetime.now()
        if self._datetime < now + datetime.timedelta(seconds=self.FADE_TIME):
            pass
        elif self._datetime < now and self.incubator.recursive:
            while self._datetime < now:
                self._datetime += datetime.timedelta(days=self.incubator.cycle.period)

        self.check_progress()

        if not self.has_next_step and self._datetime < now and self.incubator.recursive:
            self._datetime += datetime.timedelta(days=self.incubator.cycle.period)

    @classmethod
    def of(cls, data: dict, incubator: Incubator):
        self = cls(data)
        for attname, value in data.items():
            if attname == 'id':
                continue
            if hasattr(self, attname):
                setattr(self, attname, value)

        self.incubator = incubator

        return self
