import logging

import spread_knx as knx
from spread_core.bam import Funit, Provider, MultiCommand

""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"""                                  DALI 1                                """
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

DISCOVERY_DELAY = 0.5


class KnxProvider(Provider):
    def_state = None

    def __init__(self, data):
        super(KnxProvider, self).__init__(data)
        self.address = knx.Address(**data['attributes'])
        self._status = self.SUSPICIOUS

    def repair_on_fail(self) -> tuple:
        return (*super(KnxProvider, self).repair_on_fail(), BrightnessLevel,)

    def on_survey(self):
        yield MultiCommand(self.check_valid)

    def request(self, cmd, *args, sig=None, **kwargs):
        for response in self._manager.send_interface.send(cmd, self.address.state):
            self.on_request_answer(response, *args, sig=sig)

    def on_request_answer(self, response, *args, sig=None, **kwargs):
        logging.warning('Unknown response %s' % response)

    def set_valid(self, validity):
        if validity == self._status == self.SUSPICIOUS:
            validity = self.INVALID

        super(KnxProvider, self).set_valid(validity)

    def check_valid(self):
        self.GetBrightnessLevel()

    def get_info(self):
        pass

    def on_update(self, funit_type, value, sig=None, **kwargs):
        if funit_type == BrightnessLevel:
            if value is None:
                self.set_valid(self.SUSPICIOUS)
                if self._status == self.SUSPICIOUS:
                    return
            else:
                self.set_valid(self.VALID)

        super(KnxProvider, self).on_update(funit_type, value, sig)

    def prs2raw(self, p):
        if p == 0:
            return 0
        return min(max(1, round(255 * p / 100)), 255)

    def raw2prs(self, r):
        return None if r is None else min(100, round(100 * r / 255))

    def SetDiscovery(self, value, sig=None):
        pass

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        raise NotImplemented()

    def GetBrightnessLevel(self, sig=None):
        raise NotImplemented()


class KnxDimmer(KnxProvider):
    def_state = knx.commands.Dimming

    def on_request_answer(self, response, *args, sig=None):
        if isinstance(response, knx.commands.Dimming):
            self.on_update(BrightnessLevel, self.raw2prs(response.value), sig=sig)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        self.on_update(BrightnessLevel, value, sig=sig)
        self.request(knx.commands.Dimming.set(self.address, self.prs2raw(value)))

    def GetBrightnessLevel(self, sig=None):
        self.request(knx.commands.Dimming.get(self.address), sig=sig)


class KnxRelay(KnxProvider):
    def_state = knx.commands.Switching

    def on_request_answer(self, response, *args, sig=None):
        if isinstance(response, knx.commands.Switching):
            value = response.value
            funit_type = BrightnessLevel
            if value is not None:
                value = 100 if value else 0
            self.on_update(funit_type, value, sig=sig)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        self.on_update(BrightnessLevel, value, sig=sig)
        self.request(knx.commands.Switching.set(self.address, value > 0), sig=sig)

    def GetBrightnessLevel(self, sig=None):
        self.request(knx.commands.Switching.get(self.address), sig=sig)


class KNXFunit(Funit):
    entity_type = KnxProvider
    easy_get = False
    get_after_set = False
    set_cmd = None
    get_cmd = None
    set_func = None
    get_func = None

    @property
    def entity(self) -> KnxProvider:
        return super(KNXFunit, self).entity

    def set(self):
        if self.set_cmd:
            self.entity.manager.add_command(self.entity, self.set_cmd.set(self.entity.address, self.set_value))
        elif self.set_func:
            self.entity.manager.add_command(
                self.entity,
                self._apply_func(self.entity, self.set_func.__name__)
            )

        if self.get_cmd and self.get_after_set:
            self.get()

    def get(self):
        if self.easy_get:
            self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
        elif self.get_cmd:
            self.entity.manager.add_command(self.entity, self.get_cmd.get(self.entity.address))
        elif self.get_func:
            self.entity.manager.add_command(
                self.entity,
                self._apply_func(self.entity, self.get_func.__name__)
            )


KnxProvider.funit_class = KNXFunit


class BrightnessLevel(KNXFunit):
    set_func = KnxProvider.SetBrightnessLevel
    get_func = KnxProvider.GetBrightnessLevel
    value_range = range(0, 101)


class GroupLevel(BrightnessLevel):
    pass
