import datetime
import logging
import threading

from spread_core.mqtt.spread.variable import Variable
import spread_lom as lom
from spread_core.bam import Funit
from spread_core.bam.providers import Provider
from spread_core.errors.project_errors import CommandError
from spread_core.tools.timer import Timer


class LomProvider(Provider):
    binding = None

    def __init__(self, data):
        super(LomProvider, self).__init__(data)
        if 'devEui' in data['attributes']:
            self.dev_eui = data['attributes']['devEui']
        # костыль
        elif 'address' in data['attributes']:
            self.dev_eui = data['attributes']['address']
        ###
        self._current_profile = None
        self._level = None
        self.timer = None
        self.groups = data['attributes'].get('groups', []) if isinstance(self, LomDimmer) else []

    def not_repaired(self) -> tuple:
        return (*super(LomProvider, self).not_repaired(), Binding, DevEUI, CurrentProfile, Time)

    @property
    def answer_interval(self):
        return 300

    @property
    def req_attempts(self):
        return 2

    @property
    def profile(self):
        return self._current_profile

    def is_dumped(self, key): return key == PersonalProfile

    @property
    def participants(self):
        if self.is_device:
            yield self
        else:
            for d in self.manager.devices.values():
                if isinstance(d, LomDimmer) and d.is_device:
                    if self.is_broadcast or self.dev_eui in d.groups:
                        yield d

    @profile.setter
    def profile(self, profile: lom.profiles.DimmingProfile):
        if profile is None:
            profile = lom.profiles.GlobalGeoProfile.default()

        for _provider in self.participants:
            assert isinstance(_provider, LomDimmer)
            if _provider._current_profile != profile:
                _provider._current_profile = profile
                _provider.on_update(CurrentProfile, profile.serialize())
                _provider.on_timer()

    def on_timer(self):
        profile = self.profile

        self.stop_timer()

        if isinstance(profile, lom.profiles.DimmingProfile):
            pos = None
            if isinstance(profile, lom.profiles.GeoProfile):
                pos = self.manager.coordinates
            items = profile.sorted_items(pos)
            if len(items) > 0:
                if isinstance(profile, lom.profiles.TimedLevelProfile):
                    if profile.start_time.timestamp() + items[0].period <= datetime.datetime.now().timestamp():
                        self.apply_profile(self.get_value(ActiveProfile, 0))
                    else:
                        self.level = items[0].value
                        self.start_timer(items[0].period)
                # elif isinstance(profile, lom.profiles.LumProfile):
                #     if self.get_value(lom.const.ILLUMINATION, -1) == profile.
                #     self.brightness_level
                elif profile.is_timed:
                    self.level = items[-1].value
                    self.start_timer(items[0].time_left(pos))

    def on_lum(self, lum):
        if isinstance(self.profile, lom.profiles.LumProfile):
            last = None
            for item in self.profile.sorted_items():
                if last is None:
                    last = item

                if lum >= item.lum:
                    last = item
                elif last:
                    break

            if last:
                self.level = last.value

    def start_timer(self, time):
        self.stop_timer()
        self.timer = Timer.set(time, self.on_timer)
        # logging.debug(f'Запуск таймера({self.id}) на {datetime.timedelta(seconds=time)}')

    def stop_timer(self):
        if self.timer:
            try:
                self.timer.close()
            except ValueError:
                pass
            finally:
                self.timer = None

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, level: int):
        if level != self._level:
            self._level = level

            if level is None:
                brightness = None
            elif level != 0:
                brightness = max(1, min(100, round(100 * level / 254)))
            else:
                brightness = 0

            self.on_update(BrightnessLevel, brightness)

    @property
    def is_device(self):
        return self.binding == lom.const.DEVICE

    @property
    def is_broadcast(self):
        return self.binding == lom.const.MULTICAST

    @property
    def is_group(self):
        return self.binding == lom.const.GROUP

    def get_bindings(self):
        self.on_update(Binding, self.binding)
        self.on_update(DevEUI, self.dev_eui)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        if funit_type == ActiveProfile:
            if not isinstance(self.profile, (lom.profiles.TimedLevelProfile, lom.profiles.PersonalDimmingProfile)):
                if self.profile is None or self.profile.id != value:
                    self.apply_profile(value)
        elif funit_type == PersonalProfile:
            profile = lom.profiles.PersonalDimmingProfile.deserialize(value)
            self.profile = profile
        elif funit_type == Illumination:
            self.on_lum(value)
        elif funit_type == Time:
            kwargs.setdefault(Variable.TS, datetime.datetime.now())
        # elif funit_type in [Latitude, Longitude, Altitude]:
        elif funit_type == Altitude:
            if self.get_value(funit_type, None) != value:
                if value == 0:
                    # установка координат производится только пр получении Altitude == 0
                    # но значение не запрашивается, т.е. таймер запустится при приёме следующего регулярного пакета
                    self.manager.send_interface.send_fast(
                        lom.commands.SetConfig(
                            self.dev_eui,
                            [Latitude.parameter_id, Latitude(self, Variable(self.manager.coordinates[0])).set_value],
                            [Longitude.parameter_id, Longitude(self, Variable(self.manager.coordinates[1])).set_value],
                            [Altitude.parameter_id, Altitude(self, Variable(self.manager.coordinates[2])).set_value],
                        )
                    )
                else:
                    self.start_timer(1)

        super(LomProvider, self).on_update(funit_type, value, sig=sig, **kwargs)

        # if funit_type == lom.const.EXTRA_FLAGS:
        #     c277 = 10 if self.get_value(lom.const.ANSWER_RAND_INTERVAL) != 10 else 15
        #     self.send(lom.commands.SetConfig(self.dev_eui, [lom.const.ANSWER_RAND_INTERVAL, c277]))

    def apply_profile(self, profile_id):
        if profile_id == 0:
            self.profile = None
        elif profile_id in self.manager.profiles:
            self.profile = self.manager.profiles[profile_id]
        else:
            raise lom.errors.UnknownProfile(profile_id)

    def on_command_sended(self, cmd):
        # if isinstance(cmd, lom.commands.SetConfig):
        #     for key, value in cmd.items.items():
        #         self.changed_confs.setdefault(key)
        #     self._manager.add_info(self, lom.commands.GetConfig(self.dev_eui, list(cmd.items.keys())))
        return
        # if isinstance(cmd, lom.commands.ClearProfile):
        #     if self.get_value(ActiveProfile) in cmd.ids:
        #         self.on_update(ActiveProfile, 0)
        #         if self.profile and self.profile.id in cmd.ids:
        #             self.profile = None
        # elif isinstance(cmd, lom.commands.SetProfile):
        #     if isinstance(cmd.profile, lom.profiles.TimedLevelProfile):
        #         self.profile = cmd.profile
        #     elif isinstance(cmd.profile, lom.profiles.GlobalDimmingProfile):
        #         self.manager.add_profile(cmd.profile)
        #         self.profile = self.manager.profiles[cmd.profile.id]

    def make_reply(self, funit_type, variable, sig):
        if isinstance(sig, list):
            for s in sig:
                super(LomProvider, self).make_reply(funit_type, variable, s)
        else:
            super(LomProvider, self).make_reply(funit_type, variable, sig)

    def on_survey(self):
        yield

    def request(self, cmd, *args, **kwargs):
        sig = getattr(cmd, 'sig', None)

        for sub_cmd in cmd.split(6):
            threading.Event().wait(1)
            for response in self.manager.send_interface.send(sub_cmd,
                                                             timeout=self.answer_interval,
                                                             attempts=self.req_attempts):
                self.on_request_answer(sub_cmd, response, sig=sig)

    def on_request_answer(self, command, response, sig=None, *args):
        if isinstance(response, lom.commands.ConfResponse):
            for p_key, raw_value in response.value.items():
                try:
                    funit = _ConfigItem.of_parameter(p_key)
                    if funit:
                        self.on_update(funit, funit.of_raw(raw_value), sig=sig)
                except BaseException as ex:
                    logging.exception(ex)
        elif isinstance(command, lom.commands.SetConfig):
            for p_key, raw_value in command.items.items():
                try:
                    funit = _ConfigItem.of_parameter(p_key)
                    if funit:
                        if response and response.success:
                            if not self.is_device:
                                _provider = self.manager.devices.get(response.dev_eui, None)
                                if _provider:
                                    _provider.on_update(funit, funit.of_raw(raw_value))
                            else:
                                self.on_update(funit, funit.of_raw(raw_value), sig=sig)
                except BaseException as ex:
                    logging.exception(ex)
        elif isinstance(command, lom.commands.GlobalTimedLevel):
            if response and response.success:
                if not self.is_device:
                    _provider = self.manager.devices.get(response.dev_eui, None)
                    if _provider:
                        _provider.profile = command.profile
                else:
                    self.profile = command.profile
        elif isinstance(command, lom.commands.SetProfile):
            if response and response.success:
                if not self.is_device:
                    _provider = self.manager.devices.get(response.dev_eui, None)
                    if _provider and _provider.get_value(ActiveProfile) == command.profile.id:
                        _provider.profile = command.profile
                elif self.get_value(ActiveProfile) == command.profile.id:
                    self.profile = command.profile
        elif isinstance(command, lom.commands.SetActiveProfile):
            if response and response.success:
                if not self.is_device:
                    _provider = self.manager.devices.get(response.dev_eui, None)
                    if _provider:
                        _provider.on_update(ActiveProfile, command.id)
                else:
                    self.on_update(ActiveProfile, command.id, sig=sig)

    def on_exit(self):
        super(LomProvider, self).on_exit()
        self.level = None
        self.stop_timer()

    def get_info(self):
        return
        # if self.is_device:
        #     self._manager.add_info(self, lom.commands.GetConfig(self.dev_eui, [funit.parameter_id for funit in (
        #         ActiveProfile,
        #         AnswerInterval,
        #         SendInterval,
        #         Latitude,
        #         Longitude,
        #         Altitude,
        #
        #         RejoinInterval,
        #         TimeSyncInterval,
        #         NWCheckInterval,
        #         NWCheckAttempts,
        #         ConfirmEnabled,
        #         AngleThreshold,
        #
        #         DaliScanInterval,
        #         McAddressGroup1,
        #         Group1Enabled,
        #         Group2Enabled,
        #         Group3Enabled,
        #         Group4Enabled,
        #
        #         AccelSensitivity,
        #     ) if self.get_value(funit, None) is None]))

    def get_value(self, funit_type, def_value=None):
        if funit_type in self.values:
            return self.values[funit_type]
        if def_value is not None:
            return def_value
        return None

    def set_current_profile(self, value: (int, dict), sig=None):
        if isinstance(value, dict):
            """Personal profile"""
            raise CommandError.of_exec('Персональный профиль не реализлован!')
        else:
            """Global profile"""
            try:
                profile = self.manager.profiles[int(value)]
            except KeyError:
                raise CommandError.of_exec('Идентификатор {} отсутствует в списке профилей!'.format(value))

        if not isinstance(profile, lom.profiles.TimedLevelProfile):
            self.manager.add_command(self, lom.commands.ClearAllProfiles(self.dev_eui))
        self.manager.add_command(self, lom.commands.SetProfile.of(self.dev_eui, profile, mc=not self.is_device))
        self.manager.add_command(self, lom.commands.SetConfig(self.dev_eui, [ActiveProfile.parameter_id, profile.id]))


class LomGroup(LomProvider):
    binding = lom.const.GROUP

    @property
    def answer_interval(self):
        try:
            return max(*(
                p.get_value(AnswerInterval, 300)
                for p in self.manager.devices.values()
                if isinstance(p, LomDimmer) and p.is_device and self.dev_eui in p.groups
            ))
        except TypeError:
            return super(LomGroup, self).answer_interval

    @property
    def req_attempts(self):
        result = 0
        for p in self.manager.devices.values():
            if isinstance(p, LomDimmer) and p.is_device and self.dev_eui in p.groups:
                result += 1

        return max(1, result)


class LomDimmer(LomProvider):
    binding = lom.const.DEVICE

    @property
    def answer_interval(self):
        return self.get_value(AnswerInterval, super(LomDimmer, self).answer_interval)


class _LomFunit(Funit):
    entity_type = LomProvider
    set_cmd = None
    get_cmd = None
    set_func = None
    get_func = None
    easy_get = False
    value_type = int

    @property
    def entity(self) -> entity_type:
        return super(_LomFunit, self).entity

    def set(self):
        if self.set_cmd:
            self._apply_command(self.set_cmd, self.entity.dev_eui, self.set_value)
            if not self.entity.is_device:
                self.entity.on_update(self.__class__, self.of_raw(self.set_value), sig=self.signature)
        elif self.set_func:
            self.entity.manager.add_command(self.entity, self._apply_func(self.entity, self.set_func.__name__))
        else:
            super(_LomFunit, self).set()

    def get(self):
        self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)

    def _apply_command(self, cmd_class, *args, custom_signatures=None, **kwargs):
        cmd = cmd_class(*args, mc=not self.entity.is_device, **kwargs)
        setattr(cmd, 'sig', custom_signatures if custom_signatures else [self.signature])

        self.entity.manager.add_command(self.entity, cmd)


LomProvider.funit_class = _LomFunit


class Binding(_LomFunit):
    easy_get = True


class DevEUI(_LomFunit):
    # hex(int.from_bytes(int(val).to_bytes(8, 'little'), 'big'))[2:].rjust(16, '0').upper()
    easy_get = True


class Serial(_LomFunit):
    easy_get = True


class Illumination(_LomFunit):
    easy_get = True


class BrightnessLevel(_LomFunit):
    easy_get = True
    set_cmd = lom.commands.GlobalTimedLevel
    value_range = range(0, 101)

    @property
    def set_value(self):
        level = super(BrightnessLevel, self).set_value
        if level != 0:
            level = min(max(1, int(254 * level / 100)), 254)

        return lom.profiles.GlobalTimedLevel.of_level(
            lom.profiles.profile_items.TimedLevel(
                level,
                int(self.params.get('period', self.entity.manager.def_period))
            )
        )


class GroupLevel(BrightnessLevel):
    pass


class CurrentProfile(_LomFunit):
    easy_get = True
    set_func = LomProvider.set_current_profile


class PersonalProfile(_LomFunit):
    easy_get = True


class GetQueueItem:
    cmd = None

    TIMEOUT = 1
    timer = None
    queue = {}

    def __init__(self, funit: _LomFunit):
        self.funit = funit

    def __str__(self): return '%s' % self.funit.entity.dev_eui

    def __repr__(self): return str(self)

    @staticmethod
    def append_parameter(funit):
        return funit.parameter_id

    @classmethod
    def on_timer(cls, funits: [_LomFunit]):
        try:
            if funits[0].entity.dev_eui in cls.queue:
                cls.queue[funits[0].entity.dev_eui] += funits
            else:
                parameters = []
                signatures = []
                for funit in funits:
                    if funit.parameter_id not in parameters:
                        parameters.append(cls.append_parameter(funit))

                    signatures.append(funit.signature)

                funits[0]._apply_command(cls.cmd, funits[0].entity.dev_eui, parameters, custom_signatures=signatures)

        finally:
            if cls.queue:
                cls.start_timer(cls.queue.pop(list(cls.queue.keys())[0]))
            elif cls.timer:
                try:
                    cls.timer.close()
                except ValueError:
                    pass
                finally:
                    cls.timer = None

    @classmethod
    def start_timer(cls, funits: [_LomFunit]):
        cls.timer = Timer.set(cls.TIMEOUT, cls.on_timer, funits)

    @classmethod
    def add(cls, funit):
        if cls.timer is None:
            cls.start_timer([funit])
        else:
            cls.queue.setdefault(funit.entity.dev_eui, [])
            cls.queue[funit.entity.dev_eui].append(funit)


class GetQueue(GetQueueItem):
    cmd = lom.commands.GetConfig


class SetQueue(GetQueueItem):
    cmd = lom.commands.SetConfig

    @staticmethod
    def append_parameter(funit):
        return [funit.parameter_id, funit.set_value]


class _ConfigItem(_LomFunit):
    parameter_id = None

    @classmethod
    def of_parameter(cls, parameter_id):
        for s in cls.__subclasses__():
            if issubclass(s, _ConfigItem):
                if s.__name__.startswith('_'):
                    r = s.of_parameter(parameter_id)
                    if r:
                        return r
                elif s.parameter_id == parameter_id:
                    return s

        return None

    def get(self):
        GetQueue.add(self)
        # self._apply_command(lom.commands.GetConfig, self.entity.dev_eui, [self.parameter_id])

    def set(self):
        SetQueue.add(self)

        if not self.entity.is_device:
            self.entity.on_update(self.__class__, self.of_raw(self.set_value), sig=self.signature)
        # self._apply_command(lom.commands.SetConfig, self.entity.dev_eui, [self.parameter_id, int(self.set_value)])


class RejoinInterval(_ConfigItem):
    parameter_id = 265


class TimeSyncInterval(_ConfigItem):
    parameter_id = 266


class SendInterval(_ConfigItem):
    parameter_id = 267


class NWCheckInterval(_ConfigItem):
    parameter_id = 268


class NWCheckAttempts(_ConfigItem):
    parameter_id = 269


class ConfirmEnabled(_ConfigItem):
    parameter_id = 270
    index_of = [False, True]


class Latitude(_ConfigItem):
    parameter_id = 271
    multiplier = 100000


class Longitude(_ConfigItem):
    parameter_id = 272
    multiplier = 100000


class Altitude(_ConfigItem):
    parameter_id = 273


class AngleThreshold(_ConfigItem):
    parameter_id = 274


class AccelSensitivity(_ConfigItem):
    parameter_id = 275
    index_of = ['Off', 'Low', 'Middle', 'High']


class DaliScanInterval(_ConfigItem):
    parameter_id = 276


class AnswerInterval(_ConfigItem):
    parameter_id = 277


class ActiveProfile(_ConfigItem):
    parameter_id = 278
    set_cmd = lom.commands.SetActiveProfile


class _GroupEnabled(_ConfigItem):
    index_of = [False, True]


class _McAddressGroup(_ConfigItem):
    value_range = range(0, 0xffffffff)
    hex_count = 4


class Group1Enabled     (_GroupEnabled):    parameter_id = 264
class McAddressGroup1   (_McAddressGroup):  parameter_id = 7
class Group2Enabled     (_GroupEnabled):    parameter_id = 281
class McAddressGroup2   (_McAddressGroup):  parameter_id = 9
class Group3Enabled     (_GroupEnabled):    parameter_id = 282
class McAddressGroup3   (_McAddressGroup):  parameter_id = 10
class Group4Enabled     (_GroupEnabled):    parameter_id = 283
class McAddressGroup4   (_McAddressGroup):  parameter_id = 11


class Firmware(_ConfigItem):
    parameter_id = 3

    @classmethod
    def of_raw(cls, raw_value):
        return f'{raw_value >> 4}.{raw_value & 0xF}'


class UserPass(_ConfigItem):
    parameter_id = 6

    @classmethod
    def of_raw(cls, raw_value):
        return raw_value.to_bytes(4, 'little').decode()


class Time(_LomFunit):
    pass
