import json
import logging
import queue
import threading

import spread_core.mqtt as mqtt
from spread_core.bam import generator as generator
from spread_core.bam.entities import Entity
from spread_core.bam.multi_command import MultiCommand
from spread_core.errors.project_errors import ProjectError, JocketError, InvalidDevice, InitError, ClassifierError, \
    CommandError
from spread_core.protocols.dali.exceptions import CommandGenerateError
from spread_core.tools.base_interface import AbstractInterfaceClass
from spread_core.tools.corutunes import coroutine
from spread_core.tools.manager_interface import ManagerOfBroker
from spread_core.tools.settings import config, POLL_RATE

DEF_POLL_RATE = 0


class Manager(Entity, ManagerOfBroker):
    STATUS_STOPPED = 'Stopped'
    STATUS_RUNNING = 'Running'
    STATUS_INFO = 'Info'
    STATUS_SURVEY = 'Survey'

    def __init__(self, p_id,  m_id):
        Entity.__init__(self, m_id)
        self._poll_rate_pr = 0
        self._stopped = False
        self._send_interface = None
        self._providers = dict()
        self._addresses = dict()
        self._project_id = p_id
        self.poll_rate = 0
        self.ready_timer = None
        self.bus = None
        self.name = ''
        self.stack = queue.Queue()

        self.update_config()

    def _sorted_info(self):
        return self._providers

    @property
    def send_interface(self) -> AbstractInterfaceClass:
        return self._send_interface

    @send_interface.setter
    def send_interface(self, interface):
        if interface:
            interface.set_event_handler(self.event_handler())
            interface.update_config()
            interface.set_external_cmd_handler(self.external_command_handler())
            interface.set_status_handler(self.bus_status_handler())
            interface.set_bus_params(self.bus)
            interface.start()

        if self._send_interface:
            self._send_interface.set_stopped()
            del self._send_interface

        self._send_interface = interface

    @property
    def project_id(self):
        return self._project_id

    @property
    def providers(self):
        """returns copy of providers"""
        return self._providers.copy()

    def get_provider(self, provider_id):
        """returns provider by id"""
        return self._providers.get(provider_id, None)

    def providers_by_address(self, dali_type, short_address) -> list:
        """returns providers by address"""

        provider_ids = self._addresses.get(short_address, [])
        for provider_id in provider_ids:
            try:
                _provider = self._providers[provider_id]
                if _provider._dali_type == dali_type:
                    yield _provider
            except KeyError:
                continue

    @property
    def addresses(self):
        """returns copy of addresses"""
        return self._addresses.copy()

    def publish(self, topic, data, retain=False, **kwargs):
        ManagerOfBroker.publish(self, topic, data, retain=retain, **kwargs)

    def parse(self, data):
        if data['type'] != self.__class__.__name__:
            raise InitError(f'Manager type required as "{self.__class__.__name__}" but set "{data["type"]}"')
        self.bus = data['attributes'].get('bus', None)
        self.name = data.get('name', f'id:{self.id}')
        self._poll_rate_pr = data.get('attributes', {}).get(POLL_RATE, DEF_POLL_RATE)
        self.update_poll_rate()
        self.SetStatus(None, self.STATUS_RUNNING)

    def set_broker(self, mqttc):
        ManagerOfBroker.__init__(self, mqttc)

    def set_interface(self):
        raise NotImplementedError()

    def update_config(self):
        self.update_poll_rate()
        if self._send_interface:
            self._send_interface.update_config()

    def update_poll_rate(self):
        pl = config.get(POLL_RATE, None)
        if pl is None:
            pl = self._poll_rate_pr
        self.poll_rate = pl / 1000 if pl else None

    def start(self):
        self.subscribe(mqtt.project.TopicProject.managers(self.project_id))

    def on_state(self, topic: mqtt.spread.topic.SpreadTopic = None, variable: mqtt.spread.variable.Variable = None):
        if topic and variable:
            if isinstance(topic.entity_address, mqtt.spread.address.ProviderAddress):
                _provider = self.get_provider(topic.entity_address.id)
                if _provider:
                    if variable.value is not None:
                        if topic.entity_address.funit_type not in _provider.not_repaired():
                            if not _provider.is_device or not variable.preliminary:
                                _provider._invalid_values[topic.entity_address.funit_type] = variable.value

        self.set_ready_timer()

    def stop_ready_timer(self):
        if isinstance(self.ready_timer, threading.Timer):
            self.ready_timer.cancel()
            del self.ready_timer
            self.ready_timer = None

    def set_ready_timer(self):
        self.stop_ready_timer()

        self.ready_timer = threading.Timer(1, self.on_ready)
        self.ready_timer.setName('on_ready Timer')
        self.ready_timer.start()

    def on_project(self, topic, data):
        self.unsubscribe(topic)

        if isinstance(data, (bytes, bytearray)):
            data = data.decode()

        payload = json.loads(data)
        if mqtt.project.TopicProject.MANAGERS in payload:
            for item in payload[mqtt.project.TopicProject.MANAGERS]:
                if item['id'] == self.id:
                    try:
                        self.parse(item)
                    except BaseException as ex:
                        raise InitError(str(ex))
                    self.subscribe(mqtt.project.TopicProject.providers(self.project_id), True)
                    return
            raise ProjectError("В файле 'managers.json' отсутствует менеджер с ID={}".format(self.id))
        elif mqtt.project.TopicProject.PROVIDERS in payload:
            for item in payload[mqtt.project.TopicProject.PROVIDERS]:
                if item['managerID'] == self.id:
                    try:
                        self.add_provider(item)
                    except BaseException as ex:
                        raise InitError(f'{ex} for {item}')
            logging.info('Project({}) loaded'.format(self.project_id))
            self.on_project_ready()

    def on_project_ready(self):
        self.on_state()
        self.subscribe(mqtt.spread.topic.State(self.get_address('+/+/#')))

    def add_provider(self, data):
        provider = generator.generate_provider(data)
        provider.set_manager(self)
        self._providers[provider.id] = provider

    def on_ready(self):
        self.stop_ready_timer()
        self.unsubscribe(mqtt.spread.topic.State(self.get_address('+/+/#')))
        self.subscribe(mqtt.spread.topic.Set(self.get_address(None)))
        self.subscribe(mqtt.spread.topic.Renew(self.get_address(None)))

        self.set_interface()
        self.info()

    def on_exit(self):
        super(Manager, self).on_exit()
        self._stopped = True

        self.stop_ready_timer()

        if self.poll_rate is None and self.stack.qsize() == 0:
            self.stack.put((None, None))

        if self._send_interface:
            self._send_interface.set_stopped()

        if self.mqttc.connected:
            for p_id in self._providers:
                provider = self._providers[p_id]
                provider.on_exit()

            self.SetStatus(None, self.STATUS_STOPPED)

    def sub_survey(self):
        has_survey_item = True
        while has_survey_item:
            has_survey_item = False
            for provider in self.providers.values():
                if provider.surveyable:
                    has_survey_item = True
                    for cmd in provider.on_survey():
                        yield provider, cmd

    @coroutine
    def survey(self, sub):
        yield from sub

    @coroutine
    def external_command_handler(self):
        while True:
            try:
                ext = yield
                try:
                    self.on_external_command(*ext)
                except BaseException as ex:
                    logging.debug(f'EXTERNAL FRAME ERROR[{ext}]')
                    logging.exception(ex)
            except StopIteration:
                break

    def on_external_command(self, *args):
        raise NotImplementedError()

    @coroutine
    def event_handler(self):
        while True:
            try:
                event = yield
                try:
                    self.on_event(event)
                except BaseException as ex:
                    logging.debug(f'EVENT FRAME ERROR[{event}]')
                    logging.exception(ex)
            except StopIteration:
                break

    def on_event(self, event):
        raise NotImplementedError()

    @coroutine
    def bus_status_handler(self):
        while True:
            ready = yield
            self.on_bus_status(ready)

    def on_bus_status(self, ready):
        raise NotImplementedError()

    def info(self):
        while not (self._stopped or self.send_interface.ready):
            threading.Event().wait(1)

        if self._stopped:
            return

        self.SetStatus(None, self.STATUS_INFO)
        for p_id in self._sorted_info():
            if self._stopped:
                return
            provider = self._providers[p_id]
            if provider.infoble:
                try:
                    provider.get_bindings()
                    provider.check_valid()
                    provider.commit_states(provider.is_valid)
                    if provider.is_valid:
                        provider.get_info()
                        provider.override_info()
                except InvalidDevice as ex:
                    logging.warning(ex)
                    provider.set_valid(provider.INVALID)

        self.SetStatus(None, self.STATUS_SURVEY)
        self.do_next()

    def do_next(self):
        try:
            survey = self.survey(self.sub_survey())
        except StopIteration:
            survey = None

        while not self._stopped:
            try:
                provider, cmds = self.stack.get(timeout=self.poll_rate)
                if not isinstance(cmds, (list, tuple)):
                    cmds = (cmds,)

                for cmd in cmds:
                    if cmd:
                        if isinstance(cmd, MultiCommand):
                            cmd.call()
                        elif provider:
                            provider.request(cmd)
                        else:
                            self.send_interface.send(cmd)

            except queue.Empty:
                if survey:
                    self.stack.put(next(survey))
            except BaseException as ex:
                logging.exception(ex)
            else:
                self.stack.task_done()

    def add_command(self, provider, *cmds):
        if not self._stopped:
            if cmds:
                try:
                    self.stack.put((provider, cmds))
                except BaseException as ex:
                    logging.exception(ex)
            else:
                logging.warning('Empty command')

    def add_info(self, provider, cmd):
        if not self._stopped:
            self.stack.put((provider, cmd))

    def on_command(self, topic: mqtt.spread.topic.Set, variable: mqtt.spread.variable.Variable):
        self.log_command(topic, variable)
        if self._stopped:
            return

        try:
            if not self._send_interface.ready:
                raise JocketError('Bus is not ready')
            elif topic.entity_address.type == self.__class__.__name__ and topic.entity_address.id == self.id:
                super(Manager, self).on_command(topic, variable)
            elif topic.entity_address.id in self._providers:
                provider = self._providers[topic.entity_address.id]
                accessed = provider.is_valid
                if not accessed:
                    if isinstance(topic, mqtt.spread.topic.Set):
                        accessed = topic.entity_address.funit_type == 'Binding'
                    elif isinstance(topic, mqtt.spread.topic.Renew):
                        accessed = topic.entity_address.funit_type in ('Binding', 'BindingDevice', 'BindingGroup')

                if accessed:
                    provider.on_command(topic, variable)
                else:
                    raise CommandGenerateError('device is invalid!')
            else:
                raise JocketError('объект с адресом {} не найден!'.format(topic.entity_address.id))
        except (JocketError, CommandGenerateError, ClassifierError, InvalidDevice, CommandError) as ex:
            logging.warning(ex)
            if variable.session_id:
                to_topic = mqtt.spread.topic.Reply(self.project_id, variable.session_id)
                self.publish(to_topic, mqtt.spread.variable.Variable(None, error=str(ex)))

    def log_command(self, topic: mqtt.spread.topic.SpreadTopic, variable: mqtt.spread.variable.Variable):
        logging.debug(
            f'COMMAND({variable.session_id if variable.session_id else ""}): ' +
            f'{topic.__class__.__name__} {topic.entity_address.type}({topic.entity_address.id})'
            f'.{topic.entity_address.funit_type}' +
            (f'={variable.value}' if isinstance(topic, mqtt.spread.topic.Set) else '') +
            (f' params: {variable.params}' if variable.params else '') +
            f' queue: {self.stack.qsize()}'
        )

    def SetStatus(self, value, sig=None):
        if self.send_interface:
            self.on_update('Status', value)

    def get_funit(self, funit_type):
        if self.__class__.__name__ in mqtt.classifier:
            if funit_type in mqtt.classifier[self.__class__.__name__]:
                return mqtt.classifier[self.__class__.__name__][funit_type]
        return None

    def get_address(self, funit_type):
        if funit_type is None: funit_type = '#'
        return mqtt.spread.address.ManagerAddress.of(self, funit_type)
