from spread_core.bam import engineries
from spread_core.bam.engineries import light
from spread_core.bam.entities import Entity, Funit
from spread_core.mqtt.spread import address
from spread_core.mqtt.spread.topic import State
from spread_core.mqtt.spread.variable import Variable
from spread_core.tools import settings


class Subginery(Entity):
    def __init__(self, project_id, data):
        super().__init__(data['id'])
        self._project_id = project_id
        self._location_id = data['locationID']
        self._engineries = dict.fromkeys(data['engineries'])

    def get_address(self, funit_type):
        return address.SubgineryAddress(self.project_id, self.id, self.__class__.__name__, funit_type)

    @classmethod
    def get_address_broadcast(cls, p_id, funit_type='+'):
        return address.SubgineryAddress(p_id, '+', cls.__name__, funit_type)

    @property
    def project_id(self):
        return self._project_id

    @property
    def engineries(self): return self._engineries.copy()

    @property
    def location_id(self): return self._location_id

    def has_enginery(self, e_id): return e_id in self._engineries
    def set_enginery(self, enginery): self._engineries[enginery.id] = enginery


class Lighting(Subginery):
    def __init__(self, project_id, data):
        super().__init__(project_id, data)

        self.power_level = 0
        self._scenes = None
        self.bad_is_on = []
        self.bad_is_off = []
        self.bad_scenes = {}
        self.bad_presence_on = []
        self.bad_presence_off = []
        self.bad_lum_on = []
        self.bad_lum_off = []
        self.adapter = None

    def set_adapter(self, adapter):
        self.adapter = adapter

    def publish_state(self, funit_type, value, delay=0):
        if self.get_value(funit_type) != value:
            self.set_value(funit_type, value)
            self.adapter.publish(State(self.get_address(funit_type)), Variable(value), retain=True, delay=delay)

    def on_sensor_on_off(self, eng, is_on):
        if isinstance(eng, engineries.PresenceSensor):
            on_class_id = PresenceSensorsOn
            on_stack = self.bad_presence_on
            off_class_id = PresenceSensorsOff
            off_stack = self.bad_presence_off
        elif isinstance(eng, engineries.LightSensor):
            on_class_id = LightSensorsOn
            on_stack = self.bad_lum_on
            off_class_id = LightSensorsOff
            off_stack = self.bad_lum_off
        else:
            return

        if is_on is None:
            if eng.id in on_stack:
                on_stack.remove(eng.id)

            if eng.id in off_stack:
                off_stack.remove(eng.id)

        elif is_on:
            if eng.id not in off_stack:
                off_stack.append(eng.id)

            if eng.id in on_stack:
                on_stack.remove(eng.id)

        else:
            if eng.id not in on_stack:
                on_stack.append(eng.id)

            if eng.id in off_stack:
                off_stack.remove(eng.id)

        self.publish_state(on_class_id.funit_type, len(on_stack) == 0, delay=on_class_id.delay)
        self.publish_state(off_class_id.funit_type, len(off_stack) == 0, delay=on_class_id.delay)

    def on_level_changed(self, eng, level, preliminary):
        if level is None:
            if eng.id in self.bad_is_on:
                self.bad_is_on.remove(eng.id)

            if eng.id in self.bad_is_off:
                self.bad_is_off.remove(eng.id)
        elif level == 0:
            if eng.id not in self.bad_is_on:
                self.bad_is_on.append(eng.id)

            if eng.id in self.bad_is_off:
                self.bad_is_off.remove(eng.id)
        else:
            if eng.id not in self.bad_is_off:
                self.bad_is_off.append(eng.id)

            if eng.id in self.bad_is_on:
                self.bad_is_on.remove(eng.id)

        self.publish_state(On.funit_type, len(self.bad_is_on) == 0, delay=On.delay)
        self.publish_state(Off.funit_type, len(self.bad_is_off) == 0, delay=Off.delay)

        for scene_id in self.scenes:
            if eng.id in self.scenes[scene_id]:
                self.bad_scenes.setdefault(scene_id, [])

                if level is None or self.scenes[scene_id][eng.id] == level:
                    if eng.id in self.bad_scenes[scene_id]:
                        self.bad_scenes[scene_id].remove(eng.id)
                elif eng.id not in self.bad_scenes[scene_id]:
                    self.bad_scenes[scene_id].append(eng.id)

                self.publish_state(SceneOn.funit_type_of_scene(scene_id), len(self.bad_scenes[scene_id]) == 0, delay=SceneOn.delay)

        if preliminary is False:
            self.power_level -= eng.get_value(light.Power, 0)
            self.power_level += int(eng.power * eng.get_value(light.BrightnessLevel, 0) / 100)
            self.publish_state(light.Power, self.power_level, delay=light.Power.delay)

    def set_scene(self, _scene_id, _scene):
        self._scenes[_scene_id] = _scene
        self.bad_scenes.pop(_scene_id, None)
        self.publish_state(SceneOn.funit_type_of_scene(_scene_id), True)
        settings.dump.set_dump(self, f'scene_{_scene_id}', _scene)

    @property
    def scenes(self):
        if self._scenes is None:
            dump = settings.dump.get_dump_entity(self)
            self._scenes = {
                scene_id: {
                    int(e_id): level for e_id, level in dump[f'scene_{scene_id}'].items()
                } for scene_id in range(0, 99) if f'scene_{scene_id}' in dump
            }

        return self._scenes


class Incubator(Subginery):
    pass


class _SubgineriesFunit(Funit):
    entity_type = Lighting
    target_command = light.On
    target_class = engineries.Enginery

    def set(self):
        for e in self.entity.engineries.values():
            if isinstance(e, self.target_class):
                e.publish_provider_command(self.target_command, self.set_value, self.params, sig=self.signature)


Lighting.funit_class = _SubgineriesFunit


class On(_SubgineriesFunit):
    delay = 0
    target_command = light.BrightnessLevel
    target_class = engineries.Light
    set_value = 100


class Off(_SubgineriesFunit):
    delay = 0
    target_command = light.BrightnessLevel
    target_class = engineries.Light
    set_value = 0


class BrightnessLevel(_SubgineriesFunit):
    value_range = range(0, 101)
    target_command = light.BrightnessLevel
    target_class = engineries.Light


class LightSensorsOn(_SubgineriesFunit):
    delay = 0
    target_class = engineries.LightSensor


class LightSensorsOff(_SubgineriesFunit):
    delay = 0
    target_class = engineries.LightSensor
    set_value = False


class PresenceSensorsOn(_SubgineriesFunit):
    delay = 0
    target_class = engineries.PresenceSensor


class PresenceSensorsOff(_SubgineriesFunit):
    delay = 0
    target_class = engineries.PresenceSensor
    set_value = False


class SaveScene(_SubgineriesFunit):
    def set(self):
        _new_scene = dict()
        for _e_id, _enginery in self.entity.engineries.items():
            if isinstance(_enginery, engineries.Light):
                _value = _enginery.get_value(light.BrightnessLevel)
                if _value is not None:
                    _new_scene[_e_id] = _value
        self.entity.set_scene(self.set_value, _new_scene)


class SceneOn(_SubgineriesFunit):
    delay = 0

    @staticmethod
    def funit_type_of_scene(scene_id): return f'Scene{scene_id}On'

    def set(self):
        for e_id, level in self.entity.scenes[self.set_value].items():
            e = self.entity.engineries.get(e_id, None)
            if isinstance(e, self.target_class):
                e.publish_provider_command(light.BrightnessLevel, level, sig=self.signature)


Lighting.settable_units = [
    On, Off, BrightnessLevel,
    LightSensorsOn, LightSensorsOff,
    PresenceSensorsOn, PresenceSensorsOff,
    SaveScene, SceneOn
]