import json
import logging
from datetime import datetime

import spread_core.mqtt as mqtt
from spread_core.bam import engineries, subgineries
from spread_core.bam.engineries import light, sensor
from spread_core.errors.project_errors import ClassifierError, TopicError
from spread_core.tools import settings
from spread_core.tools.service_launcher import Launcher

onId = 'onId'
offId = 'offId'
isOnId = 'isOnId'
isOffId = 'isOffId'
setLevelId = 'setLevelId'
levelId = 'levelId'
groupOnId = 'groupOnId'
groupOffId = 'groupOffId'
groupSetLevelId = 'groupSetLevelId'
historyRequestId = 'historyRequestId'
historyResponseId = 'historyResponseId'
powerLevelId = 'powerLevelId'
setColorId = 'setColorId'
colorId = 'colorId'
saveScene1Id = 'saveScene1Id'
loadScene1Id = 'loadScene1Id'
isMatchScene1Id = 'isMatchScene1Id'
saveScene2Id = 'saveScene2Id'
loadScene2Id = 'loadScene2Id'
isMatchScene2Id = 'isMatchScene2Id'
lightSensorsOnId = 'lightSensorsOnId'
lightSensorsOffId = 'lightSensorsOffId'
isLightSensorsOnId = 'isLightSensorsOnId'
isLightSensorsOffId = 'isLightSensorsOffId'
presenceSensorsOnId = 'presenceSensorsOnId'
presenceSensorsOffId = 'presenceSensorsOffId'
isPresenceSensorsOnId = 'isPresenceSensorsOnId'
isPresenceSensorsOffId = 'isPresenceSensorsOffId'
setTemperatureId = 'setTemperatureId'


class Frontier(Launcher):
    _def_start_delay = 0
    _dumped = False
    funit_ids = {
        light.DimmingLight.__name__: [
            onId, offId, isOnId, setLevelId, levelId, groupOnId, groupOffId, groupSetLevelId, powerLevelId,
            historyRequestId, historyResponseId
        ],
        light.SwitchingLight.__name__: [
            onId, offId, isOnId, groupOnId, groupOffId, powerLevelId, historyRequestId, historyResponseId
        ],
        light.TunableWhiteLight.__name__: [
            onId, offId, isOnId, setLevelId, levelId, setColorId, colorId, groupOnId, groupOffId, groupSetLevelId,
            powerLevelId, historyRequestId, historyResponseId
        ],
        sensor.PresenceSensor.__name__: [
            onId, offId, isOnId
        ],
        sensor.LightSensor.__name__: [
            onId, offId, isOnId
        ],
        subgineries.Lighting.__name__: [
            onId, offId, isOnId, isOffId, saveScene1Id, loadScene1Id, isMatchScene1Id, saveScene2Id, loadScene2Id,
            isMatchScene2Id, powerLevelId, historyRequestId, historyResponseId, lightSensorsOnId, lightSensorsOffId,
            isLightSensorsOnId, isLightSensorsOffId, presenceSensorsOnId, presenceSensorsOffId, isPresenceSensorsOnId,
            isPresenceSensorsOffId, setTemperatureId
        ],
    }

    def __init__(self):
        self._manager = self
        self.server_id = None
        self.signatures = {}
        super(Frontier, self).__init__()

    def update_config(self, sig=None, frame=None):
        try:
            self.server_id = settings.config[settings.SERVER_ID]
        except:
            logging.error('server id is not set!')
            self.exit_handler(None, None)

    def start(self):
        try:
            self.subscribe(mqtt.tros3.topic.TopicCommandTros3(self.project_id))
            self.subscribe(mqtt.jocket.topic.TopicCommand(self.project_id))
            self.subscribe(mqtt.spread.topic.State(mqtt.spread.address.BroadcastAddress(self.project_id)))
            self.subscribe(mqtt.spread.topic.Reply(self.project_id, '+'))
        except BaseException as ex:
            logging.exception(ex)

    def on_message(self, mosq, obj, msg):
        try:
            topic = mqtt.of(msg.topic)

            if isinstance(topic, mqtt.tros3.topic.TopicCommandTros3):
                self.on_tros3(topic, msg.payload)
            elif isinstance(topic, mqtt.jocket.topic.TopicCommand):
                self.on_jocket(topic, msg.payload)
            elif isinstance(topic, mqtt.spread.topic.State):
                self.on_spread(topic, msg.payload)
            elif isinstance(topic, mqtt.spread.topic.Reply):
                self.on_reply(topic, msg.payload)
            else:
                logging.debug(f'[NO MOVE] {topic}')
        except TopicError as ex:
            logging.warning(ex)
        except BaseException as ex:
            logging.exception(ex)

    def on_jocket(self, _topic, data):
        _var = mqtt.jocket.variables.VariableJocket(json.loads(data.decode()))
        if _topic.session_id and _var.key:
            self.signatures[f'{_topic.session_id}|{_var.key}'] = (_var.id, _var.cl)
        _address = _topic.entity_addr
        if isinstance(_address, mqtt.jocket.address.EngineryAddress):
            if _topic.entity_addr.type == engineries.LightingArea.__name__:
                address = mqtt.spread.address.EngineryAddress(_topic.p_id, _topic.entity_addr.type, _address.entity_id, _address.funit_id)
            else:
                address = mqtt.spread.address.SubgineryAddress(_topic.p_id, _address.entity_id, _topic.entity_addr.type, _address.funit_id)

        elif isinstance(_address, mqtt.jocket.address.ProviderAddress):
            address = mqtt.spread.address.ProviderAddress(self.project_id, _address.manager_type, _address.manager_id,
                                                          _address.type, _address.entity_id, _address.funit_type)
        elif isinstance(_address, mqtt.jocket.address.ManagerAddress):
            address = mqtt.spread.address.ManagerAddress(self.project_id, _address.type, _address.entity_id, _address.funit_type)
        else:
            logging.debug(f'Unsupported address ({_address.__class__.__name__})')
            return

        if isinstance(_topic, mqtt.jocket.topic.TopicState):
            topic = mqtt.spread.topic.State(address)
        elif isinstance(_topic, mqtt.jocket.topic.TopicCommand):
            if _var.action == 'set' or _var.action == 'reset':
                topic = mqtt.spread.topic.Set(address)
            else:
                topic = mqtt.spread.topic.Renew(address)
        else:
            logging.debug(f'Unsupported topic ({_topic.__class__.__name__})')
            return

        var = mqtt.spread.variable.Variable(_var.value, key=_var.key, session=_topic.session_id, timestamp=datetime.fromisoformat(str(_var.timeStamp)))
        self.publish(topic, var, retain=isinstance(_topic, mqtt.jocket.topic.TopicState))

    def on_reply(self, _topic: mqtt.spread.topic.Reply, data):
        _var = mqtt.spread.variable.Variable.of(data)
        if _var.key:
            var_id, var_class = self.signatures.pop(f'{_topic.session_id}|{_var.key}', (None, None))
            var = mqtt.jocket.variables.VariableJocket.create_data(var_id, var_class, 'state', _var.value, key=_var.key, invalid=_var.invalid)

            if var_class == 32090001:
                var.data['data'] = var.value
            self.publish(mqtt.jocket.topic.TopicReply(self.project_id, _topic.session_id), var, retain=False)

    def on_spread(self, _topic: mqtt.spread.topic.State, data):
        _var = mqtt.spread.variable.Variable.of(data)
        _address = _topic.entity_address
        if isinstance(_topic.entity_address, (mqtt.spread.address.EngineryAddress, mqtt.spread.address.SubgineryAddress)) \
                and _topic.entity_address.type != engineries.LightingArea.__name__:
            value = _var.value
            if _address.funit_type == light.On:
                funit_id = isOnId
            elif _address.funit_type == light.BrightnessLevel:
                funit_id = levelId
            elif _address.funit_type == light.Power:
                funit_id = powerLevelId
            elif _address.funit_type == light.Temperature:
                funit_id = colorId
            elif _address.funit_type == subgineries.SceneOn.funit_type_of_scene(1):
                funit_id = isMatchScene1Id
            elif _address.funit_type == subgineries.SceneOn.funit_type_of_scene(2):
                funit_id = isMatchScene2Id
            elif _address.funit_type == subgineries.PresenceSensorsOn:
                funit_id = isPresenceSensorsOnId
            elif _address.funit_type == subgineries.PresenceSensorsOff:
                funit_id = isPresenceSensorsOffId
            elif _address.funit_type == subgineries.LightSensorsOn:
                funit_id = isLightSensorsOnId
            elif _address.funit_type == subgineries.LightSensorsOff:
                funit_id = isLightSensorsOffId
            elif _address.funit_type == subgineries.Off:
                funit_id = isOffId
            else:
                logging.debug(f'Unsupported funit "{_address.funit_type}"')
                return

            cl = self.funit_ids[_address.type].index(funit_id)

            if isinstance(_address, mqtt.spread.address.EngineryAddress):
                address = mqtt.jocket.address.EngineryAddress(_address.type, _address.id, cl)
            elif isinstance(_address, mqtt.spread.address.SubgineryAddress):
                address = mqtt.jocket.address.SubgineryAddress(_address.type, _address.id, cl)
            else:
                logging.debug(f'Unsupported entity "{_address.__class__.__name__}"')
                return

            var = mqtt.tros3.variable.VariableTRS3(None, _address.id, cl, value, _var.invalid)
            topic = mqtt.tros3.topic.TopicStateTros3(self.project_id, address)
        else:
            try:
                address = spread_address_to_mqtt(self.server_id, _address)
                funit = address.funit
                if _var.invalid:
                    value = funit.get(mqtt.const.DEF_VAL, funit.get(mqtt.const.STATE, int)())
                else:
                    value = _var.value
            except ClassifierError as ex:
                logging.debug(ex)
                return
            except BaseException:
                logging.debug(f'Unsupported address "{_address.__class__.__name__}"')
                return
            else:
                var = mqtt.jocket.variables.VariableJocket.create_data(_address.id, funit['id'], 'state', value, invalid=_var.invalid)
                topic = mqtt.jocket.topic.TopicState(self.project_id, address)

        self.publish(topic, var, retain=True)

    def on_tros3(self, _topic: mqtt.tros3.topic.TROS3Topic, data):
        _var = mqtt.tros3.variable.VariableTRS3.of(data)
        timestamp = datetime.fromisoformat(str(_var.timeStamp)) if _var.timeStamp else None

        if _topic.entity_addr.type in self.funit_ids:
            funit_name = self.funit_ids[_topic.entity_addr.type][_var.cl]

            if _topic.entity_addr.type in (light.DimmingLight.__name__, light.SwitchingLight.__name__,
                                           light.TunableWhiteLight.__name__,
                                           sensor.PresenceSensor.__name__, sensor.LightSensor.__name__):

                if funit_name == onId:
                    funit_type = light.On
                    value = True
                elif funit_name == offId:
                    funit_type = light.On
                    value = False
                elif funit_name == setLevelId:
                    funit_type = light.BrightnessLevel
                    value = _var.value
                elif funit_name == groupOnId:
                    funit_type = light.GroupLevel
                    value = 100
                elif funit_name == groupOffId:
                    funit_type = light.GroupLevel
                    value = 0
                elif funit_name == groupSetLevelId:
                    funit_type = light.GroupLevel
                    value = _var.value
                elif funit_name == setColorId:
                    funit_type = light.Temperature
                    value = _var.value
                else:
                    logging.debug(f'Unsupported funit "{funit_name}" for {_topic.entity_addr.type}({_var.id})')
                    return

                entity_address = mqtt.spread.address.EngineryAddress(_topic.p_id, _topic.entity_addr.type, _var.id, funit_type)

            elif _topic.entity_addr.type == subgineries.Lighting.__name__:
                if funit_name == onId:
                    funit_type = subgineries.On
                    value = _var.value
                elif funit_name == offId:
                    funit_type = subgineries.Off
                    value = _var.value
                elif funit_name == saveScene1Id or funit_name == saveScene2Id:
                    funit_type = subgineries.SaveScene
                    value = 1 if funit_name == saveScene1Id else 2
                elif funit_name == loadScene1Id or funit_name == loadScene2Id:
                    funit_type = subgineries.SceneOn
                    value = 1 if funit_name == loadScene1Id else 2
                elif funit_name == presenceSensorsOnId:
                    funit_type = subgineries.PresenceSensorsOn
                    value = True
                elif funit_name == presenceSensorsOffId:
                    funit_type = subgineries.PresenceSensorsOff
                    value = True
                elif funit_name == lightSensorsOnId:
                    funit_type = subgineries.LightSensorsOn
                    value = True
                elif funit_name == lightSensorsOffId:
                    funit_type = subgineries.LightSensorsOff
                    value = True
                else:
                    logging.debug(f'Unsupported funit "{funit_name}" for {_topic.entity_addr.type}({_var.id})')
                    return

                entity_address = mqtt.spread.address.SubgineryAddress(_topic.p_id, _topic.entity_addr.entity_id, _topic.entity_addr.type, funit_type)
                # entity_address = mqtt.spread.address.SubgineryAddress(_topic.p_id, e_data.location_id, _topic.entity_addr.type, funit_type)
            else:
                logging.debug(f'Unknown entity_type "{_topic.entity_addr.type}"')
                return
        else:
            logging.debug(f'Unknown entity(({_topic.entity_addr.type}){_var.id})')
            return

        self.publish(topic=mqtt.spread.topic.Set(entity_address),
                     data=mqtt.spread.variable.Variable(value, timestamp=timestamp),
                     retain=False)


def spread_address_to_mqtt(server_id: (str, int), _address: mqtt.spread.address.EntityAddress):
    if isinstance(_address, mqtt.spread.address.ProviderAddress):
        address = mqtt.jocket.address.ProviderAddress(mqtt.jocket.address.EngineryAddress.S_TYPE, server_id,
                                                      _address.manager_type, _address.manager_id,
                                                      _address.type, _address.id,
                                                      _address.funit_type)
    elif isinstance(_address, mqtt.spread.address.ManagerAddress):
        address = mqtt.jocket.address.ManagerAddress(mqtt.jocket.address.EngineryAddress.S_TYPE, server_id,
                                                     _address.type, _address.id, _address.funit_type)
    elif _address.type == engineries.LightingArea.__name__:
        address = mqtt.jocket.address.EngineryAddress(_address.type, _address.id, _address.funit_type)
    else:
        raise

    return address


def run():
    Frontier().connect()


if __name__ == '__main__':
    run()
