from typing import Union

from .can_packet import CanSend, CanDump
from ..dali.command import from_frame
from ..dali.frame import ForwardFrame
from ..payloads import Payload
from ..payloads import dali as dali_payload


class DaliSend(CanSend):
    def __init__(self, module_type, module_address, channel, end_frame, listen_collision=False):
        super(DaliSend, self).__init__(module_type, module_address, end_frame, cmd=0)
        self._channel = channel
        self.listen_collision = listen_collision

    @property
    def is_dali2(self):
        return len(self.end_frame) == 3

    @property
    def frame(self):
        b1 = 0x0
        b2 = 0x1 << self.channel - 1

        if self.send_twice:
            b1 = b1 | 0x1
        if self.is_dali2:
            b1 = b1 | 0x2
        if self.need_answer:
            b1 = b1 | 0x4

        frame = bytearray()
        frame.append(self.module_type)
        frame.append(b1)
        frame.append(b2)
        return frame + self.end_frame

    @property
    def payload(self):
        return generate_payload(self)

    @property
    def channel(self):
        return self._channel

    @property
    def data(self) -> bytearray:
        return self.end_frame

    @staticmethod
    def channel_from_frame(frame: bytearray) -> int:
        return 8 - bin(frame[2])[2:].rjust(8, '0').rindex('1')


class DaliDump(CanDump):

    def is_dali2(self):
        return (self.end_frame[1] >> 3) == 0x3

    @property
    def channel(self):
        return (self.frame[1] & 0x7) + 1

    @property
    def data(self):
        size_byte = self.size
        if self.is_error:
            return b''
        elif size_byte == 0x0:
            size = 1
        elif size_byte == 0x1:
            size = 2
        elif size_byte == 0x3:
            size = 3
        else:
            size = 0
        return self.frame[2:2+size]

    @property
    def size(self):
        return (self.frame[1] >> 3) & 0x3

    @property
    def is_error(self):
        return self.size >> 5 & 1 == 1

    @property
    def payload(self):
        return generate_payload(self)


query_classes = {}


def generate_payload(dump: Union[CanDump, CanSend]):
    frame = dump.data
    size = len(frame)

    if size < 2:
        if dump.channel in query_classes:
            return dali_payload.DaliResponse(query_classes[dump.channel], dump)
    elif size == 2:
        cmd = from_frame(ForwardFrame(16, frame))
        if cmd.is_query and isinstance(dump, DaliDump):
            query_classes[dump.channel] = cmd
        return dali_payload.DaliCommand(dump, cmd)
    elif size == 3:
        if frame[0] & 1 == 0:
            return dali_payload.Dali2Event(dump)
        else:
            cmd = from_frame(ForwardFrame(24, frame))
            if cmd.is_query and isinstance(dump, DaliDump):
                query_classes[dump.channel] = cmd
            return dali_payload.Dali2Command(dump, cmd)

    return Payload(dump)
