from spread_core.protocols.utils import bytes2str


class _Meta(type):
    _opcode = None

    @property
    def opcode(cls): return cls._opcode


class Response(metaclass=_Meta):
    def __init__(self, data):
        self.data = data

    def __str__(self):
        return self.value

    def __repr__(self):
        return f'[{self.__class__.__name__}]: {self.__str__()}'

    @property
    def value(self):
        return bytes2str(self.data).upper()


class Dictable(Response):
    _def_value = 'неизвестное значение для {}'
    _values = {}

    @property
    def value(self):
        if len(self.data) > 1:
            val_id = self.data[1]
            if val_id in self._values:
                return self._values[val_id]

            return self._def_value.format(hex(val_id))

        return super(Dictable, self).value


class Listable(Response):
    _def_value = 'неизвестное значение для {}'
    _values = {}

    @property
    def value(self):
        res = {}
        keys = list(self._values.keys())
        for key in keys:
            index = keys.index(key)
            if len(self.data) <= index:
                res[key] = 'unknown'
            elif len(self._values[key]) <= self.data[index]:
                res[key] = self._def_value.format(self.data[index])
            else:
                res[key] = self._values[key][self.data[index]]
        return res


def of(frame, opcode=None) -> Response:
    if opcode is None:
        opcode, *frame = frame

    res = __iterate_subclasses(Response, opcode)
    if res:
        return res(frame)
    return Response(frame)


def __iterate_subclasses(cls, opcode):
    for sc in cls.__subclasses__():
        if sc._opcode == opcode:
            return sc
        elif len(sc.__subclasses__()) > 0:
            r = __iterate_subclasses(sc, opcode)
            if r:
                return r

    return None


class Version(Response):

    @property
    def value(self):
        ver = self.data
        return dict(hv='{}{}'.format(ver[0:1].decode(), ver[1]),
                    fv='{}{}'.format(ver[2:3].decode(), ver[3]),
                    bv='{}{}'.format(ver[4:5].decode(), ver[5]))


class FirmwareVersion(Version):
    _opcode = 0x01


class UidMax(Response):
    _opcode = 0x05


class UidMin(Response):
    _opcode = 0x07


"""DaliModule"""


class ServiceDali(Response):
    _channel = None

    def __init__(self, *args, **kwargs):
        super(ServiceDali, self).__init__(*args, **kwargs)
        self._channel = self.data[0] + 1

    @property
    def channel(self):
        return self._channel


class LineStatusResponse(ServiceDali):
    _opcode = 0x21

    @property
    def channel(self): return True

    @property
    def value(self):
        return f'mask: {bin(self.data[0])[2:].rjust(8, "0")}'

    def is_active(self, channel: int) -> bool:
        """
        :param channel: line index in 1..8 as int
        :return: Is the line active
        """
        if channel not in range(1, 9):
            raise ValueError('channel expected in range 1..8')

        return bool((2**(channel-1)) & self.data[0])


class LineStateEvent(ServiceDali):
    _opcode = 0x22

    _states = {0: "Off", 1: 'On'}

    @property
    def state(self):
        return self.data[1]

    @property
    def is_active(self) -> bool:
        return bool(self.data[1])

    @property
    def value(self):
        state = self.state
        if self.state in self._states:
            state = self._states[state]
        return dict(channel=self.channel, state=state)


class EEService(ServiceDali, Dictable):
    _opcode = 0xee
    _values = {1: 'incorrect frame size',
               2: 'detect start bit during the stop condition. frame was rejected',
               3: 'backward frame was received in wrong time window. frame was rejected',
               4: 'forward frame was received in wrong time window. frame was rejected',
               5: 'bit timing violation'}


class ECService(ServiceDali, Dictable):
    _opcode = 0xec
    _values = {1: 'появление заднего фронта при выставленной единице. разрешение коллизии сервером',
               2: 'обнаружение нуля при выставленной единице. разрешение коллизии сервером',
               3: 'обнаружение разрешения коллизии между устройствами'}


"""Addressing"""


class UnaddressedNotify(Response):
    _opcode = 0xaf


"""Bootloader"""


class BootLoaderVersion(Version):
    _opcode = 0xb1


class BLReadyToReceiveFW(Response):
    _opcode = 0xb3


class BLFWReceived(Response):
    _opcode = 0xb5


class BLFWSuccess(Response):
    _opcode = 0xba


class BLFWRun(Response):
    _opcode = 0xbb


class BLFWError(Dictable):
    _opcode = 0xbe
    _values = {
        0x01: 'слишком большой размер файла',
        0x02: 'неизвестный код типа шифрования',
        0x03: 'ключ шифрования отсутствует в памяти',
        0x04: 'ошибка работы с внутренней памятью',
        0x05: 'ошибка контрольной суммы прошивки',
        0x06: 'в прошивке не найдена матрица совместимости',
        0x07: 'ошибка совместимости по классу модуля',
        0x08: 'ошибка совместимости по аппаратной версии',
        0x09: 'ошибка копирования прошивки в рабочую область. программа не будет запущена!!!',
        0x0A: 'неописанная ошибка'
    }


""" Modbus"""


class ChannelConfig(Listable):
    _opcode = 0x43
    _values = dict(channel=[1, 2, 3, 4],
                   baud=[1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600],
                   parity=['none', 'even', 'odd'],
                   stop=[1, 2])
