from ..address import Short, Group, InstanceNumber, InstanceType, InstanceGroup
from ..exceptions import EventError

# (default) Instance addressing, using instance type and number.
SCHEME_INST_TYPE_NUMBER = 0
# Device addressing, using short address and instance type.
SCHEME_ADDR_INST_TYPE = 1
# Device/instance addressing, using short address and instance number.
SCHEME_ADDR_INST_NUMBER = 2
# Device group addressing, using device group and instance type.
SCHEME_DEVICE_GROUP_INST_TYPE = 3
# Instance group addressing, using instance group and type.
SCHEME_INST_GROUP_TYPE = 4


class Event:
    _scheme_id = None
    short_addr = None
    instance_type = None
    instance_number = None
    device_group = None
    instance_group = None

    @staticmethod
    def of(frame):
        if not isinstance(frame, (bytes, bytearray)):
            raise EventError('Frame must be bytes')
        elif len(frame) != 3:
            raise EventError('Frame size is {}. Expected frame size is 3'.format(len(frame)))

        if frame[0] >> 7 == 0x0:
            if frame[1] >> 7 == 0x0:
                return DeviceEvent(frame)
            elif frame[1] >> 7 == 0x1:
                return DeviceInstanceEvent(frame)
        elif frame[0] >> 6 == 0x2:
            if frame[1] >> 7 == 0x0:
                return DeviceGroupEvent(frame)
            elif frame[1] >> 7 == 0x1:
                return InstanceEvent(frame)
        elif frame[0] >> 6 == 0x3:
            if frame[1] >> 7 == 0x0:
                return InstanceGroupEvent(frame)

        raise EventError('Event message decoding error: {}<...>{}<..>'.format(bin(frame[0] >> 6)[2:].rjust(2, '0'),
                                                                              bin(frame[1] >> 7)[2:].rjust(2, '0')))

    def __init__(self, frame):
        self._frame = frame
        mask = ''.join(bin(b)[2:].rjust(8, '0') for b in frame)
        self._value = int(mask[14:], 2)
        self._data = None

    @property
    def scheme_id(self):
        return self._scheme_id

    @property
    def data(self):
        if self._data:
            return self._data
        return EventData(self._value)

    @data.setter
    def data(self, owner_type):
        if owner_type == 1:
            self._data = ButtonEvent(self._value)
        elif owner_type == 3:
            self._data = PresenceEvent(self._value)
        elif owner_type == 4:
            self._data = LuminosityEvent(self._value)

    def __str__(self):
        attr = []
        if self.short_addr:
            attr.append(self.short_addr)
        if self.instance_number:
            attr.append(self.instance_number)
        if self.device_group:
            attr.append(self.device_group)
        if self.instance_group:
            attr.append(self.instance_group)
        return f'{self.data.__class__.__name__}({", ".join(str(i) for i in attr)}): {self.data.value}'


class DeviceEvent(Event):
    _scheme_id = 1

    def __init__(self, frame):
        super().__init__(frame)
        self.short_addr = Short(frame[0] >> 1)
        self.instance_type = InstanceType(frame[1] >> 2)


class DeviceInstanceEvent(Event):
    _scheme_id = 2

    def __init__(self, frame):
        super().__init__(frame)
        self.short_addr = Short(frame[0] >> 1)
        self.instance_number = InstanceNumber((frame[1] & 0x7f) >> 2)


class DeviceGroupEvent(Event):
    _scheme_id = 3

    def __init__(self, frame):
        super().__init__(frame)
        self.device_group = Group((frame[0] & 0x3f) >> 1)
        self.instance_type = InstanceType((frame[1] & 0x7f) >> 2)


class InstanceEvent(Event):
    _scheme_id = 0

    def __init__(self, frame):
        super().__init__(frame)
        self.instance_type = InstanceType((frame[0] & 0x3f) >> 1)
        self.instance_number = InstanceNumber((frame[1] & 0x7f) >> 2)


class InstanceGroupEvent(Event):
    _scheme_id = 0

    def __init__(self, frame):
        super().__init__(frame)
        self.instance_group = InstanceGroup((frame[0] & 0x3f) >> 1)
        self.instance_type = InstanceType((frame[1] & 0x7f) >> 2)


class EventData:
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value


class LuminosityEvent(EventData):
    pass


class PresenceEvent(EventData):
    """
    00 0000 ---0 - No movement detected. Correspon ding trig ger is the ‘No movement’ trigger.
    00 0000 ---1 - Movement detected. Correspondin g tri gger is the ‘Movement’ trigger.
    00 0000 -00- - Vacant. The area has become vacant. Correspon ding trig ger is the ‘Vacant’ tri gger.
    00 0000 -10- - Still vacant. The area is sti l l vacant. The event occurs at reg ular interval s as long as
                        the vacant con di tion hol ds. Corresponding trigger is the ‘Repeat’ trig ger.
    00 0000 -01- - Occupied The area has become occupied. Correspondin g tri gger is the ‘Occupied ’ trigger.
    00 0000 -11- - Still occupied. The area is sti l l occupied. The event occurs at reg ular interval s as long as
                        the occupied con di tion holds. Correspondin g trig ger is the ‘Repeat’ trig ger.
    00 0000 0--- - Presence sensor. The current event is tri ggered by a presence based sensor.
    00 0000 1--- - Movement sensor. The current event is tri ggered by a movement based sensor.
    """

    @property
    def has_move(self): return self._value & 0x1 == 0x1

    @property
    def is_occupied(self): return self._value & 0x2 == 0x2

    @property
    def is_still_occupied(self): return self._value & 0x4 == 0x4 and self.is_occupied

    @property
    def is_still_vacant(self): return self._value & 0x4 == 0x4 and not self.is_occupied

    @property
    def by_presence(self): return self._value & 0x8 == 0x0

    @property
    def by_movement(self): return self._value & 0x8 == 0x8

    def __str__(self):
        return '[{}]: {} ({})'.format(self.__class__.__name__, self.value, bin(self._value)[2:].rjust(5, '0'))

    @property
    def value(self): return self.has_move or self.is_occupied


class ButtonEvent(EventData):
    """
    buttonReleased	00 0000 0000		The button is released.
    buttonPressed 	00 0000 0001		The button is pressed.
    shortPress		00 0000 0010        The button is pressed and released, without being pressed quickly again (in
                                        case double press is enabled), or the button is pressed and quickly released
                                        (in case double press is disabled).
    doublePress		00 0000 0101		The button is pressed and released, quickly followed by another buttonpress.
    longPressStart	00 0000 1001		The button is pressed without releasing it.
    longPressRepeat	00 0000 1011		Following a long press start condition, the button is still pressed. The event
                                        occurs at regular intervals as long as the condition holds.
    longPressStop	00 0000 1100		Following a long press start condition, the button is released.
    buttonFree		00 0000 1110		The button has been stuck and is now released.
    buttonStuck		00 0000 1111		The button has been pressed for a very long time and is assumed stuck.
    """

    BUTTON_RELEASED     = 'ButtonReleased'
    BUTTON_PRESSED      = 'ButtonPressed'
    SHORT_PRESS         = 'ShortPress'
    DOUBLE_PRESS        = 'DoublePress'
    LONG_PRESS_START    = 'LongPressStart'
    LONG_PRESS_REPEAT   = 'LongPressRepeat'
    LONG_PRESS_STOP     = 'LongPressStop'
    BUTTON_FREE         = 'ButtonFree'
    BUTTON_STUCK        = 'ButtonStuck'
    DEFAULT             = 'UnknownButtonEvent'

    STATES = {
        0: BUTTON_RELEASED,
        1: BUTTON_PRESSED,
        2: SHORT_PRESS,
        5: DOUBLE_PRESS,
        9: LONG_PRESS_START,
        11: LONG_PRESS_REPEAT,
        12: LONG_PRESS_STOP,
        14: BUTTON_FREE,
        15: BUTTON_STUCK,
    }

    @property
    def value(self):
        return self.STATES.get(self._value, self.DEFAULT)
