import json
import logging
from typing import Any

from spread_core import mqtt
from spread_core.errors.project_errors import InitError
from spread_core.tools.settings import config, COMMUNICATE_TIMEOUT

READY_STATUS = 'Ready'


class AbstractInterfaceClass:
    wait_for_ready = False

    def __init__(self, mqtt_client):
        self.ready = not self.wait_for_ready
        self._mqtt_client = mqtt_client
        self._stopped = False
        self._send_topic = None
        self._dump_topic = None
        self._status_topic = None
        self.timeout = 1
        self.event_handler = None
        self.external_cmd_handler = None
        self.status_handler = None

    def set_bus_params(self, bus: str):
        self._send_topic = str(mqtt.bus.TopicBus.of_send(bus))
        self._dump_topic = str(mqtt.bus.TopicBus.of_dump(bus))
        self._status_topic = self._dump_topic.replace('Dump', mqtt.bus.STATE) + ('/%s' % mqtt.bus.STATUS)

    def update_config(self):
        self.timeout = config.get(COMMUNICATE_TIMEOUT, 1)

    def start(self):
        if self._dump_topic:
            logging.debug('Subscribed to %s' % self._dump_topic)
            self._mqtt_client.subscribe(str(self._dump_topic))

            if self._status_topic:
                logging.debug('Subscribed to %s' % self._status_topic)
                self._mqtt_client.subscribe(str(self._status_topic))
        else:
            raise InitError('SendTopic not set!')

    def set_event_handler(self, handler: iter):
        self.event_handler = handler

    def set_external_cmd_handler(self, handler: iter):
        self.external_cmd_handler = handler

    def set_status_handler(self, handler: iter):
        self.status_handler = handler

    def send(self, *args, **kwargs) -> Any:
        raise NotImplementedError()

    def on_dump(self, response, *args, **kwargs):
        """Handle target BUS-messages"""
        raise NotImplementedError()

    def on_status(self, status):
        """Handle BUS-status messages"""
        if isinstance(status, str):
            try:
                if status:
                    status = json.loads(status)['value'] == READY_STATUS
                else:
                    status = None
            except (KeyError, json.decoder.JSONDecodeError) as ex:
                logging.error('Json parsing error: %s => %s' % (status, str(ex)))
                status = False

        if self.ready != status:
            self.ready = status
            self.status_handler.send(status)

    def set_stopped(self):
        self._stopped = True

        logging.debug('Stop interface')

        if self._mqtt_client:
            logging.debug('Unsubscribed from %s' % self._dump_topic)
            self._mqtt_client.unsubscribe(str(self._dump_topic))

        if self.event_handler:
            try:
                self.event_handler.throw(StopIteration)
            except StopIteration:
                pass

        if self.external_cmd_handler:
            try:
                self.external_cmd_handler.throw(StopIteration)
            except StopIteration:
                pass
