import datetime
import json
import logging
import threading

from spread_core import mqtt
from spread_core.tools import settings


class ManagerOfBroker:
    def __init__(self, mqttc):
        self.mqttc = mqttc
        self.delayed_publishes = []
        self.delayed_timer = None
        self.log_subscribes = False

    def __stop_timer(self):
        if isinstance(self.delayed_timer, threading.Timer):
            self.delayed_timer.cancel()
            self.delayed_timer = None

    def __start_timer(self):
        self.__stop_timer()
        if self.delayed_publishes:
            publish = self.delayed_publishes.pop(0)
            self.delayed_timer = threading.Timer(publish.time_left, self.__on_timer, args=[publish])
            self.delayed_timer.name = 'DelayedPublisher'
            self.delayed_timer.start()

    def set_delayed_publish(self, publish):
        if self.delayed_publishes and self.delayed_publishes[0] == publish:
            self.__stop_timer()

        while publish in self.delayed_publishes:
            try:
                self.delayed_publishes.remove(publish)
            except:
                continue

        self.delayed_publishes.append(publish)
        self.delayed_publishes.sort(key=lambda pb: pb.time_left)

        if self.delayed_timer is None:
            self.__start_timer()

    def __on_timer(self, publish):
        self.publish(publish.topic, publish.data, publish.retain)
        self.__start_timer()

    def update_config(self):
        self.log_subscribes = settings.config.get(settings.LOG_SUBSCRIBES, False)

    def subscribe(self, topic, log=True):
        self.mqttc.subscribe(str(topic))
        if log or self.log_subscribes:
            logging.debug('Subscribed to {}'.format(topic))

    def will(self, topic, data, retain=False):
        data = data if isinstance(data, str) else None if data is None else data.pack()
        self.mqttc.will_set(str(topic), data, retain=retain)

    def unsubscribe(self, topic):
        self.mqttc.unsubscribe(str(topic))
        logging.debug('Unsubscribe from {}'.format(topic))

    def publish(self, topic, data, retain=False, delay=0, log_level=logging.DEBUG, **kwargs):
        if delay > 0:
            self.set_delayed_publish(DelayedPublish(delay, topic, data, retain))
        else:
            if log_level == logging.WARNING:
                log_func = logging.warning
            elif log_level == logging.INFO:
                log_func = logging.info
            else:
                log_func = logging.debug

            log_func(f'{"R " if retain else "  "} [{topic}]: '.ljust(100, '.') + f' {data}')

            if isinstance(data, dict):
                data = json.dumps(data)
            elif isinstance(data, mqtt.spread.variable.Variable):
                data = data.pack(**kwargs)

            self.mqttc.publish(str(topic), data, retain=retain)

    def on_exit(self):
        self.__stop_timer()
        self.delayed_publishes.clear()


class DelayedPublish:
    def __init__(self, delay, topic, data, retain):
        self.time = datetime.datetime.now().timestamp() + delay
        self.topic = topic
        self.data = data
        self.retain = retain

    @property
    def time_left(self): return self.time - datetime.datetime.now().timestamp()
    def __str__(self): return f'{datetime.datetime.fromtimestamp(self.time)} set {self.topic} as {self.data}'
    def __repr__(self): return self.__str__()
    def __iter__(self): return self.__str__()
    def __eq__(self, other): return isinstance(other, self.__class__) and other.topic == self.topic
    def __le__(self, other): return other.time_left <= self.time_left
    def __hash__(self): return hash(self.topic)