import logging
import signal
import sys
from time import sleep

import spread_core.tools.settings as settings
from spread_core.errors.project_errors import InitError
from spread_core.tools.mqtt_client import MqttClient


class Launcher:
    """
    Abstract launcher
    """
    _dumped = True
    _mqttc = None
    _manager = None
    _def_start_delay = 0

    def __init__(self):
        logging.info(self.__class__.__name__ + ' started')
        try:
            sleep(self.start_delay)
        except KeyboardInterrupt:
            return
        self.update_config()
        if self._dumped:
            settings.dump.init(settings.config.get(settings.DUMP, None))
        try:
            signal.signal(signal.SIGTERM, self.exit_handler)
            signal.signal(signal.SIGINT, self.exit_handler)
            signal.signal(signal.SIGHUP, self.update_config)
            signal.signal(signal.SIGUSR1, self.exit_handler)
            # signal.signal(signal.SIGALRM, self.exit_handler)
            if sys.platform == "linux" or sys.platform == "linux2":
                signal.signal(signal.SIGQUIT, self.exit_handler)
        except BaseException as ex:
            logging.error(ex)

    @property
    def start_delay(self):
        if self._def_start_delay > 0:
            return int(settings.config.get(settings.START_DELAY, self._def_start_delay))

        return self._def_start_delay

    @property
    def manager(self): return self._manager

    @property
    def project_id(self): return settings.config[settings.PROJECT_ID]

    @property
    def manager_id(self): return settings.config[settings.MANAGER_ID]

    @property
    def mqttc(self):
        if self._mqttc is None:
            self._mqttc = MqttClient(self,
                                     settings.config[settings.BROKER][settings.LOGIN],
                                     settings.config[settings.BROKER][settings.PASSWORD],
                                     self.on_connect, self.on_message,
                                     on_disconnect=self.on_disconnect)
        return self._mqttc

    def connect(self):
        host = settings.config[settings.BROKER][settings.HOST]
        port = settings.config[settings.BROKER][settings.PORT]
        logging.info('Connect to {}:{}'.format(host, port))
        self.mqttc.connect(host=host, port=port)
        self.mqttc.loop_forever()

    def on_connect(self):
        fatal = True
        try:
            self.manager.start()
        except InitError as ex:
            logging.warning(ex)
        except BaseException as ex:
            logging.exception(ex)
        else:
            fatal = False

        if fatal:
            self.on_exit(0, None)

    def subscribe(self, topic):
        self.mqttc.subscribe(str(topic))
        logging.debug('Subscribed to {}'.format(topic))

    def unsubscribe(self, topic):
        self.mqttc.unsubscribe(str(topic))
        logging.debug('Unsubscribe from {}'.format(topic))

    def publish(self, topic, data, retain=False):
        logging.debug(f'{"R" if retain else " "} [{topic}]: '.ljust(100, '.') + f' {data}')
        if data is None:
            _data = data
        else:
            _data = data if isinstance(data, (str, bytes, bytearray)) else data.pack()
        self.mqttc.publish(str(topic), _data, retain=retain)

    def on_exit(self, sig, frame):
        if self._manager and not isinstance(self._manager, self.__class__):
            try:
                self._manager.on_exit()
            except BaseException as ex:
                logging.exception(ex)

        try:
            settings.dump.on_exit()
        except BaseException as ex:
            logging.exception(ex)

    def exit_handler(self, sig, frame):
        try:
            self.on_exit(sig, frame)
        except BaseException as ex:
            logging.exception(ex)

        if self._mqttc and self._mqttc.is_connected():
            try:
                self._mqttc.on_disconnect = None
                self._mqttc.disconnect()
                self._mqttc.loop_stop(True)
            except BaseException as ex:
                logging.error(ex)

        logging.info('exit')

    def update_config(self, sig=None, frame=None):
        settings.config.parse()
        if self._manager and not issubclass(self._manager.__class__, self.__class__):
            self._manager.update_config()

    def on_message(self, mosq, obj, msg): pass
    def on_disconnect(self, client, userdata, rc): self.exit_handler(None, None)
