import argparse
import json
import logging
import os
import sys
import threading
import traceback
from datetime import datetime

from .settings_const import *


def log_uncaught_exceptions(ex_cls, ex, tb):
    text = '{}: {} {}:\n'.format(datetime.now(), ex_cls.__name__, ex)
    text += ''.join(traceback.format_tb(tb))
    logging.error(text)


sys.excepthook = log_uncaught_exceptions


class Dump(dict):
    def __init__(self):
        super(Dump, self).__init__()
        self.__file = None
        self.__timer = None
        self.__inited = False

    def __start_timer(self):
        self.__stop_timer()
        self.__timer = threading.Timer(self.get(DUMP_TIMEOUT, 1), self.__on_timer,)
        self.__timer.name = 'DumpTimer'
        self.__timer.start()

    def __stop_timer(self):
        if self.__timer and isinstance(self.__timer, threading.Timer):
            self.__timer.cancel()
        self.__timer = None

    def __on_timer(self, write_double=False):
        self.__timer = None
        if self:
            Dump.write(self.__file, self)
            if write_double:
                Dump.write(f'{self.__file}{os.path.extsep}{DOUBLE_POSTFIX}', self)

        else:
            try:
                os.remove(self.__file)
                os.remove(f'{self.__file}{os.path.extsep}{DOUBLE_POSTFIX}')
            except FileNotFoundError:
                pass
    @property
    def inited(self): return self.__file

    def init(self, dump_file):
        if not self.inited:
            if dump_file:
                if not os.path.exists(os.path.dirname(dump_file)):
                    os.makedirs(os.path.dirname(dump_file), exist_ok=True)

                try:
                    self.read(self, dump_file)
                    if len(self) == 0:
                        raise BaseException('%s is empty!' % dump_file)
                except BaseException as ex1:
                    dump_double_file = f'{dump_file}{os.path.extsep}{DOUBLE_POSTFIX}'
                    try:
                        logging.warning('%s reading exception: %s' % (dump_file, ex1))
                        self.read(self, dump_double_file)
                    except BaseException as ex2:
                        logging.warning('%s reading exception: %s' % (dump_double_file, ex2))
                self.__file = dump_file
            else:
                raise FileExistsError('Конфиг не содержит дамп-файла!!!')

        return self

    def get_dump(self, entity_type, entity_id, funit_type):
        entity_id = str(entity_id)
        if entity_type in self:
            if entity_id in self[entity_type]:
                if funit_type in self[entity_type][entity_id]:
                    return True, self[entity_type][entity_id][funit_type]

        return False, None

    def get_dump_entity(self, entity):
        if entity.__class__.__name__ in self:
            e_id = str(entity.id)
            if e_id in self[entity.__class__.__name__]:
                return self[entity.__class__.__name__][e_id]

        return dict()

    def set_dump(self, entity, funit_type, value):
        funit_type = str(funit_type)
        if self.inited:
            _override = False
            if entity.__class__.__name__ not in self:
                if value == KILL_ITEM:
                    return
                _override = True
                self[entity.__class__.__name__] = dict()
            e_id = str(entity.id)
            if e_id not in self[entity.__class__.__name__]:
                if value == KILL_ITEM:
                    return
                _override = True
                self[entity.__class__.__name__][e_id] = dict()
            if (funit_type not in self[entity.__class__.__name__][e_id]
                    or self[entity.__class__.__name__][e_id][funit_type] != value):
                _override = True
                if value == KILL_ITEM:
                    self[entity.__class__.__name__][e_id].pop(funit_type, None)
                    if not self[entity.__class__.__name__][e_id]:
                        self[entity.__class__.__name__].pop(e_id)
                        if not self[entity.__class__.__name__]:
                            self.pop(entity.__class__.__name__)
                else:
                    self[entity.__class__.__name__][e_id][funit_type] = value

            if _override:
                self.__start_timer()

    def __setitem__(self, key, value):
        if self.get(key) != value:
            super(Dump, self).__setitem__(key, value)
            self.__start_timer()

    def on_exit(self):
        if self.__timer and isinstance(self.__timer, threading.Timer):
            self.__timer.cancel()
            self.__on_timer(write_double=True)

    @staticmethod
    def read(owner, file_name):
        with open(file_name, "r") as file:
            res = file.read()
        res = json.loads(res)
        if res:
            for name in res.keys():
                if name == 'include' and isinstance(res[name], list):
                    for inc_item in res[name]:
                        Dump.read(owner, inc_item)
                else:
                    owner[name] = res[name]

    @staticmethod
    def write(path, data: dict):
        logging.debug(f'write dump as {path}')
        try:
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, 'w') as outfile:
                outfile.write(json.dumps(dict(data), indent=2))
        except BaseException as ex:
            logging.error(ex)


class Config(dict):
    def __init__(self):
        super(Config, self).__init__()
        self.args = None

    def __getitem__(self, item):
        if not self.args:
            self.parse()
        return super(Config, self).__getitem__(item)

    def parse(self):
        self.args = parser.parse_args()
        conf_path = self.args.config

        if isinstance(conf_path, str) and os.path.isfile(conf_path):
            try:
                Dump.read(config, conf_path)
            except BaseException as ex:
                print(f'{conf_path} reading error: \n{ex}')
                sys.exit(0)
        else:
            print('ATTENTION: Settings file not set. Please provide some settings file. Breaked...')
            sys.exit(0)

        """LOGGING"""
        log_level = config.get(LOG_LEVEL, 0)
        if log_level > 1:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO

        logging.getLogger().setLevel(log_level)
        logging.info('set logLevel as "%s"' % config.get(LOG_LEVEL, 0))


config = Config()
dump = Dump()

parser = argparse.ArgumentParser(description='Spread services launcher')
parser.add_argument('-c', '--config', type=str, help='provide a config file path')

logging.basicConfig(format=u'%(levelname)-8s [%(asctime)s]  %(message)s', level=logging.INFO, stream=sys.stdout)
