import asyncio
import datetime
import os
import signal
import sys
import threading

import aiomqtt


class Counter:
    period = 1

    def __init__(self):
        self._count = 0
        self.start_time = None

    def on_start(self):
        if self.start_time is None:
            self.start_time = datetime.datetime.now().timestamp()

    @property
    def count(self):
        self.on_start()
        return self._count

    @property
    def get_and_up(self):
        self.on_start()
        self._count += 2
        return self._count

    @property
    def speed(self): return round(self._count / self.period) if self.period != 0 else 0

    @property
    def progress(self):
        if self.start_time is None:
            return 0
        return max(0, min(100, int((datetime.datetime.now().timestamp() - self.start_time) / self.period * 100)))


class PingPong:
    PING = 'System/Speed/Ping'
    PONG = 'System/Speed/Pong'
    vendor = ''

    def __init__(self, loop=None):
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self._client = None
        self.counter = Counter()
        self.connected = asyncio.Event(loop=self.loop)
        self.subscribed = asyncio.Event(loop=self.loop)

    @property
    def client(self):
        if self._client is None:
            c = aiomqtt.Client(self.loop)
            c.username_pw_set(username='admin', password='broker')
            c.loop_start()  # See "About that loop..." below.

            c.on_connect = self.on_connect
            c.on_subscribe = self.on_subscribe
            c.on_message = self.on_message
            self._client = c
        return self._client

    async def start(self):
        await self.client.connect("127.0.0.1")
        await self.connected.wait()
        await self.subscribed.wait()

    def on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            self.connected.set()
            client.subscribe(self.vendor)
        else:
            raise BaseException('Auth failed')

    def on_subscribe(self, client, userdata, mid, granted_qos):
        self.subscribed.set()

    def on_message(self, client, userdata, message):
        raise NotImplemented()


class Ping(PingPong):
    topic = PingPong.PING
    vendor = PingPong.PONG

    async def start(self):
        await super(Ping, self).start()
        await self.client.publish(self.topic, self.counter.count).wait_for_publish()
        await asyncio.sleep(self.counter.period)
        self.stop()

    def stop(self):
        self.client.disconnect()
        sys.stdout.write('\r' + ' ' * 15 + '\r')
        sys.stdout.flush()
        if self.counter.count == 0:
            print('Failed')
        else:
            print('%s pcs/sec; %s ms/pcs' % (self.counter.speed, round(1000/self.counter.speed, 2)))

    def on_message(self, client, userdata, message):
        sys.stdout.write('\rProgress: %s' % self.counter.progress + '%')
        sys.stdout.flush()
        client.publish(Ping.topic, self.counter.get_and_up)
        # self.loop.create_task(client.publish(Ping.topic, self.counter.get_and_up).wait_for_publish())

    @staticmethod
    def run(ping, pong):
        try:
            ping.loop.run_until_complete(ping.start())
        except KeyboardInterrupt:
            print('Breaked')
        finally:
            pong.client.disconnect()
            pong.loop.stop()
            os.kill(os.getpid(), signal.SIGTERM)


class Pong(PingPong):
    topic = PingPong.PONG
    vendor = PingPong.PING

    def on_message(self, client, userdata, message):
        client.publish(self.topic, message.payload)
        # self.loop.create_task(client.publish(self.topic, message.payload).wait_for_publish())

    @staticmethod
    def run(pong):
        try:
            pong.loop.run_until_complete(pong.start())
            pong.loop.run_forever()
        finally:
            pong.loop.stop()


def run():
    pong_loop = asyncio.new_event_loop()
    asyncio.set_event_loop(pong_loop)

    ping_loop = asyncio.new_event_loop()
    asyncio.set_event_loop(ping_loop)

    ping = Ping(loop=ping_loop)
    pong = Pong(loop=pong_loop)

    threading.Thread(target=Pong.run, args=[pong], name="PongThread").start()
    threading.Thread(target=Ping.run, args=[ping, pong], name="PingThread").start()


if __name__ == '__main__':
    run()
