import json
import logging
import threading

import math

import spread_core.mqtt as mqtt
from spread_core.tools import settings
from spread_core.tools.service_launcher import Launcher


class Item:
    def __init__(self, msg):
        self._topic = msg.topic
        self._data = msg.payload
        self.retain = msg.retain == 1
        self.size = len(msg.payload)
        self.count = 1

    def __add__(self, other):
        self.size += other.size
        self.count += 1

        return self

    def __str__(self):
        return f'"{self.topic}": {self.count} times of {size(self.size)}'

    @property
    def short(self): return '/'.join(self.topic.split('/')[:2])

    @property
    def topic(self): return self._topic

    @property
    def data(self): return self._data

    def log(self):
        if len(self.data) == 0:
            data = 'empty'
        elif len(self.data) > 256:
            data = f'data of {size(len(self.data))}'
        else:
            try:
                data = json.loads(self.data.decode())
            except:
                try:
                    data = mqtt.tros3.variable.VariableTRS3.of(self.data)
                except:
                    data = " ".join([hex(b)[2:].rjust(2, "0") for b in self.data])
        logging.info(f'{self.topic}: {data}')


class Statistic(dict):
    def __init__(self):
        super(Statistic, self).__init__()

    def set(self, item: Item):
        if item.short not in self:
            self[item.short] = item
        else:
            self[item.short] += item

    def to_string(self):
        def make(t, c, s):
            return f'\n{t.ljust(15, " ")} {str(c).rjust(5, " ")}    {size(s).rjust(9, " ")}'
        _count = 0
        _size = 0
        result = '\n'
        for message in self.values():
            _count += message.count
            _size += message.size
            result += make(message.short, message.count, message.size)

        result += make("TOTAL", _count, _size) + '\n'
        return result


class TopicHandler(Launcher):
    _dumped = False
    _def_start_delay = 0

    def __init__(self, topic, verbose=False, remove=False):
        self._manager = self
        self._exit_timer = None
        self.topics = topic.split('|')
        self.verbose = verbose
        self.remove = remove
        self.restart_timer()
        self.statistic = Statistic()
        super(TopicHandler, self).__init__()

    def start(self):
        for topic in self.topics:
            logging.info('subscribed to {}'.format(topic))
            self.mqttc.subscribe(f'{topic}')

    def restart_timer(self):
        if self._exit_timer:
            self._exit_timer.cancel()
        self._exit_timer = threading.Timer(1, function=self.exit_handler, args=[1, None])
        self._exit_timer.start()

    def exit_handler(self, sig, frame):
        logging.info(self.statistic.to_string())
        super(TopicHandler, self).exit_handler(sig, frame)

    def clear(self, item: Item):
        if len(item.data) > 0:
            self.mqttc.publish(item.topic, None, retain=True)

    def on_message(self, mosq, obj, msg):
        if msg.retain == 1:
            try:
                item = Item(msg)
                self.statistic.set(item)
                if self.verbose:
                    item.log()

                if self.remove:
                    self.clear(item)
            except BaseException as ex:
                logging.exception(ex)
            finally:
                self.restart_timer()


def size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "Kb", "Mb", "Gb", "Tb", "Pb", "Eb", "Zb", "Yb")

    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2 if i > 1 else 1)
    return "%s %s" % (s, size_name[i])


def run():
    settings.parser.add_argument('-t', '--topic', type=str, help='topic for subscribe')
    settings.parser.add_argument('-r', '--remove', action='store_true', help='use for clean topics')
    settings.parser.add_argument('-v', '--verbose', action='store_true', help='detailed messages')
    TopicHandler(settings.config.args.topic, settings.config.args.verbose, settings.config.args.remove).connect()


if __name__ == '__main__':
    run()
