import datetime
import json

from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer

from spread_core.protocols.payloads.dali import DaliPayload, DaliCommand
from spread_dashboard.monitor.vendors.dali.broker_vendor import WSVendorBroker
from spread_dashboard.monitor.vendors.dali.can_vendor import WSVendorCan


class ChatConsumer(WebsocketConsumer):
    dali_requests = None
    room_name = 'default_group'

    def __init__(self, *args, **kargs):
        super(ChatConsumer, self).__init__(*args, **kargs)

    @property
    def vendor(self):
        if ChatConsumer._vendor is None:
            ChatConsumer._vendor = WSVendorBroker()
        return ChatConsumer._vendor

    def connect(self):
        self.room_group_name = 'chat_%s' % self.room_name

        if self.room_group_name not in self.channel_layer.groups:
            self.vendor.set_listener(self.room_group_name, self._on_dali)

        # Join room group
        async_to_sync(self.channel_layer.group_add)(
            self.room_group_name,
            self.channel_name
        )

        self.accept()

    def disconnect(self, close_code):
        # Leave room group
        async_to_sync(self.channel_layer.group_discard)(
            self.room_group_name,
            self.channel_name
        )

        if self.room_group_name not in self.channel_layer.groups:
            self.vendor.remove_listener(self.room_group_name)

    # Receive message from WebSocket
    def receive(self, text_data=None, bytes_data=None):
        text_data_json = json.loads(text_data)
        message = text_data_json['message']
        if message['type'] == 'add':
            self.vendor.add_subscription(self.room_group_name, *message['topics'])
        elif message['type'] == 'remove':
            self.vendor.remove_subscription(self.room_group_name, *message['topics'])

    # Receive message from room group
    def chat_message(self, event):
        message = event['message']

        # Send message to WebSocket
        self.send(text_data=json.dumps({
            'message': message
        }))

    def _on_dali(self, topic, message):
        if isinstance(message, str):
            data = message.split('#')

            try:
                time = data[1].split('TS')[1]
                time = time.split(':')[0]
                time = datetime.datetime.fromtimestamp(int(time) / 1000).time()
            except:
                time = datetime.datetime.now().time()

            frame = bytes.fromhex(data[0])
        else:
            frame = message
            data = [''.join([hex(b)[2:].rjust(2, '0') for b in frame])]
            time = datetime.datetime.now().time()

        payload = DaliPayload.of(frame, self.__class__.dali_requests.get(topic, None))
        if isinstance(payload, DaliCommand) and payload.command and payload.command.response:
            self.__class__.dali_requests[topic] = payload.command

        async_to_sync(self.channel_layer.group_send)(
            self.room_group_name,
            {
                'type': 'chat_message',
                'message': dict(
                    topic='/'.join(topic.split('/')[2:]),
                    direct=topic.split('/')[1],
                    time=str(time)[:-3],
                    hex=data[0].upper(),
                    address=payload.address if payload.has_address else '',
                    instance=payload.instance if payload.has_instance else '',
                    mark=payload.mark,
                    name=payload.name,
                    desc=payload.desc
                )
            }
        )


class ChatConsumerBroker(ChatConsumer):
    _vendor = None
    room_name = 'dali'
    dali_requests = {}

    @property
    def vendor(self):
        if ChatConsumerBroker._vendor is None:
            ChatConsumerBroker._vendor = WSVendorBroker()
        return ChatConsumerBroker._vendor


class ChatConsumerCan(ChatConsumer):
    _vendor = None
    room_name = 'raw_dali'
    dali_requests = {}

    @property
    def vendor(self):
        if ChatConsumerCan._vendor is None:
            ChatConsumerCan._vendor = WSVendorCan()
        return ChatConsumerCan._vendor
