import logging
import re
from threading import Thread

import can

from spread_dashboard.monitor.vendors.base_vendor import WSVendor


def coroutine(func):
    def inner(*args, **kwargs):
        g = func(*args, **kwargs)
        g.send(None)
        return g
    return inner


def parse_topic(topic):
    pattern = r'^Bus/\w+/Rapida/(\d+)/Dali/(\d+)/\d+$'
    return (int(i) for i in re.search(pattern, topic).groups())


def build_topic(direct, gw, module):
    return 'Bus/%s/Rapida/%s/Dali/%s/{}' % (direct, gw, module)


class WSVendorCan(WSVendor):
    def __init__(self):
        super(WSVendorCan, self).__init__()
        self.buses = {}
        self.delegator = self.delegator(self.subgen())

    def run(self) -> None:
        Thread.run(self)

    def subgen(self):
        while True:
            try:
                topic, payload = yield
                self._on_recv(topic, payload)
            except StopIteration:
                logging.debug("STOP vendering")
                break
            except BaseException as ex:
                logging.exception(ex)

    @coroutine
    def delegator(self, g):
        yield from g

    def add_subscription(self, group, *topics):
        for topic in topics:
            self.listeners[group].topics.setdefault(topic, 0)
            if self.listeners[group].topics[topic] == 0:
                try:
                    gw, module = parse_topic(topic)
                except BaseException:
                    logging.warning('Can parse topic to GATEWAY_ID and MODULE_ID')
                else:
                    self.listeners[group].topics[topic] += 1

                    if gw not in self.buses:
                        try:
                            bus = BusThread(gw, self.delegator)
                        except OSError as ex:
                            logging.error(ex)
                            continue
                        else:
                            bus.start()
                            self.buses[gw] = bus
                    else:
                        bus = self.buses[gw]

                    bus.add_module(module)

    def remove_subscription(self, group, *topics):
        for topic in topics:
            if topic in self.listeners[group].topics:
                self.listeners[group].topics[topic] -= 1
                if self.listeners[group].topics[topic] <= 0:
                    try:
                        gw, module = parse_topic(topic)
                    except:
                        logging.warning('Can parse topic to GATEWAY_ID and MODULE_ID')
                    else:
                        if gw in self.buses:
                            bus = self.buses.pop(gw)
                            bus.remove_module(module)
                            bus.stop()
                            del bus


class BusThread(Thread):
    CONTROLLER_ID = 31

    def __init__(self, gateway, delegator):
        super(BusThread, self).__init__(name='VendorBus_%s' % gateway)
        self.gateway = gateway
        self.delegator = delegator
        self._stopped = False

        self._bus = can.Bus(
            channel=self.can_name,
            **dict(
                bustype='socketcan',
                can_filters=[]
            )
        )

    @property
    def can_name(self):
        return '%scan%s' % ('' if self.gateway == 0 else 'r', self.gateway)

    def add_module(self, module):
        filters = self._bus.filters
        if filters is None:
            filters = []
        filters += [
            {
                "can_id": (module << 5) + self.CONTROLLER_ID,
                "can_mask": 1023,
            },
            {
                "can_id": (self.CONTROLLER_ID << 5) + module,
                "can_mask": 1023,
            }
        ]

        self._bus.set_filters(filters)

    def remove_module(self, module):
        self._bus.set_filters(
            [f
             for f in self._bus.filters
             if f['can_id'] != (module << 5) + self.CONTROLLER_ID
             and f['can_id'] != (self.CONTROLLER_ID << 5) + module]
        )

    def stop(self):
        logging.info('Stop listening %s' % self.can_name)
        self._stopped = True
        self._bus.stop_all_periodic_tasks()
        self._bus.shutdown()

    def run(self) -> None:
        while not self._stopped:
            try:
                logging.info('Start listening %s' % self.can_name)
                for msg in self._bus:
                    try:
                        if msg.data[0] != 0x01:
                            continue
                        if (msg.arbitration_id >> 5) == self.CONTROLLER_ID:
                            channels = (
                                i for i in range(0, 8)
                                if (msg.data[2] >> i) & 1 == 1
                            )
                            topic = build_topic('Send', self.gateway, msg.arbitration_id & self.CONTROLLER_ID)
                            payload = msg.data[3:]
                        else:
                            topic = build_topic('Dump', self.gateway, (msg.arbitration_id >> 5) & self.CONTROLLER_ID)
                            payload = msg.data[2:]
                            channels = [msg.data[1] & 0x07]

                        for ch in channels:
                            try:
                                self.delegator.send((topic.format(ch+1), payload))
                            except BaseException as ex:
                                logging.exception(ex)
                                continue

                    except BaseException as ex:
                        logging.exception(ex)
            except KeyboardInterrupt:
                self.stop()
            except BaseException as ex:
                if not self._stopped:
                    logging.exception(ex)
