import json
import logging
import os
import uuid
import zipfile

import spread_core.mqtt as mqtt
from spread_core.mqtt.project import TopicProject
from spread_core.bam.generator import generate_provider
from spread_core.bam.dali.managers import Dali
from spread_core.bam.lom.manager import Lom
from spread_core.bam.knx.manager import Knx
from spread_core.bam.mercury.managers import Mercury
from spread_dashboard.project import models
from spread_dashboard.utils import app
from . import configs


class Project:
    def __init__(self):
        ctp, project_id, _, self.name = self.read_pid()
        app().project_id = project_id
        configs.Config.PROJECT_ID = project_id
        if ctp:
            configs.Frontier.SERVER_ID = self.read_sid(ctp)

    @staticmethod
    def read_pid(ctp_path=None):
        """
        :param ctp_path: project.ctp path
        :return: ctp, project_id, root_location_id, name
        """
        if ctp_path is None:
            ctp_path = configs.Projector.ctp_file
        if os.path.exists(ctp_path):
            ctp = zipfile.ZipFile(ctp_path, 'r')
            entity = json.loads(ctp.read(TopicProject.HEADER + '.json').decode())[TopicProject.HEADER]
            return (
                ctp,
                entity['project']['id'],
                entity['project']['rootLocationID'],
                entity['project'].get('title', entity['project'].get('name', 'Project %s' % entity['project']['id']))
            )
        return None, None, None, None

    @staticmethod
    def read_sid(ctp):
        for entity in json.loads(ctp.read(TopicProject.SERVERS + '.json').decode())[TopicProject.SERVERS]:
            if entity['type'] == 'Spread':
                return entity['id']

    @classmethod
    def read_ctp(cls):
        ctp = zipfile.ZipFile(configs.Projector.ctp_file, 'r')

        cls.read_ctp_file(models.Location, TopicProject.LOCATIONS, ctp)
        cls.read_ctp_file(models.Subginery, TopicProject.SUBGINERIES, ctp)
        cls.read_ctp_file(models.Manager, TopicProject.MANAGERS, ctp)
        cls.read_ctp_file(models.Enginery, TopicProject.ENGINERIES, ctp)
        cls.read_ctp_file(models.Provider, TopicProject.PROVIDERS, ctp)

    @staticmethod
    def read_ctp_file(model, file, ctp):
        stack = []
        for entity in json.loads(ctp.read(file + '.json').decode())[file]:
            try:
                assert issubclass(model, models.Entity)
                model.objects.update_or_create(id=entity['id'], defaults=model.separator(entity))
                stack.append(entity['id'])
            except BaseException as ex:
                logging.error('Filling error: %s' % ex)
        model.objects.exclude(id__in=stack).delete()

    @classmethod
    def generate_configs(cls, force=False):
        ctp, project_id, _, _ = cls.read_pid(None)

        if force or project_id != configs.Config.PROJECT_ID:
            configs.Tree.reset()

        configs.Config.PROJECT_ID = project_id
        configs.Frontier.SERVER_ID = cls.read_sid(ctp)

        tree = configs.Tree()

        configs.Common().save(generate_sv=True)
        configs.Dashboard().save(generate_sv=True)
        if configs.Frontier.exist():
            configs.Frontier().save(generate_sv=True)
        configs.Projector().save(generate_sv=True)
        configs.Scheduler().save(generate_sv=True)
        configs.Scripter().save(generate_sv=True)
        configs.Cycler().save(generate_sv=True)
        configs.EventsSqlite().save(generate_sv=True)
        configs.LightingEquipment().save(generate_sv=True)

        app().publish_web_token()

        adapters = []
        for manager in json.loads(ctp.read(TopicProject.MANAGERS + '.json').decode())[TopicProject.MANAGERS]:
            adapter = None
            if manager['type'] == Dali.__name__:
                bus = manager['attributes']['bus']
                m = configs.DaliManager(bus)
                if bus not in adapters:
                    adapters.append(bus)
                    adapter = configs.CanAdapter(bus)
            elif manager['type'] == Mercury.__name__:
                bus = manager['attributes']['bus']
                m = configs.MercuryManager(bus)
                if bus not in adapters:
                    adapters.append(bus)
                    adapter = configs.TcpAdapter(bus)
            elif manager['type'] == Lom.__name__:
                m = configs.LomManager()
            elif manager['type'] == Knx.__name__:
                bus = manager['attributes']['bus']
                m = configs.KnxManager(bus=bus)
                if bus not in adapters:
                    adapter = configs.KnxNetIpAdapter(bus)
                    adapters.append(bus)
            else:
                continue

            if adapter:
                adapter.save(generate_sv=True)
                tree.decrease(adapter.config_dest, adapter.file)
            m.managerID = manager['id']
            m.save(generate_sv=True)
            tree.decrease(m.config_dest, m.file)

        tree.clear()


class Settings(dict):
    def __init__(self):
        super(Settings, self).__init__()
        self.read()

    def read(self) -> None:
        self.update(configs.Dashboard().serialize())
        for file in self.pop('include', []):
            if os.path.exists(file):
                with open(file, 'r') as stream:
                    self.update(json.load(stream))


def broker_auth() -> dict:
    guid = str(uuid.uuid4())
    reply = str(mqtt.spread.topic.Reply(app().project_id, ''))
    return dict(web_port=2883, login='admin', password='broker', reply=reply, guid=guid)


def config(model: models.Provider):
    guid = uuid.uuid4()
    provider = generate_provider(model.serialize()) if model else None
    manager_model = models.Manager.objects.get(id=model.managerID) if model else None
    manager = (Dali if manager_model.type == Dali.__name__ else Mercury)(app().project_id, manager_model.id)
    manager.parse(manager_model.serialize())
    provider.set_manager(manager)
    return dict(broker=broker_auth(),
                project=dict(state=str(mqtt.spread.topic.State.of_provider(provider, '+')),
                             get=str(mqtt.spread.topic.Renew.of_provider(provider, '{}')),
                             set=str(mqtt.spread.topic.Set.of_provider(provider, '{}')),
                             reply=str(mqtt.spread.topic.Reply(app().project_id, guid)),
                             session_id=str(guid),
                             entity_id=model.id if model else None))
