import base64
from datetime import datetime

import cbor
import math

from spread_core.tools import utils
from . import profiles
from .const import *


class LFrame:
    _cmd_id = None
    _fport = 1

    def __init__(self, dev_eui, pack=None, mc=False):
        """
        :param dev_eui: str - Идентификатор контроллера
        :param pack: bytes
        """
        self.pending = False
        self.immediately = True
        pack = [] if pack is None else pack
        self.dev_eui = dev_eui
        self.is_mc = mc
        self._pack = list(pack[1:])

    def __str__(self):
        return f'{self.__class__.__name__}(<{self.dev_eui}>): {self.pack}'

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return isinstance(other, LFrame) and self.__class__ == other.__class__ and self.dev_eui == other.dev_eui and self.pack == other.pack

    def check_response(self, response):
        if self.is_mc:
            return True

        return self.dev_eui == response.dev_eui

    pack = property(fget=lambda self: [self._cmd_id] + self._pack)
    fport = property(fget=lambda self: self._fport)


class LCommand(LFrame):

    @staticmethod
    def frame_to_data(pack):
        return pack[1:]

    @property
    def frame(self):
        data = self.pack
        data = cbor.dumps(data)
        data = base64.b64encode(data).decode()
        data = {'data': data, 'fport': self.fport, 'imme': True}

        if self.is_mc:
            data['mcaddr'] = self.dev_eui
        else:
            data['dev_eui'] = self.dev_eui

        return data

    def split(self, max_length: int):
        yield self

    def part_count(self, max_length): return 1


class SetConfig(LCommand):
    """Команды конфигурации устройства в соответствии

    :param dev_eui: идентификатор контроллера
    :param _items: пары значений:
                conf_id - идентификатор параметра конфигурации
                value - значение параметра конфигурации
    """
    _cmd_id = 1

    def __init__(self, dev_eui, *_items: [int, int], **kwargs):
        if len(_items) == 1 and isinstance(_items[0][0], list):
            _items = _items[0]

        super(SetConfig, self).__init__(dev_eui, [self._cmd_id, *_items], **kwargs)

    @property
    def items(self) -> dict:
        return {conf_id: value for conf_id, value in self._pack}

    def split(self, max_length: int) -> [LCommand]:
        count = self.part_count(max_length)
        for i in range(0, count):
            cmd = SetConfig(self.dev_eui, *self._pack[i*max_length:(i+1)*max_length], mc=self.is_mc)
            cmd.immediately = i == 0
            cmd.pending = i == 0 and count > 1
            yield cmd

    def part_count(self, max_length): return math.ceil(len(self._pack) / max_length)

    def check_response(self, response: LFrame):
        return response._cmd_id == self._cmd_id and super(SetConfig, self).check_response(response)


class SetProfile(LCommand):
    """
    :param dev_eui: идентификатор контроллера
    :param profile: объект DimmingProfile или pack
    """

    _profile_class = None

    def __init__(self, dev_eui: str, *profile: (list, profiles.DimmingProfile), **kwargs):
        if isinstance(profile[0], profiles.DimmingProfile):
            self.profile = profile[0]
            profile = profile[0].pack
        else:
            self.profile = self._profile_class(*profile)

        super(SetProfile, self).__init__(dev_eui, [self._cmd_id, *profile], **kwargs)

    @classmethod
    def of(cls, dev_eui: str, profile: profiles.DimmingProfile, mc: bool):
        res = utils.get_subclass(cls, lambda scl: issubclass(scl, cls) and scl._profile_class == profile.__class__)
        return res(dev_eui, profile, mc=mc)

    def split(self, max_length: int):
        for i in range(0, self.part_count(max_length)):
            cmd = self.__class__(
                self.dev_eui,
                *list(self.pack[1:3] + self.pack[3+i*(max_length - 3):3+(i+1)*(max_length - 3)]),
                mc=self.is_mc
            )
            cmd.immediately = i == 0
            cmd.pending = i == 0 and self.part_count(max_length) > 1
            yield cmd

    def part_count(self, max_length):
        return math.ceil(len(self.profile.items)/(max_length-3))


class SetGlobalGeoProfile(SetProfile):
    """Добавление/изменение записей в глобальном профиле диммирования на основе координат места установки"""
    _cmd_id = 2
    _profile_class = profiles.GlobalGeoProfile


class SetGlobalLumProfile(SetProfile):
    """Добавление/изменение записей в глобальном профиле диммирования на основе уровня освещенности"""
    _cmd_id = 3
    _profile_class = profiles.GlobalLumProfile


class SetPersonalGeoProfile(SetProfile):
    """Добавление/изменение записей в персональном профиле диммирования на основе координат места установки"""
    _cmd_id = 4
    _profile_class = profiles.PersonalGeoProfile


class SetPersonalLumProfile(SetProfile):
    """Добавление/изменение записей в персональном профиле диммирования на основе уровня освещенности"""
    _cmd_id = 5
    _profile_class = profiles.PersonalLumProfile


class SetGlobalUTCProfile(SetProfile):
    """Добавление/изменение записей в глобальном профиле диммирования на основе абсолютного времени"""
    _cmd_id = 6
    _profile_class = profiles.GlobalUTCProfile


class SetPersonalUTCProfile(SetProfile):
    """Добавление/изменение записей в персональный профиль диммирования на основе абсолютного времени"""
    _cmd_id = 7
    _profile_class = profiles.PersonalUTCProfile


class RemoveProfileItem(LCommand):
    """Удаление записей из профиля диммирования

    :param dev_eui: идентификатор контроллера
    :param _profile_id: идентификатор профиля
    :param _items: идентификаторы записей в профиле, которые подлежат удалению
    """

    _cmd_id = 8
    profile_id = property(fget=lambda self: self.pack[0])
    items = property(fget=lambda self: self.pack[1])

    def __init__(self, dev_eui, _profile_id, _items: list, **kwargs):
        super(RemoveProfileItem, self).__init__(dev_eui, [self._cmd_id, int(_profile_id), list(_items)], **kwargs)


class GlobalTimedLevel(SetProfile):
    """Команда установки глобального уровня диммирования на N секунд"""
    _cmd_id = 9
    _profile_class = profiles.GlobalTimedLevel


class ScanLine(LCommand):
    """Команда сканирования линии DALI"""
    _cmd_id = 12


class PersonalTimedLevel(SetProfile):
    """Команда установки персонального уровня диммирования на N секунд"""
    _cmd_id = 21
    _profile_class = profiles.PersonalTimedLevel


class ClearProfile(LCommand):
    """Очистка/удаление профиля диммирования

    При получении данной команды устройство должно удалить все записи из профиля диммирования,
    т.о. профиль перестает существовать. Если указанный профиль является глобальным и активным в настоящий момент,
    то устройство должно перейти на использование профиля 0

    :param dev_eui: идентификатор контроллера
    :param ids: идентификаторы профиля
    """
    _cmd_id = 10

    def __init__(self, dev_eui, ids: list, **kwargs):
        try:
            super(ClearProfile, self).__init__(dev_eui, [self._cmd_id, [int(i) for i in ids]], **kwargs)
        except BaseException as ex:
            raise ValueError("profile_id required as int")

    ids = property(fget=lambda self: self._pack[0])


class SetActiveProfile(LCommand):
    """Команда установки активного глобального профиля диммирования

    :param dev_eui: идентификатор контроллера
    :param _id: идентификатор профиля
    """
    _cmd_id = 20
    id = property(fget=lambda self: self._pack[0])

    def __init__(self, dev_eui, _id: int, **kwargs):
        try:
            super(SetActiveProfile, self).__init__(dev_eui, [self._cmd_id, int(_id)], **kwargs)
        except:
            raise ValueError("profile_id required as int")


class ExtendLine(LCommand):
    """Расширение линии DALI"""
    _cmd_id = 25


class GetProfileId(LCommand):
    """Запрос текущего активного глобального профиля"""
    _cmd_id = 26


class GetConfig(LCommand):
    """Запрос текущей конфигурации устройства

    :param dev_eui: идентификатор контроллера
    :param _items: идентификаторы параметра конфигурации
    """
    _cmd_id = 27

    def __init__(self, dev_eui, param_ids: [int], **kwargs):
        super(GetConfig, self).__init__(dev_eui, [self._cmd_id, list(param_ids)], **kwargs)

    def split(self, max_length: int) -> [LCommand]:
        count = self.part_count(max_length)
        for i in range(0, count):
            cmd = GetConfig(self.dev_eui, self._pack[0][i*max_length:(i+1)*max_length], mc=self.is_mc)
            cmd.immediately = i == 0
            cmd.pending = i == 0 and count > 1
            yield cmd

    def part_count(self, max_length): return math.ceil(len(self._pack[0]) / max_length)


class ClearAllProfiles(LCommand):
    """Полная очистка всех профилей"""
    _cmd_id = 99


"""Answers
 https://docs.google.com/document/d/18l70dzPSCUIDfsLIdP_LjyV6RtVBtqlP2eaA_djga4k/edit#heading=h.zf2tqqun8v7x"""


class LAnswer(LFrame):
    _funit_type = []

    def __init__(self, dev_eui, pack, **kwargs):
        self._cmd_id = pack[0]
        super(LAnswer, self).__init__(dev_eui, pack, **kwargs)

    @property
    def success(self): return True

    @property
    def failed(self): return False

    @property
    def valid(self): return True

    def __str__(self):
        if self.__class__ == LAnswer:
            cl = utils.get_subclass(LCommand, lambda sc: issubclass(sc, LCommand) and sc._cmd_id == self._cmd_id)
            name = cl.__name__ if cl else self.__class__.__name__
            return f'{name}<{self.dev_eui}>: { "SUCCESS" if self.success else "FAILED" if self.failed else "SERVICE" } {self.pack}'
        else:
            return super(LAnswer, self).__str__()

    @property
    def value(self):
        return {
            self._funit_type[i]: self._pack[i] for i in range(0, len(self._pack))
            for i in range(len(self._pack))
            if i < len(self._funit_type) and self._funit_type[i] is not None
        }

    @property
    def funit_type(self) -> [str]: return self._funit_type


class ScanResponse(LAnswer):
    """Результат сканирования линии DALI"""
    _cmd_id = 12
    _funit_type = [SCAN]


class ProfileResponse(LAnswer):
    """Ответ на запрос текущего активного профиля

    Данная команда передаются на FPort = 1. Ответ должен быть отправлен в случайный тайм слот.
    """
    _funit_type = [ACTIVE_PROFILE]
    _cmd_id = 26


class ConfResponse(LAnswer):
    """Ответ на запрос текущей конфигурации устройства
    """
    _cmd_id = 27

    @property
    def value(self):
        return {parameter: value for parameter, value in self._pack}


class ConfCommit(LAnswer):
    _cmd_id = 28

    @property
    def valid(self): return 0 <= self._pack[0] <= 1

    @property
    def failed(self): return self._pack[0] == 0

    @property
    def success(self): return self._pack[0] == 1


"""Events https://docs.google.com/document/d/18l70dzPSCUIDfsLIdP_LjyV6RtVBtqlP2eaA_djga4k/edit#heading=h.qr2wbqi6qkhm"""


class LEvent(LFrame):
    _fport = 2
    _flags = []
    _funit_type = []

    def flags(self, byte: int) -> {}:
        return {self._flags[i]: bool(byte & 2**i) for i in range(len(self._flags))}

    @property
    def value(self):
        return {
            self._funit_type[i]: self._pack[i]
            for i in range(len(self._pack))
            if i < len(self._funit_type) and self._funit_type[i] is not None
        }


class LomStatusOrdinary(LEvent):
    """Очередной пакет данных

    Данная команда передаются на FPort = 2. Очередной пакет отправляется в случайный таймслот в рамках настройки
    DATA_TX_INTERVAL.
    """

    _cmd_id = 30
    _funit_type = [TIME, OPER_TIME, LATITUDE, LONGITUDE, ALTITUDE, ACTIVE_PROFILE, TEMP, TEMP_MIN, TEMP_MAX,
                   ILLUMINATION, TILT, FLAGS, None, BRIGHTNESS_LEVEL]
    _flags = ['Powered', 'Activated']

    @property
    def value(self):
        result = super(LomStatusOrdinary, self).value
        result[TIME] = str(datetime.fromtimestamp(self._pack[0])).replace(' ', 'T')
        result[LATITUDE] = self._pack[2] / 100000
        result[LONGITUDE] = self._pack[3] / 100000
        result[FLAGS] = self.flags(self._pack[11])
        return result


class LomStatusExtra(LEvent):
    """Внеочередной пакет данных

    Данная команда передаются на FPort = 2.
    В зависимости от ситуации, может отправляться как немедленно, так и в случайный таймслот.
    В ситуациях, когда в результате возникновения какой-либо проблемы, для которой генерируется внеочередной
    пакет, могут случиться коллизии, должен использоваться механизм отложенной доставки. Например: отключение питания.
    В противном случае пакет должен генерироваться и отправляться немедленно.
    Например: изменение наклона выше критического уровня.
    """

    _cmd_id = 31
    _funit_type = [TIME, OPER_TIME, ACTIVE_PROFILE, TEMP, ILLUMINATION, TILT, EXTRA_FLAGS]
    _flags = ['PowerFail', 'PowerRestored', 'HighTilt', 'Shock']

    @property
    def value(self):
        result = super(LomStatusExtra, self).value
        result[TIME] = str(datetime.fromtimestamp(self._pack[0])).replace(' ', 'T')
        result[EXTRA_FLAGS] = self.flags(self._pack[6])
        return result


class DaliStatus(LEvent):
    """Выявлены проблемы при сканировании статусов устройств сети DALI

    Данная команда передаются на FPort = 3.
    Внеочередной пакет отправляемый немедленно по факту выявления проблем в светильниках, подключенных к шине DALI.
    Пакет передается только в том случае, если битовые поля Lamp Failure, Limit Error, Power Failure установлены в 1.
    """
    _cmd_id = 32
    _fport = 3
    _funit_type = [DALI_STATUS]
    _flags = ['ControlGear', 'LampFailure', 'LightOn', 'LimitError', 'FadeRunning', 'ResetState', 'MissingSA', 'PowerFailure']

    def value(self):
        return {self._pack[2*n]: self.flags(self._pack[0]) for n in range(int(len(self._pack)/2))}


def event_of(dev_eui, pack):
    if len(pack) > 0:
        for sc in LEvent.__subclasses__():
            if issubclass(sc, LEvent) and sc._cmd_id == pack[0]:
                return sc(dev_eui, pack)

    return LEvent(dev_eui, pack)


def request_of(dev_eui, pack):
    if len(pack) > 0:
        res = utils.get_subclass(LCommand, lambda sc: issubclass(sc, LCommand) and sc._cmd_id == pack[0])
        if res:
            return res(dev_eui, *res.frame_to_data(pack))

    return LCommand(dev_eui, pack)


def response_of(dev_eui, pack):
    if len(pack) > 0:
        for sc in LAnswer.__subclasses__():
            if issubclass(sc, LAnswer) and sc._cmd_id == pack[0]:
                return sc(dev_eui, pack)

    return LAnswer(dev_eui, pack)
