import datetime

from spread_core.tools import utils
from spread_lom import profile_items

UNSET_POSITION = (0, 0, 0)


# [CMD_ID=2] global geo	CMD_ID, PROFILE_TYPE,	PROFILE_ID,			 [ID, OFFSET_TYPE, OFFSET_VALUE, DIMMING_LEVEL]
# [CMD_ID=3] global lum	CMD_ID, PROFILE_TYPE,	PROFILE_ID, 		 [ID, ILLUMINATION, DIMMING_LEVEL]
# [CMD_ID=4] person geo	CMD_ID,                 PROFILE_ID, SA/[SN], [ID, OFFSET_TYPE, OFFSET_VALUE, DIMMING_LEVEL]
# [CMD_ID=5] person lum	CMD_ID,                 PROFILE_ID, SA/[SN], [ID, ILLUMINATION, DIMMING_LEVEL]
# [CMD_ID=6] global tim	CMD_ID, PROFILE_TYPE,	PROFILE_ID, 		 [ID, HOURS, MINUTES, DIMMING_LEVEL]
# [CMD_ID=7] person tim	CMD_ID,                 PROFILE_ID, SA/[SN], [ID, HOURS, MINUTES, DIMMING_LEVEL]

class DimmingProfile:
    _items_index = 2
    _timed = False
    _items_class = profile_items.DimmingItem
    _type = None

    id = property(fget=lambda self: self.pack[1])
    is_timed = property(lambda self: self._timed)
    pack = property(lambda self: self._pack)

    def __init__(self, *_pack: (int, list)):
        self._pack = list(_pack)

    def __str__(self):
        return f'{self.__class__.__name__}({self.id}): {" ".join(str(i) for i in self.pack)}'

    def __repr__(self): return self.__str__()

    def __eq__(self, other):
        return isinstance(other, DimmingProfile) and self.pack == other.pack

    def sorted_items(self, position):
        return sorted(self.items, key=lambda _item: _item.time_left(position))

    @property
    def items(self):
        res = []
        for item_pack in self.pack[self._items_index:]:
            item = self._items_class(*item_pack)
            res.append(item)

        return res

    def add_item(self, item: profile_items.DimmingItem):
        self._pack.append(item.pack)

    def serialize(self) -> dict: return {}


class GlobalDimmingProfile(DimmingProfile):
    """
    :param _type: тип профиля: 0 - с указанием уровня диммирования; 1 - с указанием номера DALI сцены.
    :param _id: идентификатор профиля
    :param _items: записи профиля
    """
    LEVEL_CONTROL = 0
    SCENE_CONTROL = 1
    TYPES = ['Level', 'Scene']
    operand = property(fget=lambda self: self._pack[0])

    def __init__(self, _type: int, _id: int, *_items: list):
        super(GlobalDimmingProfile, self).__init__(_type, _id, *_items)

    @classmethod
    def of_level(cls, _id: int, *_items: profile_items.DimmingItem) -> DimmingProfile:
        return cls(cls.LEVEL_CONTROL, _id, *[i.pack for i in _items])

    @classmethod
    def of_scene(cls, _id: int, *_items: profile_items.DimmingItem) -> DimmingProfile:
        return cls(cls.SCENE_CONTROL, _id, *[i.pack for i in _items])

    def serialize(self) -> dict:
        return dict(type=self._type, operand=self.TYPES[self.operand], id=self.id, items=[item.serialize() for item in self.items])

    @classmethod
    def deserialize(cls, data):
        _type = data['type']
        args = []
        if 'operand' in data:
            args.append(cls.TYPES.index(data['operand']))
        if 'id' in data:
            args.append(int(data.pop('id')))
        scl = utils.get_subclass(cls, lambda scl: issubclass(scl, cls) and scl._type == _type)
        args += [scl._items_class.deserialize(i).pack for i in data['items']]
        return scl(*args)


class PersonalDimmingProfile(DimmingProfile):
    """Добавление/изменение записей в персональном профиле диммирования

    :param dev_eui: идентификатор контроллера
    :param _id: идентификатор профиля
    :param _address: короткий адрес устройства/массив байтов серийного номера светильника (4 байта, LSB first)
    :param _items: записи профиля
    """

    _items_index = 3
    id = property(fget=lambda self: self.pack[0])
    address = property(fget=lambda self: self.pack[1])

    def __init__(self, _id: int, _address: (list, int), *_items: list):
        super(PersonalDimmingProfile, self).__init__(_id, _address, *_items)

    @classmethod
    def of_level(cls, _id: int, _address: (list, int), *items: profile_items.DimmingItem) -> DimmingProfile:
        return cls(_id, _address, *[item.pack for item in items])

    def serialize(self) -> dict:
        return dict(type=self._type, id=self.id, address=self.address, items=[item.serialize() for item in self.items])

    @classmethod
    def deserialize(cls, data):
        _type = data['type']
        args = [data['id'], data['address']]
        scl = utils.get_subclass(cls, lambda scl: issubclass(scl, cls) and scl._type == _type)
        args += [scl._items_class.deserialize(i).pack for i in data['items']]
        return scl(*args)


class UTCProfile(DimmingProfile):
    _type = 'UTC'
    _timed = True


class GlobalUTCProfile(UTCProfile, GlobalDimmingProfile):
    """Глобальный профиль диммирования на основе абсолютного времени"""
    _items_class = profile_items.UTCItem


class PersonalUTCProfile(UTCProfile, PersonalDimmingProfile):
    """Персональный профиль диммирования на основе абсолютного времени"""
    _items_class = profile_items.UTCItem


class LumProfile(DimmingProfile):
    _type = 'Lum'

    def sorted_items(self, _=None):
        return sorted(self.items, key=lambda _item: _item.lum)


class GlobalLumProfile(LumProfile, GlobalDimmingProfile):
    """Глобальный профиль диммирования на основе уровня освещенности"""
    _items_class = profile_items.LumItem


class PersonalLumProfile(LumProfile, PersonalDimmingProfile):
    """Персональный профиль диммирования на основе уровня освещенности"""
    _items_class = profile_items.LumItem


class GeoProfile(DimmingProfile):
    _type = 'Geo'
    _timed = True


class GlobalGeoProfile(GeoProfile, GlobalDimmingProfile):
    """Глобальный профиль диммирования на основе координат места установки"""
    _items_class = profile_items.GlobalGeoItem

    @classmethod
    def default(cls):
        return cls.of_level(0, profile_items.GlobalGeoItem.of_sunrise(1, 0, 0), profile_items.GlobalGeoItem.of_sunset(2, 0, 254))


class PersonalGeoProfile(GeoProfile, PersonalDimmingProfile):
    """Персональный профиль диммирования на основе координат места установки"""
    _items_class = profile_items.GeoItem


class TimedLevelProfile(DimmingProfile):
    """
    :param dev_eui: идентификатор контроллера
    :param _items: записи профиля
    """
    _type = 'Temp'
    _items_index = 0
    id = -1

    def __init__(self, *pack: int):
        super(TimedLevelProfile, self).__init__(*pack)
        self.start_time = datetime.datetime.now(datetime.timezone.utc)

    @property
    def items(self):
        return [self._items_class(*self.pack)]

    @classmethod
    def of_level(cls, *_items: profile_items.DimmingItem) -> DimmingProfile:
        return cls(*[i.pack for i in _items])

    def serialize(self) -> dict:
        return dict(type=self._type, items=[dict(period=item.period, value=item.value) for item in self.items if isinstance(item, profile_items.TimedLevel)])


class GlobalTimedLevel(TimedLevelProfile, GlobalDimmingProfile):
    """Профиль глобального уровня диммирования на N секунд"""
    _items_class = profile_items.TimedLevel

    def __init__(self, *pack: list):
        if not isinstance(pack[0], int):
            pack = pack[0]
        TimedLevelProfile.__init__(self, *pack)

    @classmethod
    def of_level(cls, _item: profile_items.DimmingItem, **kwargs) -> DimmingProfile:
        return cls(_item.pack)

    @classmethod
    def of_scene(cls, *_items: profile_items.DimmingItem) -> DimmingProfile:
        raise NotImplemented()


class PersonalTimedLevel(TimedLevelProfile, PersonalDimmingProfile):
    """Профиль персонального уровня диммирования на N секунд"""
    _items_class = profile_items.TimedLevelAddressed
