import datetime
import logging
import threading

from spread_core.tools import settings
from spread_core.tools.base_interface import AbstractInterfaceClass
from .commands import MCommand, OpenSession, CloseSession
from .errors import TimeOutError, AccessError


class _Interface(AbstractInterfaceClass):
    SESSION_TIME = 230

    def __init__(self, mqtt_client, p1='111111', p2='222222'):
        super(_Interface, self).__init__(mqtt_client)
        self._passwords = {1: p1, 2: p2}
        self._deprecate_time = {}
        self.has_failed = []
        self.hexed_passwords = []

    def update_config(self):
        self.SESSION_TIME = settings.config.get(settings.SESSION_TIME, self.SESSION_TIME)

        has_hexed = list(self.hexed_passwords)
        for hexed_address in settings.config.get(settings.HEXED, []):
            if hexed_address not in has_hexed:
                self.set_hexed(hexed_address, True)
            else:
                has_hexed.remove(hexed_address)

        for has_hexed_address in has_hexed:
            self.set_hexed(has_hexed_address, False)

    def check_access(self, address, level):
        ts = datetime.datetime.now().timestamp()
        if level > 0:
            if ts - self._deprecate_time.get(address, {level: 99999})[level] >= self.SESSION_TIME:
                _open_session = OpenSession(address, level, self._passwords[level], address in self.hexed_passwords)
                logging.debug(f'  -> {_open_session}')
                r = self._apply(_open_session)
                if r is None or not r.success:
                    raise AccessError(address, level)
                logging.debug(f'    <- {r}')

                self._deprecate_time.setdefault(address, {})
                self._deprecate_time[address][level] = ts

    def set_hexed(self, address: int, hexed: bool):
        if hexed and address not in self.hexed_passwords:
            self.hexed_passwords.append(address)
        elif not hexed and address in self.hexed_passwords:
            self.hexed_passwords.remove(address)

    def send(self, cmd: MCommand):
        logging.debug(f'-> {cmd}')
        if cmd.is_timed:
            self.check_access(cmd.addr_int, cmd.access_level)
        result = self._apply(cmd)
        logging.debug(f'  <- {result}')
        return result

    def _apply(self, cmd):
        if isinstance(cmd, CloseSession):
            self._deprecate_time[cmd.addr_int] = {}


class BrokerInterface(_Interface):
    def __init__(self, mqtt_client):
        super().__init__(mqtt_client)
        self.response = None
        self._response_event = threading.Event()

    def set_event_handler(self, _): pass
    def set_external_cmd_handler(self, _): pass
    def set_status_handler(self, _): pass

    def on_response(self, data):
        self.response = bytes.fromhex(data.split('#')[0])
        if not self._response_event.is_set():
            self._response_event.set()

    def _apply(self, cmd: MCommand):
        super()._apply(cmd)
        data = ''.join(hex(b).replace('0x', '').rjust(2, '0') for b in cmd.pack).upper()
        flags = 'RS{}'.format(3 + cmd.framesize)

        self._response_event.clear()
        self.response = None

        self._mqtt_client.publish(
            topic=self._send_topic,
            payload='#'.join([data, flags]))

        self._response_event.wait(self.timeout)
        if self.response:
            return cmd.set_response(self.response)
        else:
            raise TimeOutError(cmd)


# i = SocketInterface('10.10.1.32', 4001, p1='111111', p2='222222')
# res = i.send(SerialCommand(30))
# print(res)
# res = i.send(CurrentDateTime(30))
# print(res)
# # res = i.send(OpenCloseTime(30, 0))
# res = i.send(StoredEnergy(30, period=0x3, tariff=1, month=10))
# powerP0 = i.send(AUXPower(30, power_id=0, phase=0))
# rate1 = i.send(AUXPowerRate(30, phase=0))
# voltage1 = i.send(AUXVoltage(30, phase=1))
# amperage1 = i.send(AUXAmperage(30, phase=1))
# angle12 = i.send(AUXAngle(30, phase=1))
# frequency = i.send(AUXFrequency(30))
# temp = i.send(AUXTemp(30))
# res = i.send(ReadLoadState(30))
# res = i.send(SetLoad(30, True))
# res = i.send(CloseSession(30))
