import json
import logging

import spread_core.bam as bam
import spread_core.mqtt as mqtt
from spread_core.errors.project_errors import ProjectError, ProviderNotPresent, EngineryNotPresent, ClassifierError
from spread_core.tools.manager_interface import ManagerOfBroker

PROVIDERS = 'providers'
ID = 'id'
TYPE = 'type'
MANAGER_ID = 'managerID'
RECIPE = 'recipe'
INGREDIENTS = 'ingredients'


def log_command(fn):
    def log(self, topic, variable):
        logging.debug(
            f'COMMAND({variable.session_id if variable.session_id else ""}): {topic.__class__.__name__} {topic.entity_address.type}({topic.entity_address.id})'
            f'.{topic.entity_address.funit_type}' +
            (f'={variable.value}' if isinstance(topic, mqtt.spread.topic.Set) else '') +
            (f' params: {variable.params}' if variable.params else '')
        )
        fn(self, topic, variable)

    return log


class LightingEquipment(ManagerOfBroker):
    EVENT_TOPICS = ('CurrentLuminosity', 'CurrentPresence', 'LastEvent')
    STATE_TOPICS = ('On', 'BrightnessLevel', 'Temperature', 'EngineryHoldTime') + EVENT_TOPICS

    def __init__(self, project_id, mqttc):
        super(LightingEquipment, self).__init__(mqttc)
        self.id = project_id
        self._managers = dict()
        self._subgineries = dict()
        self._engineries = {}
        self._lighting_areas = dict()
        self.loaded = False
        self.button_presets = {}

    def start(self):
        if self.loaded:
            for s in (self._engineries, self._subgineries, self._lighting_areas):
                for e in s.values():
                    values = {**e.values}
                    e.values = {}
                    for funit_type, value in values.items():
                        e.publish_state(funit_type=funit_type, value=value)

            self.exploit_topics(self.subscribe)
        else:
            self.subscribe(mqtt.project.TopicProject.managers(self.id))

    def on_local(self, topic: mqtt.local.LAButtonPresets, data):
        if data:
            self.button_presets.setdefault(topic.entity_id, bam.engineries.lighting_area.ButtonPreset(topic.entity_id))
            self.button_presets[topic.entity_id].update(json.loads(data))
        else:
            self.button_presets.pop(topic.entity_id)

    def on_project(self, topic, data):
        self.unsubscribe(topic)
        data = json.loads(data)
        if mqtt.project.TopicProject.MANAGERS in data:
            for item in data[mqtt.project.TopicProject.MANAGERS]:
                if item[ID] not in self._managers:
                    self._managers[item[ID]] = {'data': item, PROVIDERS: {}}
            self.subscribe(mqtt.project.TopicProject.providers(self.id))
        elif mqtt.project.TopicProject.PROVIDERS in data:
            for item in data[mqtt.project.TopicProject.PROVIDERS]:
                if item[MANAGER_ID] in self._managers:
                    if item[ID] not in self._managers[item[MANAGER_ID]][PROVIDERS]:
                        manager = self._managers[item[MANAGER_ID]]
                        manager[PROVIDERS][item[ID]] = bam.Recipe(self.id, item[ID], item[TYPE], manager['data'])
            self.subscribe(mqtt.project.TopicProject.subgineries(self.id))
        elif mqtt.project.TopicProject.SUBGINERIES in data:
            for s_data in data[mqtt.project.TopicProject.SUBGINERIES]:
                try:
                    self._subgineries[s_data[ID]] = bam.generator.generate_subginery(self.id, s_data)
                    self._subgineries[s_data[ID]].set_adapter(self)
                except ProjectError as ex:
                    logging.warning(ex)
            self.subscribe(mqtt.project.TopicProject.engineries(self.id))
        elif mqtt.project.TopicProject.ENGINERIES in data:
            try:
                for e_data in data[mqtt.project.TopicProject.ENGINERIES]:
                    e_id = e_data[ID]
                    if RECIPE in e_data and TYPE in e_data[RECIPE] and INGREDIENTS in e_data[RECIPE]:
                        if e_data[RECIPE][TYPE] == 'Simple' and len(e_data[RECIPE][INGREDIENTS]) > 0:
                            try:
                                _eng = bam.generate_enginery(self.id, e_data)
                                _eng.set_adapter(self)
                                self._engineries[e_id] = _eng
                            except ProjectError as ex:
                                logging.warning(ex)
                                continue
                            else:
                                for subginery in self._subgineries.values():
                                    if subginery.has_enginery(e_id):
                                        _eng.set_subginery(subginery)
                                        subginery.set_enginery(_eng)
                                for manager in self._managers.values():
                                    recipe = manager[PROVIDERS].get(e_data[RECIPE][INGREDIENTS][0][ID])
                                    if recipe and isinstance(recipe, bam.Recipe):
                                        recipe.set_enginery(_eng)
                                        _eng.set_provider(recipe)
                                        break
                for e_data in data[mqtt.project.TopicProject.ENGINERIES]:
                    if RECIPE in e_data and TYPE in e_data[RECIPE] and INGREDIENTS in e_data[RECIPE]:
                        if e_data[RECIPE][TYPE] == 'Piecemeal' and e_data[TYPE] == bam.engineries.LightingArea.__name__:
                            la = bam.generate_enginery(self.id, e_data)
                            la.set_adapter(self)
                            for p_data in sorted(e_data[RECIPE][INGREDIENTS], key=lambda ing: ing.get('index', -1)):
                                for manager in self._managers.values():
                                    recipe = manager[PROVIDERS].get(p_data[ID])
                                    if recipe and isinstance(recipe, bam.Recipe):
                                        recipe.set_la(la)
                                        la.set_ingredient(p_data[TYPE], recipe)
                                        break
                            self._lighting_areas[e_data[ID]] = la
            except BaseException as ex:
                logging.exception(ex)
            finally:
                self.on_ready()

    def on_ready(self):
        self.loaded = True
        logging.info('Project({}) loaded'.format(self.id))
        for _la in self._lighting_areas.values():
            _la.get_info()

        self.exploit_topics(self.subscribe)

    def exploit_topics(self, action):
        """Make action with topics"""

        action(mqtt.local.LAButtonPresets.broadcast(self.id))

        for fut in self.STATE_TOPICS:
            action(mqtt.spread.topic.State(mqtt.spread.address.ProviderAddress(self.id, '+', '+', '+', '+', funit_type=fut)))

        for entity in [
            bam.engineries.DimmingLight, bam.engineries.SwitchingLight, bam.engineries.TunableWhiteLight,
            bam.engineries.LightingArea,
            bam.subgineries.Lighting
        ]:

            for funit in entity.settable_units or []:
                action(mqtt.spread.topic.Set(entity.get_address_broadcast(self.id, funit)))

            for funit in entity.gettable_units or []:
                action(mqtt.spread.topic.Renew(entity.get_address_broadcast(self.id, funit)))

    @log_command
    def on_command(self, topic: mqtt.spread.topic.SpreadTopic, variable: mqtt.spread.variable.Variable):
        e_id = int(topic.entity_address.id)

        try:
            if topic.entity_address.type == bam.engineries.LightingArea.__name__ and e_id in self._lighting_areas:
                self._lighting_areas[e_id].on_command(topic, variable)
            elif isinstance(topic.entity_address, mqtt.spread.address.SubgineryAddress) and e_id in self._subgineries:
                self._subgineries[e_id].on_command(topic, variable)
            elif isinstance(topic.entity_address, mqtt.spread.address.EngineryAddress) and e_id in self._engineries:
                self._engineries[e_id].on_command(topic, variable)
            else:
                raise EngineryNotPresent(f'{topic.entity_address.type}({e_id}) is not present')
        except (EngineryNotPresent, ClassifierError) as ex:
            # if variable.session_id:
            #     to_topic = mqtt.spread.topic.Reply(self.id, variable.session_id)
            #     self.publish(to_topic, dict(error=dict(code=123, message='{}'.format(ex))), False)
            logging.debug(ex)

    def on_state(self, topic: mqtt.spread.topic.State, variable: mqtt.spread.variable.Variable):
        m_id = int(topic.entity_address.manager_id)
        if m_id in self._managers:
            recipe = self._managers[m_id][PROVIDERS].get(topic.entity_address.id)
            if recipe:
                for eng in recipe.engs:
                    eng.on_state(topic, variable)
                for la in recipe.las:
                    la.on_state(topic, variable)
        else:
            raise ProviderNotPresent(topic.entity_address.type, topic.entity_address.id)

    def on_event(self, topic: mqtt.spread.topic.State, variable: mqtt.spread.variable.Variable):
        if variable.invalid:
            return

        m_id = topic.entity_address.manager_id
        if m_id in self._managers:
            p_id = topic.entity_address.id
            for _la in self._lighting_areas.values():
                if p_id in _la.luminosities:
                    _la.on_luminosity_event(p_id, variable)

                elif p_id in _la.presences:
                    _la.on_presence_event(p_id, variable)

                elif p_id in _la.buttons:
                    _la.on_button_event(p_id, variable)
        else:
            raise ProviderNotPresent(topic.entity_address.type, topic.entity_address.entity_id)

    def on_exit(self):
        super(LightingEquipment, self).on_exit()
        self.exploit_topics(self.unsubscribe)

        for stack in [self._engineries.values(), self._lighting_areas.values()]:
            for eng in stack:
                try:
                    eng.set_invalid()
                except BaseException as ex:
                    logging.exception(ex)
