import spread_core.protocols.dali as dali
from ..payloads import Payload


class DaliPayload(Payload):
    @staticmethod
    def of(frame: bytes, query_class=None):
        if len(frame) < 2:
            return DaliResponse(query_class, frame)
        elif len(frame) == 2:
            return DaliCommand(frame)
        elif len(frame) == 3:
            if frame[0] & 1 == 0:
                return Dali2Event(frame)
            else:
                return Dali2Command(frame)

        raise ValueError('frame size more than 3')

    @property
    def command(self) -> dali.command.Command:
        if not hasattr(self, '_command'):
            self._command = dali.command.from_frame(dali.frame.ForwardFrame(8*len(self.frame), self.frame))
        return self._command

    @property
    def has_address(self):
        return False

    @property
    def has_instance(self):
        return False

    @property
    def address(self):
        raise ValueError()

    @property
    def instance(self):
        raise ValueError()

    @staticmethod
    def pack_instance(inst) -> str:
        if isinstance(inst, dali.address.InstanceNumber):
            return f'IN{inst._value}'
        elif isinstance(inst, dali.address.InstanceType):
            return f'IT{inst._value}'
        elif isinstance(inst, dali.address.InstanceGroup):
            return f'IG{inst._value}'
        elif isinstance(inst, dali.address.InstanceBroadcast):
            return 'IB'
        elif isinstance(inst, dali.address.FeatureInstanceType):
            return f'FT{inst._value}'
        elif isinstance(inst, dali.address.FeatureInstanceGroup):
            return f'FG{inst._value}'
        elif isinstance(inst, dali.address.FeatureInstanceNumber):
            return f'FN{inst._value}'
        elif isinstance(inst, dali.address.FeatureDevice):
            return f'FD{inst._value}'
        elif isinstance(inst, dali.address.FeatureInstanceBroadcast):
            return f'FB'
        return 'X3'

    @staticmethod
    def pack_address(addr) -> str:
        if isinstance(addr, dali.address.Short):
            return f'D{addr.address}'
        elif isinstance(addr, dali.address.Group):
            return f'G{addr.group}'
        elif isinstance(addr, dali.address.Broadcast):
            return 'B'
        elif isinstance(addr, dali.address.BroadcastUnaddressed):
            return 'BU'
        return 'X3'


class DaliCommand(DaliPayload):
    mark = 'dali'

    @property
    def has_address(self):
        return isinstance(self.command, (dali.gear.general._StandardCommand, dali.gear.general.DAPC))

    @property
    def has_params(self):
        if isinstance(self.command, dali.gear.general.DAPC):
            return True
        return isinstance(self.command, dali.gear.general._StandardCommand) and self.command._hasparam

    @property
    def params(self):
        if isinstance(self.command, dali.gear.general.DAPC):
            return [self.command.power]
        return [self.command.param]

    @property
    def has_instance(self):
        return False

    @property
    def instance(self):
        return None

    @property
    def address(self):
        return self.pack_address(self.command.destination)

    @property
    def desc(self) -> str:
        if self.has_params:
            return ' '.join([self.name, ' '.join([str(p) for p in self.params])])
        return self.name

    @property
    def name(self) -> str:
        return self.command.__class__.__name__


class Dali2Command(DaliCommand):
    mark = 'dali2'

    @property
    def has_address(self):
        return isinstance(self.command, dali.device.general._DeviceCommand) and not isinstance(self.command, dali.device.general._SpecialDeviceCommand)

    @property
    def has_instance(self):
        return isinstance(self.command, dali.device.general._StandardInstanceCommand)

    @property
    def has_params(self):
        return isinstance(self.command, (dali.device.general._SpecialDeviceCommandOneParam, dali.device.general._SpecialDeviceCommandTwoParam))

    @property
    def params(self):
        if isinstance(self.command, dali.device.general._SpecialDeviceCommandOneParam):
            return [self.command._opcode]
        elif isinstance(self.command, dali.device.general._SpecialDeviceCommandTwoParam):
            return [self.command._instance, self.command._opcode]
        else:
            return []

    @property
    def instance(self):
        return self.pack_instance(self.command.instance)


class DaliResponse(DaliPayload):
    def __init__(self, query: dali.command.Command, *args, **kwargs):
        super(DaliResponse, self).__init__(*args, **kwargs)
        self.query = query

    @property
    def is_dali2(self):
        return isinstance(self.query, dali.device.general._DeviceCommand)

    @property
    def has_address(self):
        if isinstance(self.query, dali.gear.general._GearCommand):
            return isinstance(self.query, (dali.gear.general._StandardCommand, dali.gear.general.DAPC))
        elif isinstance(self.query, dali.device.general._DeviceCommand):
            return not isinstance(self.query, dali.device.general._SpecialDeviceCommand)
        return False

    @property
    def address(self):
        return self.pack_address(self.query.destination)

    @property
    def instance(self):
        return self.pack_instance(self.query.instance)

    @property
    def has_instance(self): return isinstance(self.query, dali.device.general._StandardInstanceCommand)

    @property
    def name(self) -> str:
        if self.query:
            return self.query.__class__.__name__
        return ''

    @property
    def mark(self):
        if self.query:
            if self.is_dali2:
                return Dali2Command.mark
            else:
                return DaliCommand.mark
        return 'response'

    @property
    def desc(self) -> str:
        if len(self.frame) == 0:
            return 'empty'
        elif self.query and self.query.response:
            frame = dali.frame.BackwardFrame(self.frame[0])
            resp = self.query.response(frame)
            if isinstance(resp, (dali.gear.general.QueryFadeTimeAndRateResponse, dali.gear.general.QueryDeviceTypeResponse)):
                return str(resp)
            elif isinstance(resp, dali.command.BitmapResponse):
                return bin(resp.value.as_integer)[2:].rjust(8, '0')
            elif isinstance(resp, dali.command.YesNoResponse):
                return str(resp.value)
            else:
                return str(resp.value.as_integer)
        else:
            return super(DaliResponse, self).desc


class Dali2Event(DaliPayload):
    mark = 'event'

    d1 = d2 = val = _address = _instance = None

    def __init__(self, *args, **kwargs):
        super(Dali2Event, self).__init__(*args, **kwargs)
        self.__parse()

    @property
    def has_address(self):
        return self._address is not None

    @property
    def address(self):
        return self.pack_address(self._address)

    @property
    def has_instance(self):
        return self._instance is not None

    @property
    def instance(self):
        return self.pack_instance(self._instance)

    @property
    def desc(self) -> str:
        return ' '.join([self.name, str(self.val)])

    @property
    def name(self) -> str: return self._name

    def __parse(self):
        frame = self.frame

        if frame[0] >> 7 == 0x0:
            short = frame[0] >> 1
            self.d1 = f'D{short}'
            self._address = dali.address.Short(short)
            if frame[1] >> 7 == 0x0:
                tp = frame[1] >> 2
                self._instance = dali.address.InstanceType(tp)
                self.d2 = f'IT{tp}'
            elif frame[1] >> 7 == 0x1:
                number = (frame[1] & 0x7f) >> 2
                self._instance = dali.address.InstanceNumber(number)
                self.d2 = f'IN{number}'
        elif frame[0] >> 6 == 0x2:
            if frame[1] >> 7 == 0x0:
                group = (frame[0] & 0x3f) >> 1
                self.d1 = f'G{group}'
                self._address = dali.address.Group(group)
                tp = (frame[1] & 0x7f) >> 2
                self.d2 = f'IT{tp}'
                self._instance = dali.address.InstanceType(tp)
            elif frame[1] >> 7 == 0x1:
                self.tp = (frame[0] & 0x3f) >> 1
                self._instance = dali.address.InstanceType(self.tp)
                self.d1 = f'IT{self.tp}'
                self.d2 = f'IN{(frame[1] & 0x7f) >> 2}'
        elif frame[0] >> 6 == 0x3:
            if frame[1] >> 7 == 0x0:
                self.d1 = f'IG{(frame[0] & 0x3f) >> 1}'
                self.tp = (frame[1] & 0x7f) >> 2
                self.d1 = f'IT{self.tp}'
                self._instance = dali.address.InstanceType(self.tp)

        if isinstance(self._instance, dali.address.InstanceType) and self._instance._value == 3:
            self.val = frame[2] & 0x03 != 0x00
            self._name = 'Presence'
        else:
            self.val = int.from_bytes(frame, 'big') & int('1'*10, 2)
            self._name = 'Luminosity'
